namespace Brjnk.Common
{
    using System;
    using System.Collections.Concurrent;
    using System.Threading;

    /// <summary>
    /// </summary>
    public class AsyncProcessing<T> : IDisposable
    {
        private AsyncProcessingMode mode;
        private ConcurrentQueue<AsyncTask<T>> tasks;
        private AutoResetEvent syncEvent;
        private Thread thread;
        private SynchronizationContext callbackSyncContext;

        public AsyncProcessing(AsyncProcessingMode mode, SynchronizationContext callbackSyncContext)
        {
            if (mode != AsyncProcessingMode.Thread && mode != AsyncProcessingMode.ThreadPool)
                throw new ArgumentOutOfRangeException("mode", "AsyncProcessing: Hodnota mode typu AsyncProcessingMode nen platn");

            tasks = new ConcurrentQueue<AsyncTask<T>>();
            this.mode = mode;
            this.callbackSyncContext = callbackSyncContext;
            startThreadRequired = mode == AsyncProcessingMode.ThreadPool;
            if (mode == AsyncProcessingMode.Thread)
            {
                thread = new Thread(new ThreadStart(() => Loop()));
                thread.Start();
                syncEvent = new AutoResetEvent(false);
            }
        }

        public AsyncProcessing(AsyncProcessingMode mode)
            : this(mode, null)
        { }

        private object lck = new Object();

        public void StartAsync(Func<T> func, Action<T, object> callback, object userObject, Action<Exception, object> exceptionCallback)
        {
            lock (lck)
            {
                if (disposed) throw new ObjectDisposedException("AsyncProcessing");
                tasks.Enqueue(CreateAsyncTask(func, callback, userObject, exceptionCallback));
                //pokud neb smyka, tak j spustit
                //pi modu AsyncProcessingMode b smyka pod
                if (startThreadRequired && mode == AsyncProcessingMode.ThreadPool)
                {
                    ThreadPool.QueueUserWorkItem(new WaitCallback((o) => Loop()));
                    startThreadRequired = false;
                }
                if (mode == AsyncProcessingMode.Thread)
                    syncEvent.Set();
            }
        }

        public void StartAsync(Func<T> func, Action<T, object> callback)
        { StartAsync(func, callback, null, null); }

        private AsyncTask<T> CreateAsyncTask(Func<T> func, Action<T, object> callback, object userObject, Action<Exception, object> exceptionCallback)
        {
            if (object.ReferenceEquals(callbackSyncContext, null))
                return new AsyncTask<T>(func, SynchronizationContext.Current, callback, userObject, exceptionCallback);
            else
                return new AsyncTask<T>(func, callbackSyncContext, callback, userObject, exceptionCallback);
        }

        private volatile bool startThreadRequired;

        private void Loop()
        {
            startThreadRequired = false;
            bool lastTask = false;
            while (true)
            {
                if (lastTask) break;
                if (disposed) { return; };

                AsyncTask<T> task;
                lock (lck)
                {
                    task = PeekOrDequeue(false);
                    if (mode == AsyncProcessingMode.ThreadPool && tasks.IsEmpty) { lastTask = true; startThreadRequired = true; };
                }

                if (task == null) //ve front nen dn task
                {
                    //pokud pracuje s ThreadPoolem, tak ho nebude blokovat a uvoln vlkno
                    if (mode == AsyncProcessingMode.ThreadPool) break;
                    //pokud pracuje s vlastnm vlknem, bude pozastaveno
                    else if (mode == AsyncProcessingMode.Thread) syncEvent.WaitOne();
                }
                else
                {
                    try
                    {
                        task.Start();
                    }
                    finally
                    {
                        //Odstrann fronty probh a po dokonen lohy.
                        //Kdyby byla fronta przdn a zrove jet probhala loha,
                        //pidn dal lohy by zpsobilo pout dalho vlkna z ThreadPoolu
                        var a = PeekOrDequeue(true);
                    }
                }
            }
        }

        private AsyncTask<T> PeekOrDequeue(bool dequeue)
        {
            AsyncTask<T> task;
            bool success;
            if (dequeue) success = tasks.TryDequeue(out task);
            else success = tasks.TryPeek(out task);
            if (success) return task;
            return null;
        }

        private volatile bool disposed = false;

        public void Dispose()
        {
            disposed = true;
            //pokud vlkno ek, odblokuje se aby se ukonilo
            if (mode == AsyncProcessingMode.Thread) syncEvent.Set();
        }

        /// <summary>
        /// Spust funkci ve vlkn ThreadPool. Funkce jsou vykonny paraleln ve vlknech ThreadPoolu.
        /// </summary>
        public static void StartAsyncParallel(Func<T> func, Action<T, object> callback, object userObject = null, Action<Exception, object> exceptionCallback = null)
        {
            var task = new AsyncTask<T>(func, SynchronizationContext.Current, callback, userObject, exceptionCallback);
            ThreadPool.QueueUserWorkItem(new WaitCallback(t =>
            {
                ((AsyncTask<T>)task).Start();
            }), task);
        }
    }

    public enum AsyncProcessingMode
    {
        /// <summary>
        /// Pouv vlastn vytvoen vlkno
        /// </summary>
        Thread,

        ///<summary>
        /// Pouv vlkno z ThreadPoolu. Voln v rmci jedn instance tdy AsyncProcessing nejsou volny paraleln.
        ///</summary>
        ThreadPool
    }
}