namespace Brjnk.MVVM
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Collections.Specialized;
    using System.Linq;
    using Brjnk.Common;

    /// <summary>
    /// Zaobal poloky kolekce. Kolekce je aktualizovna.
    /// Pi NotifyCollectionChangedAction.Reset jsou vechny poloky znovu vytvoeny!
    /// </summary>
    public class CollectionWrapper<TSource, TWrapper> : ICollection<TWrapper>, IAddRemoveEvents, INotifyCollectionChanged, IList<TWrapper> where TWrapper : IWrapper<TSource>
    {
        private ObservableCollection<TSource> source;
        private List<TWrapper> wrapped = new List<TWrapper>();

        private WeakEventListener listener;

        public CollectionWrapper(ObservableCollection<TSource> source)
        {
            this.source = source;
            listener = new WeakEventListener(Changed);
            CollectionChangedEventManager.AddListener(source, listener);
            Refresh();
        }

        private bool Changed(Type managerType, object sender, EventArgs e)
        {
            var args = e as NotifyCollectionChangedEventArgs;
            switch (args.Action)
            {
                case NotifyCollectionChangedAction.Add:
                    args = AddItems(args.NewItems, args.NewStartingIndex);
                    break;

                case NotifyCollectionChangedAction.Move:
                    if (args.NewItems.Count != 1 || args.OldItems.Count != 1) throw new NotSupportedException("Nen podporovno Move vce prvk.");
                    args = Move(args.OldStartingIndex, args.NewStartingIndex);
                    break;

                case NotifyCollectionChangedAction.Remove:
                    args = Remove(args.OldStartingIndex, args.OldItems.Count);
                    break;

                case NotifyCollectionChangedAction.Replace:
                    args = Replace(args.NewItems, args.NewStartingIndex);
                    break;

                case NotifyCollectionChangedAction.Reset:
                    args = Refresh();
                    break;

                default:
                    throw new Exception("Neplatn NotifyCollectionChangedAction");
            }
            OnCollectionChanged(args);
            return true;
        }

        private NotifyCollectionChangedEventArgs Refresh()
        {
            foreach (var item in wrapped) OnItemRemoved(item);

            for (int i = wrapped.Count - 1; i >= 0; i--)
            {
                var item = wrapped[i];
                wrapped.RemoveAt(i);
                //OnItemRemoved(item);
            }

            for (int i = 0; i < source.Count; i++)
            {
                var newItem = Wrap(source[i]);
                wrapped.Add(newItem);
                OnItemAdded(newItem);
            }
            return new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset);
        }

        private NotifyCollectionChangedEventArgs Remove(int index, int count)
        {
            IList removed = new List<TWrapper>();
            for (int i = index; i < index + count; i++)
            {
                removed.Add(wrapped[i]);
            }

            wrapped.RemoveRange(index, count);

            foreach (var item in removed) OnItemRemoved((TWrapper)item);
            return new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Remove, removed, index);
        }

        private NotifyCollectionChangedEventArgs Move(int oldIndex, int newIndex)
        {
            var obj = wrapped[oldIndex];
            wrapped.RemoveAt(oldIndex);
            wrapped.Insert(newIndex, obj);
            return new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Move, obj, newIndex, oldIndex);
        }

        private NotifyCollectionChangedEventArgs AddItems(System.Collections.IList newItems, int startingIndex)
        {
            //pout objekt z promn adding, pokud je dostupn
            if (newItems.Count == 1 && adding != null && object.ReferenceEquals(adding.GetWrappedObject(), newItems[0]))
            {
                wrapped.Insert(startingIndex, adding);
                OnItemAdded(adding);
                adding = default(TWrapper);
                return new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, adding, startingIndex);
            }

            //pidat nov msto do kolekce
            for (int i = 0; i < newItems.Count; i++)
                wrapped.Add(default(TWrapper));

            //pesunout poloky za vkldamm mstem (za startingIndex) na konec (do vytvoenho msta)
            for (int i = wrapped.Count - newItems.Count - 1; i >= startingIndex; i--)
                wrapped[i + newItems.Count] = wrapped[i];

            List<TWrapper> newWrappers = new List<TWrapper>();

            //vloit nov poloky
            for (int i = 0; i < newItems.Count; i++)
            {
                var newObject = Wrap((TSource)newItems[i]);
                wrapped[i + startingIndex] = newObject;
                newWrappers.Add(newObject);
                OnItemAdded(newObject);
            }

            return new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, newWrappers, startingIndex);
        }

        private NotifyCollectionChangedEventArgs Replace(System.Collections.IList newItems, int startingIndex)
        {
            if (newItems.Count == 1 && adding != null && object.ReferenceEquals(adding.GetWrappedObject(), newItems[0]))
            {
                var oldObject = wrapped[startingIndex];
                wrapped[startingIndex] = adding;
                OnItemRemoved(oldObject);
                OnItemAdded(adding);
                return new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Replace,adding, oldObject,  startingIndex);
            }

            List<TWrapper> newObjects = new List<TWrapper>();
            List<TWrapper> oldObjects = new List<TWrapper>();

            for (int i = 0; i < newItems.Count; i++)
            {
                var newObject = Wrap(source[i + startingIndex]);
                var oldObject = wrapped[i + startingIndex];
                wrapped[i + startingIndex] = newObject;
                OnItemAdded(newObject);
                OnItemRemoved(oldObject);
                oldObjects.Add(oldObject);
                newObjects.Add(newObject);
            }

            return new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Replace, newObjects, oldObjects, startingIndex);
        }

        private TWrapper Wrap(TSource item)
        {
            return (TWrapper)Activator.CreateInstance(typeof(TWrapper), new object[] { item });
        }

        private void OnItemAdded(TWrapper newObject)
        {
            if (ItemAdded != null) ItemAdded(newObject);
            if (ItemsAddedUntyped != null) ItemsAddedUntyped(this, new ItemsEventArgs(Enumerable.Repeat(newObject, 1)));
        }

        private void OnItemRemoved(TWrapper obj)
        {
            if (ItemRemoved != null) ItemRemoved(obj);
            if (ItemsRemovedUntyped != null) ItemsRemovedUntyped(this, new ItemsEventArgs(Enumerable.Repeat(obj, 1)));
        }

        protected void OnCollectionChanged(NotifyCollectionChangedEventArgs e)
        {
            if (CollectionChanged != null) CollectionChanged(this, e);
        }

        public event NotifyCollectionChangedEventHandler CollectionChanged;

        public event Action<TWrapper> ItemAdded;

        public event Action<TWrapper> ItemRemoved;

        //Pi pidvn obalu je uloen do tto promn.
        //Obalovan objekt je pak pidn do zdrojov kolkce.
        //Pi zpracovn udlosti pidn prvku do zdrojov kolekce je pak pouit objekt z tto promn.
        private TWrapper adding;

        public void Add(TWrapper item)
        {
            adding = item;
            source.Add(item.GetWrappedObject());
        }

        public void Clear()
        {
            source.Clear();
        }

        public bool Contains(TWrapper item)
        {
            return wrapped.Contains(item);
        }

        public void CopyTo(TWrapper[] array, int arrayIndex)
        {
            throw new NotImplementedException();
        }

        public int Count
        {
            get { return wrapped.Count; }
        }

        public bool IsReadOnly
        {
            get { return false; }
        }

        public bool Remove(TWrapper item)
        {
            int index = wrapped.IndexOf(item);
            if (index < 0) return false;
            source.RemoveAt(index);
            return true;
        }

        public IEnumerator<TWrapper> GetEnumerator()
        {
            return wrapped.GetEnumerator();
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return wrapped.GetEnumerator();
        }

        public TWrapper this[int index]
        {
            get { return wrapped[index]; }
            set
            {
                adding = value;
                source[index] = value.GetWrappedObject();
            }
        }

        public TWrapper FindWrapper(TSource obj)
        {
            return this.Where(s => object.ReferenceEquals(s.GetWrappedObject(), obj)).FirstOrDefault();
        }

        public event EventHandler<ItemsEventArgs> ItemsAddedUntyped;

        public event EventHandler<ItemsEventArgs> ItemsRemovedUntyped;

        public int IndexOf(TWrapper item)
        {
            for (int i = 0; i < this.Count; i++)
            {
                if (object.ReferenceEquals(this[i], item)) return i; ;
            }
            return -1;
        }

        public void Insert(int index, TWrapper item)
        {
            throw new NotImplementedException("Metoda nen pouvan, zatm nebyla implementovna");
        }

        public void RemoveAt(int index)
        {
            this.Remove(this[index]);
        }
    }
}