namespace Brjnk.MVVM
{
    using System;
    using System.Collections.Generic;
    using System.ComponentModel;
    using System.Linq;

    /// <summary>
    /// </summary>
    public class DataItem : INotifyPropertyChanged
    {
        #region Static

        /// <summary>
        /// Seznam vlastnost (PropertyInfo). Hodnoty jsou pidny metodou DefineProperty.
        /// Kolekce neobsahuje vechny vlastnosti, nejsou zjiovny automaticky (reflection).
        /// Poloky jsou pidny metodou DefinePropery nebo GetPropertyInfo.
        /// </summary>
        private static List<PropertyInfo> properties = new List<PropertyInfo>();

        /// <summary>
        /// Definuje informace o vlastnostech sledovanch objekt.
        /// Poloky jsou pidny metodou DefineSubProperty.
        /// </summary>
        private static Dictionary<Tuple<string, string>, string[]> subProperties = new Dictionary<Tuple<string, string>, string[]>();

        /// <summary>
        /// Definuje tdu (tdy) vlastnosti
        /// </summary>
        protected static void DefineProperty(string name, string[] classes)
        {
            properties.RemoveAll(p => p.Name == name);
            properties.Add(new PropertyInfo(name, classes));
        }

        protected static void DefineSubProperty(string[] classes, string name, string baseProperty)
        {
            subProperties.Add(new Tuple<string, string>(baseProperty, name), classes);
        }

        #endregion Static

        /// <summary>
        /// Seznam vlastnost, kter jsou sledovny a jejich hodnoty.
        /// </summary>
        private Dictionary<string, object> watchedProperties = new Dictionary<string, object>();

        /// <summary>
        /// Seznam sledovanch objekt (potomk v rmci stromu).
        /// K jedn vlastnosti me psluet vce sledovanch objekt (jedn-li se o kolekci)
        /// </summary>
        private Dictionary<object, WatchedObjectInfo> watchedObjects = new Dictionary<object, WatchedObjectInfo>();

        /// <summary>
        /// Rodi tohoto uzlu ve stromu
        /// </summary>
        public DataItem Parent { get; protected set; }

        private int updatingCount = 0;

        public void BeginUpdate()
        {
            updatingCount++;
            BeginChildrenUpdate();
        }

        private void BeginChildrenUpdate()
        {
            foreach (var obj in watchedObjects.Keys)
                if (obj is DataItem) (obj as DataItem).BeginUpdate();
        }

        public void EndUpdate()
        {
            EndChildrenUpdate();
            updatingCount--;
            if (updatingCount < 0) throw new InvalidOperationException("UpdatingCount je zporn");
            if (updatingCount == 0)
                OnChanged(new ChangedEventArgs(this));
        }

        private void EndChildrenUpdate()
        {
            foreach (var obj in watchedObjects.Keys)
                if (obj is DataItem) (obj as DataItem).EndUpdate();
        }

        /// <summary>
        /// Definuje, e vlastnost danho nzvu m bt sledovna a zmny objektu ve vlastnosti zpsob vyvoln udlosti Changed.
        /// Sledovan objekt by ml implementovat DataItem nebo IAddRemoveEvents nebo PropertyChanged. V opanm ppad nebude sledovn probhat
        /// </summary>
        protected void DefineWatchedProperty(string name, object defaultValue)
        {
            watchedProperties.Add(name, defaultValue);
            StartWatchingObject(GetPropertyInfo(name), defaultValue);
        }

        /// <summary>
        /// Vrt hodnotu sledovan vlastnosti
        /// </summary>
        protected T GetWatchedProperty<T>(string name)
        {
            return (T)watchedProperties[name];
        }

        private string updatingProperty = null;

        /// <summary>
        ///  Nastav hodnotu sledovan vlastnosti
        /// </summary>
        protected void SetWatchedProperty(string property, object value)
        {
            //notifikovat zmnu property
            updatingProperty = property;

            var oldValue = watchedProperties[property];
            var propInfo = GetPropertyInfo(property);
            OnChanged(new ChangedEventArgs(this, propInfo, oldValue, value));
            WatchedPropertyChanged(propInfo, oldValue, value);
            OnPropertyChanged(property);

            //zmnit
            StopWatchingObject(oldValue);
            watchedProperties[property] = value;
            StartWatchingObject(propInfo, value);

            updatingProperty = null;
        }

        protected void StartWatchingObject(string property, object watchedObject)
        {
            StartWatchingObject(GetPropertyInfo(property), watchedObject);
        }

        /// <summary>
        /// Spust sledovn pednho objektu. Po skonen sledovn mus bt zavolna metoda StopWatchingObject, aby mohl bt sledovan objekt uvolnn z pamti.
        /// </summary>
        private void StartWatchingObject(PropertyInfo property, object watchedObject)
        {
            if ((object)(watchedObject) == null) return;
            if ((watchedObject is DataItem || watchedObject is INotifyPropertyChanged || typeof(IAddRemoveEvents).IsAssignableFrom(watchedObject.GetType())))
            {
                if (watchedObjects.Any(o => object.ReferenceEquals(o.Key, watchedObject) && o.Value.Property.Name != property.Name))
                    throw new InvalidOperationException("Jeden objekt neme bt sledovn vcekrt v rmci rznch nzv vlastnosti");
            }

            AddWatchedObject(watchedObject, property);

            if (watchedObject is DataItem)
            {
                var item = (watchedObject as DataItem);
                item.Changed += new EventHandler<ChangedEventArgs>(WatchedObject_Changed);
                item.Parent = this;
                for (int i = 0; i < updatingCount; i++) item.BeginUpdate();
            }
            else if (typeof(IAddRemoveEvents).IsAssignableFrom(watchedObject.GetType()))
            {
                var target = watchedObject as IAddRemoveEvents;
                foreach (object item in target)
                    StartWatchingObject(property, item);
                target.ItemsAddedUntyped += new EventHandler<ItemsEventArgs>(DataItem_ItemsAddedUntyped);
                target.ItemsRemovedUntyped += new EventHandler<ItemsEventArgs>(DataItem_ItemsRemovedUntyped);
            }
            else if (watchedObject is INotifyPropertyChanged)
                (watchedObject as INotifyPropertyChanged).PropertyChanged += new PropertyChangedEventHandler(WatchedObject_PropertyChanged);
        }

        protected void StopWatchingObject(object watchedObject)
        {
            if ((object)(watchedObject) == null) return;
            if (watchedObject is DataItem)
            {
                var item = (watchedObject as DataItem);
                item.Changed -= new EventHandler<ChangedEventArgs>(WatchedObject_Changed);
                for (int i = 0; i < updatingCount; i++) item.EndUpdate();
                // item.Parent = null;
            }
            else if (typeof(IAddRemoveEvents).IsAssignableFrom(watchedObject.GetType()))
            {
                var target = watchedObject as IAddRemoveEvents;
                var property = GetPropertyInfo(watchedObject);
                foreach (object item in target)
                    StartWatchingObject(property, item);
                target.ItemsAddedUntyped -= new EventHandler<ItemsEventArgs>(DataItem_ItemsAddedUntyped);
                target.ItemsRemovedUntyped -= new EventHandler<ItemsEventArgs>(DataItem_ItemsRemovedUntyped);
            }
            else if (watchedObject is INotifyPropertyChanged)
            {
                (watchedObject as INotifyPropertyChanged).PropertyChanged -= new PropertyChangedEventHandler(WatchedObject_PropertyChanged);
            }
            RemoveWatchedObject(watchedObject);
        }

        /// <summary>
        /// Pokud sledovan objekt je potomkem DataItem, pi jeho zmn je vykonna tato funkce
        /// </summary>
        private void WatchedObject_Changed(object sender, ChangedEventArgs e)
        {
            e.AddNodePropertyInfo(GetPropertyInfo(sender));
            e.AddNodeObject(this);
            OnChanged(e);
        }

        /// <summary>
        /// Pokud sledovan objekt je potomkem IAddRemoveEvents, pi jeho zmn je vykonna tato funkce
        /// </summary>
        private void DataItem_ItemsAddedUntyped(object sender, ItemsEventArgs e)
        {
            var property = GetPropertyInfo(sender);
            foreach (var item in e.Items.Cast<object>().ToList())
            {
                StartWatchingObject(property, item);
                var args = new ChangedEventArgs(sender, ChangeType.ChildAdded, item);
                WatchedObject_Changed(sender, args);
            }
        }

        /// <summary>
        /// Pokud sledovan objekt je potomkem IAddREmoveEvents, pi jeho zmn je vykonna tato funkce
        /// </summary>
        private void DataItem_ItemsRemovedUntyped(object sender, ItemsEventArgs e)
        {
            foreach (var item in e.Items.Cast<object >().ToList())
            {
                var args = new ChangedEventArgs(sender, ChangeType.ChildRemoved, item);
                args.AddNodeObject(this);
                args.AddNodePropertyInfo(GetPropertyInfo(sender));
                OnChanged(args);

                StopWatchingObject(item);
            }
        }

        /// <summary>
        /// Pokud sledovan objekt implementuje INotifyPropertyChanged, jeho zmny jsou zpracovny touto funkc
        /// </summary>
        private void WatchedObject_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            var property = GetPropertyInfo(sender);
            var args = new ChangedEventArgs(sender, ProvidePropertyInfo(e.PropertyName, property));
            args.AddNodeObject(this);
            args.AddNodePropertyInfo(property);
            OnChanged(args);
        }

        /// <summary>
        /// Poskytne informace k vlastnosti sledovanho objektu (z kolekce subProperties)
        /// </summary>
        protected virtual PropertyInfo ProvidePropertyInfo(string name, PropertyInfo baseProperty)
        {
            string[] classes;
            subProperties.TryGetValue(new Tuple<string, string>(baseProperty.Name, name), out classes);
            return new PropertyInfo(name, classes);
        }

        /// <summary>
        /// Metoda je zavolna pi zmn hodnoty sledovan vlastnosti.
        /// Metoda nen zavolna pi zmn stavu objektu vlastnosti a jeho dalch sledovanch vlastnostech!
        /// </summary>
        /// <param name="name"></param>
        protected virtual void WatchedPropertyChanged(PropertyInfo name, object oldValue, object newValue)
        { }

        /// <summary>
        /// Nalezne PropertyInfo podle jmna. Pokud neexistuje, bude vytvoeno
        /// </summary>
        protected PropertyInfo GetPropertyInfo(string name)
        {
            var property = properties.Where(p => p.Name == name).FirstOrDefault();
            if (property == null)
            {
                property = new PropertyInfo(name, null);
                properties.Add(property);
            }
            return property;
        }

        /// <summary>
        /// Nalezne PropertyInfo poble objektu, kter je hodnotou tto vlastnosti. Mus se jednat o sledovanou vlastnost.
        /// </summary>
        private PropertyInfo GetPropertyInfo(object sender)
        {
            return GetWatchedObjectInfo(sender).Property;
            //return watchedObjects.Where(w => object.ReferenceEquals(w.Object, sender)).First().Property;
        }

        private void AddWatchedObject(object obj, PropertyInfo property)
        {
            WatchedObjectInfo info;
            if (watchedObjects.TryGetValue(obj, out info))
            {
                info.Count++;
            }
            else
            {
                watchedObjects.Add(obj, new WatchedObjectInfo(property));
            }
        }

        private WatchedObjectInfo GetWatchedObjectInfo(object obj)
        {
            return watchedObjects[obj];
        }

        private void RemoveWatchedObject(object obj)
        {
            WatchedObjectInfo info;
            if (watchedObjects.TryGetValue(obj, out info))
            {
                if (info.Count <= 1) watchedObjects.Remove(obj);
                else info.Count--;
            }
            else
            {
                throw new InvalidOperationException("Kl nebyl nalezen");
            }
        }

        public event EventHandler<ChangedEventArgs> Changed;

        protected void OnChanged(ChangedEventArgs e)
        {
            if (updatingCount > 0) return;
            if (Changed != null) Changed(this, e);
        }

        public event PropertyChangedEventHandler PropertyChanged;

        protected void OnPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            ProcessPropertyChanged(propertyName);
        }

        protected virtual void ProcessPropertyChanged(string propertyName)
        {
            if (updatingProperty == propertyName) return;
            OnChanged(new ChangedEventArgs(this, GetPropertyInfo(propertyName)));
        }
    }
}