using System;
using System.ComponentModel;
using Brjnk.Common.ValueTypes;

namespace Brjnk.Common
{
    [TypeConverter(typeof(StringValueConverter<Distance>))]
    public struct Distance : IFormattable, IPhysicalQuantity
    {
        //1 Foot = 0.3048 Metr
        private const double Foot = 0.3048;

        //1 Yard = 0.9144 Metr
        private const double Yard = 0.9144;

        //1 mile = 1.6 Kilometer
        private const double Mile = 1609.344;

        //distance in meters
        private double _distance;

        public Distance(GpsCoordinates Coordinate1, GpsCoordinates Coordinate2)
        {
            _distance = GetDistance(Coordinate1, Coordinate2);
        }

        public Distance(double metres)
        {
            _distance = metres;
        }

        /// <summary>
        /// Return distance between two points in metres
        /// </summary>
        /// <param name="Coordinate1">First point</param>
        /// <param name="Coordinate2">Second point</param>
        /// <returns>Distance in Metres</returns>
        /// <remarks></remarks>
        public static double GetDistance(GpsCoordinates Coordinate1, GpsCoordinates Coordinate2)
        {
            if (Coordinate1 == Coordinate2) return 0;
            double lat1 = Coordinate1.LatTotalDegree;
            double lat2 = Coordinate2.LatTotalDegree;
            double lon1 = Coordinate1.LonTotalDegree;
            double lon2 = Coordinate2.LonTotalDegree;

            lat1 = lat1 * (Math.PI / 180);
            lat2 = lat2 * (Math.PI / 180);
            lon1 = lon1 * (Math.PI / 180);
            lon2 = lon2 * (Math.PI / 180);

            const int EarthRadius = 6378137;
            //meters
            return EarthRadius * Math.Acos(Math.Sin(lat2) * Math.Sin(lat1) + Math.Cos(lat2) * Math.Cos(lat1) * Math.Cos(lon2 - lon1));
        }

        /// <summary>
        /// Implicitn konverze vrac vzdlenost v metrech
        /// </summary>
        public static implicit operator double(Distance distance)
        {
            return distance._distance;
        }

        public static implicit operator Distance(double a)
        {
            return new Distance(a);
        }

        /// <summary>
        /// Return distance with optimal unit as string.
        /// </summary>
        public override string ToString()
        {
            var unit = UnitsSettings.Instance.Distance;
            return ToString(unit, 3);
        }

        /// <summary>
        /// Return distance with optimal unit as string.
        /// </summary>
        public string ToString(MeasurmentSystem MeasurmentSystem = MeasurmentSystem.Metric)
        {
            if (MeasurmentSystem == MeasurmentSystem.Metric)
            {
                if (_distance < 100)
                    //metry, 2 desetinn msta
                    return ToString(DistanceUnit.Meter, 2);
                else if (_distance < 1000)
                    //metry, 1 desetinn msto
                    return ToString(DistanceUnit.Meter, 1);
                else if (_distance < 10000)
                    //km, 2 desetinn msta
                    return ToString(DistanceUnit.Kilometre, 2);
                else if (_distance < 100000)
                    //km, 1 desetinn msto
                    return ToString(DistanceUnit.Kilometre, 1);
                else
                    //km
                    return ToString(DistanceUnit.Kilometre, 0);
            }
            else
            {
                if (_distance < Mile)
                    return ToString(DistanceUnit.Yard, 1);
                else if (_distance < 10 * Mile)
                    return ToString(DistanceUnit.Mile, 2);
                else if (_distance < 100 * Mile)
                    return ToString(DistanceUnit.Mile, 1);
                else
                    return ToString(DistanceUnit.Mile, 0);
            }
        }

        /// <summary>
        /// Return distance in specificed unit as string.
        /// </summary>
        /// <param name="Unit"></param>
        /// <param name="DecimalPlaceCount">Count of decimal places.</param>
        /// <returns></returns>
        /// <remarks></remarks>
        public string ToString(DistanceUnit Unit, int DecimalPlaceCount = 0)
        {
            switch (Unit)
            {
                case DistanceUnit.Meter:
                    return string.Format("{0:N" + DecimalPlaceCount + "} m", GetDistanceValue(DistanceUnit.Meter));
                case DistanceUnit.Kilometre:
                    return string.Format("{0:N" + DecimalPlaceCount + "} km", GetDistanceValue(DistanceUnit.Kilometre));
                case DistanceUnit.Yard:
                    return string.Format("{0:N" + DecimalPlaceCount + "} yd", GetDistanceValue(DistanceUnit.Yard));
                case DistanceUnit.Mile:
                    return string.Format("{0:N" + DecimalPlaceCount + "} mil", GetDistanceValue(DistanceUnit.Mile));
                default:
                    throw new ArgumentException("Unsupported distance unit", "Unit");
            }
        }

        /// <summary>
        /// Return distance in specifiked unit
        /// </summary>
        /// <param name="Unit"></param>
        /// <returns></returns>
        /// <remarks></remarks>
        public double GetDistanceValue(DistanceUnit Unit)
        {
            switch (Unit)
            {
                case DistanceUnit.Meter:
                    return _distance;
                case DistanceUnit.Kilometre:
                    return _distance / 1000;
                case DistanceUnit.Yard:
                    return _distance / Yard;
                case DistanceUnit.Mile:
                    return _distance / Mile;
                default:
                    throw new ArgumentException("Invalid value of argument", "Unit");
            }
        }

        public string ToString(string format, IFormatProvider formatProvider)
        {
            switch (format.ToLower())
            {
                case "km":
                    return ToString(DistanceUnit.Kilometre, 3);
                case "mil":
                    return ToString(DistanceUnit.Mile, 3);
                case "m":
                    return ToString(DistanceUnit.Meter, 2);
                case "yd":
                    return ToString(DistanceUnit.Yard, 2);
                default:
                    return ToString();
            }
        }

        public static Distance operator +(Distance d1, Distance d2)
        {
            return new Distance((double)d1 + (double)d2);
        }

        public static Distance operator +(Distance d1, double d2)
        {
            return new Distance(d1.GetDistanceValue(DistanceUnit.Meter) + d2);
        }

        static Distance()
        {
            Parser = new ValueParser();
            Parser.AddUnit("m", 1);
            Parser.AddUnit("km", 1000);
            Parser.AddUnit("mil", Mile);
            Parser.AddUnit("yd", Yard);
        }

        public static readonly ValueParser Parser;

        public double GetValue()
        {
            return _distance;
        }
    }

    public enum MeasurmentSystem
    {
        Metric,
        USMeasurmentSystem
    }

    public enum DistanceUnit
    {
        Meter,
        Kilometre,

        Yard,
        Mile
    }
}