namespace Brjnk.CoordinateSystem2D
{
    using System;
    using System.Windows;
    using System.Windows.Controls;
    using System.Windows.Media;

    /// <summary>
    /// UI prvek vykreslujc osu grafu i s popisky. Popisky osy jsou zskvny z instance tdy AxisDefinition.
    /// </summary>
    public class Axis : FrameworkElement
    {
        private readonly static double verticalTextAngle = 270;

        protected override void OnRender(System.Windows.Media.DrawingContext drawingContext)
        {
            drawingContext.DrawRectangle(Background, null, new Rect(0, 0, this.ActualWidth, this.ActualHeight));

            var axis = Definition;
            double scalePerDip = (To - From) / GetWidth();
            var labels = axis.GetAxisLabels(From, To, scalePerDip * (MinLabelWidth + LabelsMargin * 2));

            double lastLabelFrom = 0;
            double lastLabelTo = 0;
            double lastLabelValue = 0;
            bool lastLabelLarge = false;
            string lastLabelString = null;
            bool first = true;

            foreach (var label in labels.Labels)
            {
                var value = ToRelative(label.Value);
                //LabelSize.Big se od LabelSize.Large li pouze tunost psma
                if (label.Size == LabelSize.Big || label.Size == LabelSize.Large)
                {
                    DrawValue(value, label.Size, drawingContext);
                    lastLabelTo = lastLabelValue + (value - lastLabelValue) / 2;
                    if (first) lastLabelTo = 0;
                    ProcessPreviousLable(lastLabelFrom, lastLabelTo, lastLabelValue, lastLabelString, lastLabelLarge, drawingContext);
                    lastLabelString = label.Text;
                    lastLabelValue = value;
                    lastLabelFrom = lastLabelTo;
                    lastLabelLarge = label.Size == LabelSize.Large;
                    first = false;
                }
                else
                {
                    DrawValue(value, label.Size, drawingContext);
                }
            }

            ProcessPreviousLable(lastLabelFrom, GetWidth(), lastLabelValue, lastLabelString, lastLabelLarge, drawingContext);
            DrawCaption(labels.Caption, drawingContext);
            OnLabelsGenerated(labels);
        }

        private void DrawCaption(string caption, DrawingContext drawingContext)
        {
            var typeface = new Typeface(TextBlock.GetFontFamily(this), FontStyles.Normal, FontWeights.Normal, FontStretches.Normal);
            var fText = new FormattedText(caption, System.Globalization.CultureInfo.CurrentCulture,
                TextBlock.GetFlowDirection(this),
                typeface,
                TextBlock.GetFontSize(this), CaptionBrush);

            double scaleX = (GetWidth() / 4) / fText.Width;
            double scaleY = (GetHeight() - RulerHeight) / fText.Height;
            double scale = Math.Min(scaleX, scaleY);
            scale = Math.Min(scale, 1);
            if (scale <= 0) return;

            if (Orientation == AxisOrientation.Horizontal)
            {
                drawingContext.PushTransform(new ScaleTransform(scale, scale, GetWidth(), GetHeight()));
                double x = GetWidth() - fText.Width; double y = GetHeight() - fText.Height;
                drawingContext.DrawRectangle(CaptionBackground, null, new Rect(x, y, fText.Width, fText.Height));
                drawingContext.DrawText(fText, new Point(x, y));
                drawingContext.Pop();
            }
            else
            {
                double centerX = (fText.Height / 2) * scale;
                double centerY = (fText.Width / 2) * scale;
                drawingContext.PushTransform(new ScaleTransform(scale, scale, centerX, centerY));
                drawingContext.PushTransform(new RotateTransform(verticalTextAngle, centerX, centerY));
                double x = centerX - fText.Width / 2; double y = centerY - fText.Height / 2;
                drawingContext.DrawRectangle(CaptionBackground, null, new Rect(x, y, fText.Width, fText.Height));
                drawingContext.DrawText(fText, new Point(x, y));
                drawingContext.Pop();
                drawingContext.Pop();
            }
        }

        private void ProcessPreviousLable(double from, double to, double value, string text, bool bold, DrawingContext context)
        {
            if (!string.IsNullOrEmpty(text))
            {
                DrawLabel(value, 2 * Math.Min(value - from - LabelsMargin, to - value - LabelsMargin), text, bold, context);
            }
        }

        /// <summary>
        /// Pevede skutenou hodnotu na relativn vzdlenost od potku viditeln osy (v dip)
        /// </summary>
        private double ToRelative(double value)
        {
            double scalePerDip = (To - From) / GetWidth();
            return (value - From) / scalePerDip;
        }

        private void RenderRequired(object sender, EventArgs args)
        {
            this.InvalidateVisual();
        }

        private double GetWidth()
        {
            if (Orientation == AxisOrientation.Horizontal) return this.ActualWidth;
            return this.ActualHeight;
        }

        private double GetHeight()
        {
            if (Orientation == AxisOrientation.Horizontal) return this.ActualHeight;
            return this.ActualWidth;
        }

        /// <summary>
        /// Vykresl rku na ose
        /// </summary>
        private void DrawValue(double relativeValue, LabelSize size, DrawingContext context)
        {
            Pen currentPen;
            double height;
            switch (size)
            {
                case LabelSize.Small:
                    pen4.Thickness = this.RulerThickness;
                    pen4.Brush = this.RulerBrush4;
                    currentPen = pen4;
                    height = this.RulerHeight / 2;
                    break;

                case LabelSize.Medium:
                    pen3.Thickness = this.RulerThickness;
                    pen3.Brush = this.RulerBrush3;
                    currentPen = pen3;
                    height = 3 * this.RulerHeight / 4;
                    break;

                case LabelSize.Big:
                    pen2.Thickness = this.RulerThickness;
                    pen2.Brush = this.RulerBrush2;
                    currentPen = pen2;
                    height = this.RulerHeight;
                    break;

                case LabelSize.Large:
                    pen1.Thickness = this.RulerThickness;
                    pen1.Brush = this.RulerBrush1;
                    currentPen = pen1;
                    height = this.RulerHeight;
                    break;

                default:
                    throw new ArgumentException("Size");
            }

            if (Orientation == AxisOrientation.Horizontal)
                context.DrawLine(currentPen, new Point(relativeValue, 0), new Point(relativeValue, height));
            else
                context.DrawLine(currentPen,
                    new Point(GetHeight() - height, GetWidth() - relativeValue),
                    new Point(GetHeight(), GetWidth() - relativeValue));
        }

        private Pen pen1 = new Pen();
        private Pen pen2 = new Pen();
        private Pen pen3 = new Pen();
        private Pen pen4 = new Pen();

        private void DrawLabel(double relativeValue, double textWidth, string text, bool bold, DrawingContext context)
        {
            var typeface = new Typeface(TextBlock.GetFontFamily(this), FontStyles.Normal, bold ? FontWeights.Bold : FontWeights.Normal, FontStretches.Normal);
            FormattedText fText = new FormattedText(text, System.Globalization.CultureInfo.CurrentCulture, TextBlock.GetFlowDirection(this), typeface, TextBlock.GetFontSize(this), this.Foreground);

            double scaleX = textWidth / fText.Width;
            double scaleY = (this.GetHeight() - RulerHeight) / fText.Height;
            double scale = Math.Min(scaleX, scaleY);
            scale = Math.Min(MaxFontZoom, scale);
            if (scale < MinFontZoom) return;

            if (Orientation == AxisOrientation.Horizontal)
            {
                DrawTextHorizontal(fText, new Point(relativeValue, RulerHeight + (GetHeight() - RulerHeight) / 2), scale, context);
            }
            else
            {
                DrawTextVertical(fText, new Point((GetHeight() - RulerHeight) / 2, GetWidth() - relativeValue), scale, context);
            }
        }

        private void DrawTextVertical(FormattedText text, Point point, double scale, DrawingContext context)
        {
            context.PushTransform(new ScaleTransform(scale, scale, point.X, point.Y));
            context.PushTransform(new RotateTransform(verticalTextAngle, point.X, point.Y));
            context.DrawText(text, new Point(point.X - text.Width / 2, point.Y - text.Height / 2));
            context.Pop();
            context.Pop();
        }

        private void DrawTextHorizontal(FormattedText text, Point point, double scale, DrawingContext context)
        {
            context.PushTransform(new ScaleTransform(scale, scale, point.X, point.Y));
            context.DrawText(text, new Point(point.X - text.Width / 2, point.Y - text.Height / 2));
            context.Pop();
        }

        # region Properties

        public AxisDefinition Definition
        {
            get { return (AxisDefinition)GetValue(DefinitionProperty); }
            set { SetValue(DefinitionProperty, value); }
        }

        public static readonly DependencyProperty DefinitionProperty =
            DependencyProperty.Register("Definition", typeof(AxisDefinition), typeof(Axis), new PropertyMetadata(null, DefinitionChanged));

        private static void DefinitionChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            var axis = d as Axis;
            var oldValue = (e.OldValue as AxisDefinition);
            if ((object)oldValue != null) oldValue.RedrawRequired -= axis.RenderRequired;
            var newValue = (e.NewValue as AxisDefinition);
            newValue.RedrawRequired += axis.RenderRequired;
            axis.InvalidateVisual(); // pekresl podle nov definice
        }

        public double From
        {
            get { return (double)GetValue(FromProperty); }
            set { SetValue(FromProperty, value); }
        }

        public static readonly DependencyProperty FromProperty =
            DependencyProperty.Register("From", typeof(double), typeof(Axis), new FrameworkPropertyMetadata(0.0, FrameworkPropertyMetadataOptions.AffectsRender));

        public double To
        {
            get { return (double)GetValue(ToProperty); }
            set { SetValue(ToProperty, value); }
        }

        public static readonly DependencyProperty ToProperty =
            DependencyProperty.Register("To", typeof(double), typeof(Axis), new FrameworkPropertyMetadata(100.0, FrameworkPropertyMetadataOptions.AffectsRender));

        /// <summary>
        /// Vka pravtka, zbytek vertiklnho prostoru bude vyplnn popiskami hodnot osy
        /// </summary>
        public double RulerHeight
        {
            get { return (double)GetValue(RulerHeightProperty); }
            set { SetValue(RulerHeightProperty, value); }
        }

        public static readonly DependencyProperty RulerHeightProperty =
            DependencyProperty.Register("RulerHeight", typeof(double), typeof(Axis), new FrameworkPropertyMetadata(15.0, FrameworkPropertyMetadataOptions.AffectsRender));

        /// <summary>
        /// ka rky pravtka
        /// </summary>
        public double RulerThickness
        {
            get { return (double)GetValue(RulerThicknessProperty); }
            set { SetValue(RulerThicknessProperty, value); }
        }

        public static readonly DependencyProperty RulerThicknessProperty =
            DependencyProperty.Register("RulerThickness", typeof(double), typeof(Axis), new FrameworkPropertyMetadata(2.0, FrameworkPropertyMetadataOptions.AffectsRender));

        /// <summary>
        /// ttec pravtka prvn rovn (nap 0)
        /// </summary>
        public Brush RulerBrush1
        {
            get { return (Brush)GetValue(RulerBrush1Property); }
            set { SetValue(RulerBrush1Property, value); }
        }

        public static readonly DependencyProperty RulerBrush1Property =
            DependencyProperty.Register("RulerBrush1", typeof(Brush), typeof(Axis), new FrameworkPropertyMetadata(Brushes.Black, FrameworkPropertyMetadataOptions.AffectsRender));

        /// <summary>
        /// ttec pravtka druh rovn (nap 1,2,3...)
        /// </summary>
        public Brush RulerBrush2
        {
            get { return (Brush)GetValue(RulerBrush2Property); }
            set { SetValue(RulerBrush2Property, value); }
        }

        public static readonly DependencyProperty RulerBrush2Property =
            DependencyProperty.Register("RulerBrush2", typeof(Brush), typeof(Axis), new FrameworkPropertyMetadata(Brushes.Black, FrameworkPropertyMetadataOptions.AffectsRender));

        /// <summary>
        /// ttec pravtka tet rovn (nap 0.5, 1.5 ...)
        /// </summary>
        public Brush RulerBrush3
        {
            get { return (Brush)GetValue(RulerBrush3Property); }
            set { SetValue(RulerBrush3Property, value); }
        }

        public static readonly DependencyProperty RulerBrush3Property =
            DependencyProperty.Register("RulerBrush3", typeof(Brush), typeof(Axis), new FrameworkPropertyMetadata(Brushes.DarkGray, FrameworkPropertyMetadataOptions.AffectsRender));

        /// <summary>
        /// ttec pravtka tvrt rovn (nap 0.1, 0.2, 0.3)
        /// </summary>
        public Brush RulerBrush4
        {
            get { return (Brush)GetValue(RulerBrush4Property); }
            set { SetValue(RulerBrush4Property, value); }
        }

        public static readonly DependencyProperty RulerBrush4Property =
            DependencyProperty.Register("RulerBrush4", typeof(Brush), typeof(Axis), new FrameworkPropertyMetadata(Brushes.Gray, FrameworkPropertyMetadataOptions.AffectsRender));

        /// <summary>
        /// Minimln ppust ka popisku hodnoty osy. (Hodnota osy prvn rovn)
        /// </summary>
        public double MinLabelWidth
        {
            get { return (double)GetValue(MinLabelWidthProperty); }
            set { SetValue(MinLabelWidthProperty, value); }
        }

        public static readonly DependencyProperty MinLabelWidthProperty =
             DependencyProperty.Register("MinLabelWidth", typeof(double), typeof(Axis), new FrameworkPropertyMetadata(37.0, FrameworkPropertyMetadataOptions.AffectsRender));

        /// <summary>
        /// Minimln mon piblen velikosti psma.
        /// Pokud nebude dostatek prostoru k vykreslen, popisek nebude vykreslen vbec.
        /// Velikost psma je urena z TextBlock.FontSize
        /// </summary>
        public double MinFontZoom
        {
            get { return (double)GetValue(MinFontZoomProperty); }
            set { SetValue(MinFontZoomProperty, value); }
        }

        public static readonly DependencyProperty MinFontZoomProperty =
            DependencyProperty.Register("MinFontZoom", typeof(double), typeof(Axis), new FrameworkPropertyMetadata(0.1, FrameworkPropertyMetadataOptions.AffectsRender));

        /// <summary>
        /// Maximln mon piblen velikosti psma. Velikost psma je urena z TextBlock.FontSize
        /// </summary>
        public double MaxFontZoom
        {
            get { return (double)GetValue(MaxFontZoomProperty); }
            set { SetValue(MaxFontZoomProperty, value); }
        }

        public static readonly DependencyProperty MaxFontZoomProperty =
            DependencyProperty.Register("MaxFontZoom", typeof(double), typeof(Axis), new FrameworkPropertyMetadata(double.MaxValue, FrameworkPropertyMetadataOptions.AffectsRender));

        /// <summary>
        /// Okraj, kter bude pidn na pravou i levou stranu kadho popisku hodnoty seln osy
        /// </summary>
        public double LabelsMargin
        {
            get { return (double)GetValue(LabelsMarginProperty); }
            set { SetValue(LabelsMarginProperty, value); }
        }

        public static readonly DependencyProperty LabelsMarginProperty =
            DependencyProperty.Register("LabelsMargin", typeof(double), typeof(Axis), new FrameworkPropertyMetadata(0.0, FrameworkPropertyMetadataOptions.AffectsRender));

        public AxisOrientation Orientation
        {
            get { return (AxisOrientation)GetValue(OrientationProperty); }
            set { SetValue(OrientationProperty, value); }
        }

        public static readonly DependencyProperty OrientationProperty =
            DependencyProperty.Register("Orientation", typeof(AxisOrientation), typeof(Axis), new FrameworkPropertyMetadata(AxisOrientation.Horizontal, FrameworkPropertyMetadataOptions.AffectsRender));

        public static readonly DependencyProperty ForegroundProperty =
            TextBlock.ForegroundProperty.AddOwner(typeof(Axis), new FrameworkPropertyMetadata() { AffectsRender = true });

        /// <summary>
        /// ttec popisk hodnot osy (TextBlock.ForegrounProperty)
        /// </summary>
        public Brush Foreground
        {
            get { return (Brush)GetValue(ForegroundProperty); }
            set { SetValue(ForegroundProperty, value); }
        }

        /// <summary>
        /// ttec popisk osy (nikoliv popisk jednotlivch hodnot osy)
        /// </summary>
        public Brush CaptionBrush
        {
            get { return (Brush)GetValue(CaptionBrushProperty); }
            set { SetValue(CaptionBrushProperty, value); }
        }

        public static readonly DependencyProperty CaptionBrushProperty =
            DependencyProperty.Register("CaptionBrush", typeof(Brush), typeof(Axis), new FrameworkPropertyMetadata(Brushes.Black, FrameworkPropertyMetadataOptions.AffectsRender));

        public Brush CaptionBackground
        {
            get { return (Brush)GetValue(CaptionBackgroundProperty); }
            set { SetValue(CaptionBackgroundProperty, value); }
        }

        public static readonly DependencyProperty CaptionBackgroundProperty =
            DependencyProperty.Register("CaptionBackground", typeof(Brush), typeof(Axis), new FrameworkPropertyMetadata(new SolidColorBrush(Color.FromArgb(200, 255, 255, 255)), FrameworkPropertyMetadataOptions.AffectsRender));

        public Brush Background
        {
            get { return (Brush)GetValue(BackgroundProperty); }
            set { SetValue(BackgroundProperty, value); }
        }

        public static readonly DependencyProperty BackgroundProperty =
            Control.BackgroundProperty.AddOwner(typeof(Axis), new FrameworkPropertyMetadata() { AffectsRender = true });

        #endregion

        #region events

        private void OnLabelsGenerated(GetAxisLabelsResult labels)
        {
            RaiseEvent(new LabelsGeneratedEventArgs(this.Orientation, labels) { RoutedEvent = LabelsGeneratedEvent });
        }

        public event LabelsGeneratedHandler LabelsGenerated
        {
            add { AddHandler(LabelsGeneratedEvent, value); }
            remove { RemoveHandler(LabelsGeneratedEvent, value); }
        }

        public static readonly RoutedEvent LabelsGeneratedEvent = EventManager.RegisterRoutedEvent(
            "LabelsGenerated", RoutingStrategy.Bubble, typeof(LabelsGeneratedHandler), typeof(Axis));

        #endregion
    }

    public enum AxisOrientation
    {
        /// <summary>
        /// Osa x
        /// </summary>
        Horizontal = 0,

        /// <summary>
        /// Osa y
        /// </summary>
        Vertical = 1
    }

    public delegate void LabelsGeneratedHandler(object sender, LabelsGeneratedEventArgs e);

    public class LabelsGeneratedEventArgs : RoutedEventArgs
    {
        public AxisOrientation Orientation { get; protected set; }

        public GetAxisLabelsResult Labels { get; protected set; }

        public LabelsGeneratedEventArgs(AxisOrientation orientation, GetAxisLabelsResult labels)
        {
            this.Orientation = orientation;
            this.Labels = labels;
        }
    }
}