namespace Brjnk.CoordinateSystem2D
{
    using System;
    using System.Collections.Generic;
    using System.Windows;

    /// <summary>
    /// </summary>
    public abstract class AxisDefinition : System.Windows.Freezable
    {
        public string Caption
        {
            get { return ((string)GetValue(CaptionProperty)); }
            set { SetValue(CaptionProperty, value); }
        }

        public static readonly DependencyProperty CaptionProperty =
            DependencyProperty.Register("Caption", typeof(string), typeof(AxisDefinition), new PropertyMetadata(string.Empty, AxisChanged));

        protected static readonly DependencyPropertyKey IsLinearPropertyKey =
            DependencyProperty.RegisterReadOnly("IsLinear", typeof(bool), typeof(AxisDefinition), new PropertyMetadata(true));

        public static readonly DependencyProperty IsLinearProperty =
            IsLinearPropertyKey.DependencyProperty;

        public bool IsLinear
        {
            get { return (bool)GetValue(IsLinearProperty); }
            protected set { SetValue(IsLinearPropertyKey, value); }
        }

        protected static void AxisChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            var axis = d as AxisDefinition;
            axis.OnRedrawRequired(EventArgs.Empty);
        }

        /// <summary>
        /// Vygeneruje popisky osy.
        /// StepLarge, StepBig, StepSmall, StepMedium uruje rozestup mezi jednotlivmi velikostmi popisk.
        /// Je-li StepLarge, StepBig nebo StepSmall 0, popisky pslun velikosti nebudou vytvoeny
        /// StepLarge mus bt celoseln nsobek StepBig atd.
        /// </summary>
        /// <param name="labelFormatter">Funkce, kter vrt textov popisek bodu na ose</param>
        /// <returns></returns>
        protected IEnumerable<LabelInfo> GenerateLabels(double from, double to, CalculateStepsResult r)
        {
            if (r.stepSmall == 0 || from == to) yield break;
            if (r.stepLarge < 0 || r.stepBig < 0 || r.stepMedium < 0 || r.stepSmall < 0) throw new ArgumentOutOfRangeException("Step neme bt zporn");
            if (to < from) throw new ArgumentException("From neme bt vt ne To.");
            if (!IsMultiple(r.stepLarge, r.stepBig) && r.stepBig != 0) throw new ArgumentException("StepLarge mus bt nsobkem StepBig.");
            if (!IsMultiple(r.stepBig, r.stepMedium) && r.stepMedium != 0) throw new ArgumentException("StepBig mus bt nsobkem StepMedium.");
            if (!IsMultiple(r.stepMedium, r.stepSmall) && r.stepSmall != 0) throw new ArgumentException("StepMedium mus bt nsobkem StepSmall.");

            //  int start;
            //  double highestStep = Math.Max(Math.Max(r.stepLarge, r.stepBig), r.stepSmall);
            // start = (int)Math.Floor(from / highestStep) - 1;

            int step3Multiple = (int)Math.Round(r.stepMedium / r.stepSmall);
            int step2Multiple = (int)Math.Round(r.stepBig / r.stepSmall);
            int step1Multiple = (int)Math.Round(r.stepLarge / r.stepSmall);

            //Funkce pro vyhodnocen textovho popisku. Nen-li definovn, uije se vchoz.
            Func<double, string> largeFormatter = r.formatterLarge == null ? defaultFormatter : r.formatterLarge;
            Func<double, string> bigFormatter = r.formatterBig == null ? defaultFormatter : r.formatterBig;

            //int i = (start - 1) * (int)Math.Round(highestStep / r.stepSmall);
            long i = (long)Math.Floor(from / r.stepSmall) - 1;
            while (true)
            {
                i++;
                double value = i * r.stepSmall;

                //Vyfiltrovat nedouc popisky (popisky mimo danou oblast)
                if (value < from) continue;
                if (value > to) break;

                //Stanovit velikost popisku
                LabelSize size = LabelSize.Small;
                if (r.stepLarge != 0 && i % step1Multiple == 0) size = LabelSize.Large;
                else if (i == 0 && r.LargeZero) size = LabelSize.Large;
                else if (r.largeLabelSelector != null && r.largeLabelSelector(value)) size = LabelSize.Large;
                else if (r.stepBig != 0 && i % step2Multiple == 0) size = LabelSize.Big;
                else if (r.stepMedium != 0 && i % step3Multiple == 0) size = LabelSize.Medium;
                else if (r.AutoMediumStep && IsAutoMedium(i, step2Multiple)) size = LabelSize.Medium;

                // Zjistit textov popisek
                string text = null;
                if (size == LabelSize.Large || size == LabelSize.Big)
                {
                    double labelValue = r.Reverse ? -value : value;
                    text = size == LabelSize.Large ? largeFormatter(labelValue) : bigFormatter(labelValue);
                };

                yield return new LabelInfo(size, value, text);
            }
        }

        /// <summary>
        /// Vrt true, pokud popisek s danm indexem je v polovin stepBig.
        /// Pokud je StepBig rozdlen na lich poet dl, jako medium jsou povaovny 2 prostedn.
        /// </summary>
        private bool IsAutoMedium(long i, int step2Multiple)
        {
            double rest = i % step2Multiple;
            if (i < 0) rest = (step2Multiple + rest);
            if (step2Multiple % 2 == 0)
            {
                if (rest == step2Multiple / 2) return true;
            }
            else
            {
                int a = step2Multiple / 2;
                if (rest == a) return true;
                else if (rest == a + 1) return true;
            }
            return false;
        }

        private static Func<double, string> defaultFormatter = new Func<double, string>(d => d.ToString());

        private bool IsMultiple(double number, double multiplier)
        {
            double mod = number % multiplier;
            if (Math.Abs(mod) < 0.000001) return true;
            if (Math.Abs(mod - multiplier) < 0.000001) return true;
            return false;
        }

        /// <summary>
        /// Vrt popisky pro osu
        /// </summary>
        /// <param name="from">Potek viditeln sti osy v netransformovanch (internch) souadnicch</param>
        /// <param name="to">Konec viditeln sti osy v netransformovanch (internch) souadnicch</param>
        /// <param name="minSpacing">Minimln vzdlenost mezi 2 popisky v netransformovanch (internch) souadnicch</param>
        /// <returns>LabelInfo definujc popisky osy. Vrcen souadnice jsou opt v netransformovanch (internch) souadnicch. </returns>
        public abstract GetAxisLabelsResult GetAxisLabels(double from, double to, double minSpacing);

        public abstract string GetLabelTextFull(double value);

        public virtual double Transform(double value)
        {
            return value;
        }

        public virtual double TransformBack(double value)
        {
            return value;
        }

        protected void OnRedrawRequired(EventArgs args)
        {
            if (RedrawRequired != null) RedrawRequired(this, args);
        }

        public event EventHandler RedrawRequired;

        protected override abstract Freezable CreateInstanceCore();
    }
}