namespace Brjnk.CoordinateSystem2D.Layers
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.ComponentModel;
    using System.Linq;
    using System.Windows;
    using System.Windows.Controls;
    using System.Windows.Media;
    using Brjnk.Common;

    /// <summary>
    /// </summary>
    public class PhotosLayer : LayerBase
    {
        public PhotosLayer()
        {
            positionChangedListener = new WeakEventListener(RefreshPosition);
            thumbnailChangedListener = new WeakEventListener(RefreshThumbnail);
            rotationChangedListener = new WeakEventListener(RefreshRotation);
            isVisibleChangedListener = new WeakEventListener(RefreshVisibility);
            this.Content = canvas;
        }

        private Dictionary<IPhoto, PhotoInternal> photos = new Dictionary<IPhoto, PhotoInternal>();

        private Brjnk.Common.UI.Canvas canvas = new Brjnk.Common.UI.Canvas();
        private PhotoWindow photoWindow;

        protected override void InfoPopupOpening(object sender, OpenPopupEventArgs e)
        {
            var items = FindPhotosAt(e.RelativePosition);
            if (items == null) this.InfoContent = null;
            else this.InfoContent = LayerHelper.CreateContent(PhotoInfoTemplate, PhotoInfoTemplateSelector, items.SelectMany(p => p.GetAllPhotos()));
        }

        protected override void MenuOpening(object sender, OpenPopupEventArgs e)
        {
            var items = FindPhotosAt(e.RelativePosition);
            if (items == null) this.MenuContent = null;
            else this.MenuContent = LayerHelper.CreateContent(PhotoMenuTemplate, PhotoMenuTemplateSelector, items.SelectMany(p => p.GetAllPhotos()));
        }

        // Pi kliknut na miniaturu se zobraz fotka v novm okn
        public override void MouseClick(MousePosition mousePosition)
        {
            var items = FindPhotosAt(mousePosition.RelativePosition);
            if (items == null) return;
            ShowImages(items.SelectMany(p => p.GetAllPhotos()));
        }

        private List<PhotoInternal> FindPhotosAt(Point relativePosition)
        {
            List<HitTestResult> hits = new List<HitTestResult>();
            VisualTreeHelper.HitTest(
                canvas,
                o => (o as FrameworkElement).Visibility == Visibility.Visible ? HitTestFilterBehavior.Continue : HitTestFilterBehavior.ContinueSkipSelfAndChildren,
                 a => { hits.Add(a); return HitTestResultBehavior.Continue; },
                new PointHitTestParameters(relativePosition));
            if (hits.Count == 0) return null;

            List<PhotoInternal> selected = new List<PhotoInternal>();

            foreach (var result in hits)
            {
                selected.AddRange(photos.Values.Where(p => object.ReferenceEquals(p.Element, result.VisualHit)));
            }
            return selected;
        }

        private void ShowImages(IEnumerable<PhotoInternal> images)
        {
            ShowImages(images.Select(i => i.Photo).ToList());
        }

        public void ShowImages(IList<IPhoto> images)
        {
            if (photoWindow == null)
            {
                photoWindow = new PhotoWindow();
                photoWindow.Closed += new EventHandler((sender, e) => photoWindow = null);
                photoWindow.Show();
            }
            photoWindow.SetTooltipTemplate(TooltipTemplate, TooltipTemplateSelector);
            photoWindow.ShowImages(images);
        }

        protected override void ViewUpdated()
        {
            //aktualizovat obrazovkov souadnice
            foreach (PhotoInternal photo in photos.Values)
            {
                photo.UpdateScreenPosition(view);
            }
        }

        protected override void ViewUpdateCompleted()
        {
            if (GroupsEnabled)
            {
                var items = photos.Values.ToArray();
                //minimln vzdlenost, kterou mohou bt fotografie od sebe
                //pokud jsou ble zobraz se jako skupina
                double minDistance = Math.Max(this.RenderSize.Width, this.RenderSize.Height);

                for (int i = 0; i < items.Length - 1; i++)
                {
                    var item1 = items[i];
                    if (item1 == null) continue;
                    var nearItems = GetItemsInGroup(items, minDistance, i + 1, item1);
                    if (nearItems != null && nearItems.Count != 0)
                    {
                        nearItems.Insert(0, item1);
                        item1.SetGroup(nearItems, GroupedPhotosOrder);
                        item1.Show();
                        items[i] = null;
                    }
                }

                for (int i = 0; i < items.Length; i++)
                {
                    var item = items[i];
                    if (item != null)
                    {
                        item.Show();
                        item.ResetGroup();
                    }
                }
            }
            ViewUpdated();
        }

        private List<PhotoInternal> GetItemsInGroup(PhotoInternal[] items, double minDistance, int start, PhotoInternal item1)
        {
            Lazy<List<PhotoInternal>> nearItems = new Lazy<List<PhotoInternal>>();

            for (int n = start; n < items.Length; n++)
            {
                var item2 = items[n];
                if (item2 == null) continue;
                if ((item1.ScreenPosition - item2.ScreenPosition).Length < minDistance)
                {
                    item2.Hide();
                    nearItems.Value.Add(item2);
                    items[n] = null;
                }
            }
            if (nearItems.IsValueCreated == false) return null;
            return nearItems.Value;
        }

        protected override void ViewChanged(ViewInfoBase oldValue, ViewInfoBase viewInfoBase)
        {
            TransformationChanged();
        }

        protected override void TransformationChanged()
        {
            foreach (PhotoInternal photo in photos.Values)
            {
                photo.UpdateTransformed(view);
            }
            ViewUpdated();
        }

        #region properties

        public System.Collections.Specialized.INotifyCollectionChanged Photos
        {
            get { return (System.Collections.Specialized.INotifyCollectionChanged)GetValue(PhotosProperty); }
            set { SetValue(PhotosProperty, value); }
        }

        public static readonly DependencyProperty PhotosProperty =
            DependencyProperty.Register("Photos", typeof(System.Collections.Specialized.INotifyCollectionChanged), typeof(PhotosLayer), new UIPropertyMetadata(null, PhotosChanged));

        public DataTemplate PhotoInfoTemplate
        {
            get { return (DataTemplate)GetValue(PhotoInfoTemplateProperty); }
            set { SetValue(PhotoInfoTemplateProperty, value); }
        }

        public static readonly DependencyProperty PhotoInfoTemplateProperty =
            DependencyProperty.Register("PhotoInfoTemplate", typeof(DataTemplate), typeof(PhotosLayer), new UIPropertyMetadata(null));

        public DataTemplate PhotoMenuTemplate
        {
            get { return (DataTemplate)GetValue(PhotoMenuTemplateProperty); }
            set { SetValue(PhotoMenuTemplateProperty, value); }
        }

        public static readonly DependencyProperty PhotoMenuTemplateProperty =
            DependencyProperty.Register("PhotoMenuTemplate", typeof(DataTemplate), typeof(PhotosLayer), new UIPropertyMetadata(null));

        public DataTemplateSelector PhotoInfoTemplateSelector
        {
            get { return (DataTemplateSelector)GetValue(PhotoInfoTemplateSelectorProperty); }
            set { SetValue(PhotoInfoTemplateSelectorProperty, value); }
        }

        public static readonly DependencyProperty PhotoInfoTemplateSelectorProperty =
            DependencyProperty.Register("PhotoInfoTemplateSelector", typeof(DataTemplateSelector), typeof(PhotosLayer), new UIPropertyMetadata(null));

        public DataTemplateSelector PhotoMenuTemplateSelector
        {
            get { return (DataTemplateSelector)GetValue(PhotoMenuTemplateSelectorProperty); }
            set { SetValue(PhotoMenuTemplateSelectorProperty, value); }
        }

        public static readonly DependencyProperty PhotoMenuTemplateSelectorProperty =
            DependencyProperty.Register("PhotoMenuTemplateSelector", typeof(DataTemplateSelector), typeof(PhotosLayer), new UIPropertyMetadata(null));

        public Size RenderSize
        {
            get { return (Size)GetValue(RenderSizeProperty); }
            set { SetValue(RenderSizeProperty, value); }
        }

        public static readonly DependencyProperty RenderSizeProperty =
            DependencyProperty.Register("RenderSize", typeof(Size), typeof(PhotosLayer), new UIPropertyMetadata(new Size(32, 32), RenderSizeChanged));

        /// <summary>
        /// DataTemplate pro Tooltip nad otevenm obrzkem (tj. obrzkem zobrazenm v okn).
        /// ablonu tooltipu zobrazenho nad miniaturou definuje PhotoInfoTemplate!
        /// </summary>
        public DataTemplate TooltipTemplate
        {
            get { return (DataTemplate)GetValue(TooltipTemplateProperty); }
            set { SetValue(TooltipTemplateProperty, value); }
        }

        public static readonly DependencyProperty TooltipTemplateProperty =
            DependencyProperty.Register("TooltipTemplate", typeof(DataTemplate), typeof(PhotosLayer), new UIPropertyMetadata(null, PhotoWindowTooltipTemplateChanged));

        /// <summary>
        ///
        /// </summary>
        public DataTemplateSelector TooltipTemplateSelector
        {
            get { return (DataTemplateSelector)GetValue(TooltipTemplateSelectorProperty); }
            set { SetValue(TooltipTemplateSelectorProperty, value); }
        }

        public static readonly DependencyProperty TooltipTemplateSelectorProperty =
            DependencyProperty.Register("TooltipTemplateSelector", typeof(DataTemplateSelector), typeof(PhotosLayer), new UIPropertyMetadata(null, PhotoWindowTooltipTemplateChanged));

        /// <summary>
        /// True, pokud m bt povoleno seskupen fotografi, kter jsou pobl.
        /// </summary>
        public bool GroupsEnabled
        {
            get { return (bool)GetValue(GroupsEnabledProperty); }
            set { SetValue(GroupsEnabledProperty, value); }
        }

        public static readonly DependencyProperty GroupsEnabledProperty =
            DependencyProperty.Register("GroupsEnabled", typeof(bool), typeof(PhotosLayer), new UIPropertyMetadata(true, GroupsEnabledChanged));

        /// <summary>
        /// Delegt Comparison slouc pro seazen seskupench poloek
        /// </summary>
        public Comparison<IPhoto> GroupedPhotosOrder
        {
            get { return (Comparison<IPhoto>)GetValue(GroupedPhotosOrderProperty); }
            set { SetValue(GroupedPhotosOrderProperty, value); }
        }

        public static readonly DependencyProperty GroupedPhotosOrderProperty =
            DependencyProperty.Register("GroupedPhotosOrder", typeof(Comparison<IPhoto>), typeof(PhotosLayer), new UIPropertyMetadata(null, PhotosOrderChanged));

        private static void GroupsEnabledChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
        {
            var target = sender as PhotosLayer;
            if (target.GroupsEnabled)
            {
                target.ViewUpdateCompleted();
            }
            else
            {
                foreach (var p in target.photos.Values)
                {
                    p.ResetGroup();
                    p.Show();
                }
            }
        }

        private static void PhotosOrderChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
        {
            var target = sender as PhotosLayer;
            target.ViewUpdated();
        }

        private static void PhotosChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
        {
            var target = sender as PhotosLayer;
            var collectionNew = e.NewValue as System.Collections.Specialized.INotifyCollectionChanged;
            var collectionOld = e.OldValue as System.Collections.Specialized.INotifyCollectionChanged;
            if ((object)collectionNew != null) collectionNew.CollectionChanged += new System.Collections.Specialized.NotifyCollectionChangedEventHandler(target.collectionNew_CollectionChanged);
            if (collectionOld != null) collectionOld.CollectionChanged -= new System.Collections.Specialized.NotifyCollectionChangedEventHandler(target.collectionNew_CollectionChanged);
            target.photos.Clear();
            target.canvas.Children.Clear();
            target.AddPhotos(target.Photos as IEnumerable);
        }

        private static void RenderSizeChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
        {
            var target = sender as PhotosLayer;
            var size = (Size)e.NewValue;
            foreach (var photo in target.photos.Values)
            {
                photo.UpdateSize(size);
            }
            target.ViewUpdateCompleted();
        }

        private static void PhotoWindowTooltipTemplateChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
        {
            var target = sender as PhotosLayer;
            if (target.photoWindow != null) target.photoWindow.SetTooltipTemplate(target.TooltipTemplate, target.TooltipTemplateSelector);
        }

        #endregion properties

        private void collectionNew_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == System.Collections.Specialized.NotifyCollectionChangedAction.Reset)
            { photos.Clear(); this.canvas.Children.Clear(); }
            else
            {
                AddPhotos(e.NewItems);

                //Pokud se jedn o Move, oldItems obsahuje poloky, kter nemmaj bt odstranny (zstanou v seznamu na jinm mst)
                if (e.Action != System.Collections.Specialized.NotifyCollectionChangedAction.Move)
                    RemovePhotos(e.OldItems);
            }
        }

        private void RemovePhotos(IEnumerable items)
        {
            if (items == null) return;
            foreach (IPhoto item in items)
            {
                PhotoInternal photoInternal;
                if (photos.TryGetValue(item, out photoInternal))
                {
                    photos.Remove(item);
                    canvas.Children.Remove(photoInternal.Element);
                }
            }
            ViewUpdateCompleted();
        }

        private void AddPhotos(IEnumerable items)
        {
            if (items == null) return;
            foreach (IPhoto item in items)
            {
                if (photos.ContainsKey(item) == false)
                {
                    var photoInternal = new PhotoInternal(item, view, this.RenderSize);
                    photos[item] = photoInternal;
                    canvas.Children.Add(photoInternal.Element);
                    AddPropertyChangedListener(item);
                }
            }
            ViewUpdateCompleted();
        }

        private WeakEventListener positionChangedListener;
        private WeakEventListener thumbnailChangedListener;
        private WeakEventListener rotationChangedListener;
        private WeakEventListener isVisibleChangedListener;

        private void AddPropertyChangedListener(IPhoto item)
        {
            PropertyChangedEventManager.AddListener(item, positionChangedListener, "Position");
            PropertyChangedEventManager.AddListener(item, thumbnailChangedListener, "Thumbnail");
            PropertyChangedEventManager.AddListener(item, rotationChangedListener, "PhotoOrientation");
            PropertyChangedEventManager.AddListener(item, isVisibleChangedListener, "IsVisible");
        }

        private bool RefreshThumbnail(Type managerType, object sender, EventArgs e)
        {
            var photo = sender as IPhoto;
            photos[photo].UpdateThumbnail();
            return false;
        }

        private bool RefreshPosition(Type managerType, object sender, EventArgs e)
        {
            var photo = sender as IPhoto;
            photos[photo].UpdateTransformed(view);
            return false;
        }

        private bool RefreshRotation(Type managerType, object sender, EventArgs e)
        {
            var photo = sender as IPhoto;
            photos[photo].UpdateRotation((sender as IPhoto).Orientation);
            return false;
        }

        private bool RefreshVisibility(Type managerType, object sender, EventArgs e)
        {
            var photo = sender as IPhoto;
            photos[photo].UpdateVisibility();
            return false;
        }
    }
}