namespace Brjnk.CoordinateSystem2D
{
    using System;
    using System.Windows;
    using System.Windows.Media.Animation;
    using Brjnk.Common;

    /// <summary>
    /// Uruje pohled pro Area2D
    /// </summary>
    public class ViewInfo : ViewInfoBase
    {
        public ViewInfo(out SetPropertyHandler updater)
            : base(out updater)
        { }

        /// <summary>
        /// Posune pohled o vektor. Vzdlenost je v DIP
        /// </summary>
        /// <param name="v"></param>
        public void Move(Vector v, bool suppressAnim = false)
        {
            var center = new Point(CenterInternal.X + v.X / ZoomX, CenterInternal.Y + v.Y / ZoomY);
            SetView(center, Definition.ZoomX, Definition.ZoomY, suppressAnim);
        }

        public void SetBounds(Point minPoint, Point maxPoint, bool sameXY)
        {
            var point1 = TransformBack(minPoint);
            var point2 = TransformBack(maxPoint);
            Point center = new Point(point1.X + (point2.X - point1.X) / 2, point1.Y + (point2.Y - point1.Y) / 2);
            double zoomX, zoomY;

            if (point1 == point2)
            {
                zoomX = 1; zoomY = 1;
            }
            else
            {
                zoomX = base.AreaWidthDip / Math.Abs(point2.X - point1.X);
                zoomY = base.AreaHeightDip / Math.Abs(point2.Y - point1.Y);
            }

            if (sameXY)
            {
                double zoom = Math.Min(zoomX, zoomY);
                zoomX = zoom;
                zoomY = zoom;
            }

            SetView(center, zoomX, zoomY);
        }

        /// <summary>
        /// Posune pohled o vektor. Vzdlenost je v DIP a v souadnicch obrazovky.
        /// (Obrazovka m jin smr osy y ne matematick soustava souadnic)
        /// </summary>
        public void MoveByMouse(Vector v)
        {
            Move(new Vector(v.X, -v.Y), true);
        }

        /// <param name="relativePoint">Sted biblen vzhledem k levmu hornmu rohu kontroly</param>
        public void ZoomByMouse(Point relativePoint, int zoomSteps)
        {
            double x = ZoomX * Math.Pow(2, zoomSteps);
            double y = ZoomY * Math.Pow(2, zoomSteps);
            CoerceZoom(ref x, ref  y);

            Point mouse = FromRelativeToInternal(relativePoint);
            Vector toCenter = CenterInternal - mouse;

            var center = mouse + new Vector(toCenter.X * (ZoomX / x), toCenter.Y * (ZoomY / y));

            SetView(center, x, y, false);
        }

        public override void SetZoom(double x, double y)
        {
            CoerceZoom(ref x, ref  y);
            SetView(Definition.Center, x, y);
        }

        private void CoerceZoom(ref double x, ref double y)
        {
        }

        private void SetView(Point newCenter, double zoomX, double zoomY, bool suppressAnim)
        {
            CoerceZoom(ref zoomX, ref  zoomY);

            var def = new ViewDefinition(newCenter, zoomX, zoomY);
            if (suppressAnim || Animations == ViewAnimationMode.Disabled)
            {
                BeginAnimation(null);
                Definition = def;
                OnChangeCompleted();
            }
            else
            {
                var anim = new ViewAnimation() { From = Definition, To = def, Duration = TimeSpan.FromMilliseconds(500) };
                if (Animations == ViewAnimationMode.LowPerformance) Timeline.SetDesiredFrameRate(anim, 5);
                BeginAnimation(anim);
                Definition = def;
            }
        }

        #region Anim

        private AnimationClock animClock;

        private void BeginAnimation(ViewAnimation anim)
        {
            if (animClock != null) StopAnim();

            if (anim != null)
            {
                this.animClock = anim.CreateClock();
                this.animClock.CurrentTimeInvalidated += new EventHandler(animClock_CurrentTimeInvalidated);
                this.animClock.Completed += new EventHandler(AnimCompleted);
            }
        }

        private void StopAnim()
        {
            this.animClock.Completed -= new EventHandler(AnimCompleted);
            this.animClock.CurrentTimeInvalidated -= new EventHandler(animClock_CurrentTimeInvalidated);
            animClock = null;
            OnChangeCompleted();
        }

        private void AnimCompleted(object sender, EventArgs e)
        {
            StopAnim();
        }

        private void animClock_CurrentTimeInvalidated(object sender, EventArgs e)
        {
            this.SetValue(DefinitionPropertyKey, animClock.GetCurrentValue(null, null));
        }

        protected override void SetView(Point center, double zoomX, double zoomY)
        {
            SetView(center, zoomX, zoomY, true);
        }

        #endregion Anim

        public ViewAnimationMode Animations
        {
            get { return (ViewAnimationMode)GetValue(AnimationsProperty); }
            protected set { SetValue(AnimationsPropertyKey, value); }
        }

        private static readonly DependencyPropertyKey AnimationsPropertyKey =
            DependencyProperty.RegisterReadOnly("Animations", typeof(ViewAnimationMode), typeof(ViewInfo), new PropertyMetadata(ViewAnimationMode.Enabled));

        public static readonly DependencyProperty AnimationsProperty = AnimationsPropertyKey.DependencyProperty;
    }

    public enum ViewAnimationMode
    {
        Disabled = 0,
        Enabled = 1,
        LowPerformance = 2
    }
}