namespace Brjnk.GpxExplorer.Data.ImportExport
{
    using System;
    using System.Collections.Generic;
    using System.Xml.Serialization;
    using Brjnk.GpxExplorer.Data.GPX;
    using Brjnk.GpxExplorer.Data.GPX.SerializationClasses;
    using Brjnk.GpxExplorer.ViewModel;

    /// <summary>
    /// </summary>
    public class GpxWriter : IDataWriter
    {
        private bool geocaching;

        public GpxWriter(bool geocaching)
        {
            this.geocaching = geocaching;
        }

        public void Write(IEnumerable<WaypointVM> waypoints, IEnumerable<RouteVM> routes, IEnumerable<TrackVM> tracks, System.IO.Stream target)
        {
            if (geocaching) SerializationOptions.SetGeocachingExport();
            else SerializationOptions.SetDefaultExport();

            Bounds bounds = GetBounds(waypoints, routes, tracks);
            var data = new gpxType();
            data.wpt = GetWaypoints(waypoints);
            data.rte = GetRoutes(routes);
            data.trk = GetTracks(tracks);

            data.metadata = new metadataType()
            {
                time = DateTime.UtcNow,
                timeSpecified = true,
                bounds = new boundsType() { minlat = (decimal)bounds.Min.Y, minlon = (decimal)bounds.Min.X, maxlat = (decimal)bounds.Max.Y, maxlon = (decimal)bounds.Max.X }
            };

            var ns = new XmlSerializerNamespaces();
            ns.Add("groundspeak", "http://www.groundspeak.com/cache/1/1");
            ns.Add("groundspeak", "http://www.groundspeak.com/cache/1/0");

            ns.Add("gpxx", "http://www.garmin.com/xmlschemas/GpxExtensions/v1");
            ns.Add("gpxx", "http://www.garmin.com/xmlschemas/GpxExtensions/v2");
            ns.Add("gpxx", "http://www.garmin.com/xmlschemas/GpxExtensions/v3");

            XmlSerializer serializer;
            if (SerializationOptions.GpxVersion == GpxVersion.v1_1)
            {
                serializer = new XmlSerializer(typeof(gpxType));
                serializer.Serialize(target, data, ns);
            }
            else
            {
                var version1_0 = new gpx1_0(data);
                serializer = new XmlSerializer(typeof(gpx1_0));
                serializer.Serialize(target, version1_0, ns);
            }
        }

        private List<trkType> GetTracks(IEnumerable<TrackVM> tracks)
        {
            var retVal = new List<trkType>();
            foreach (var item in tracks)
            {
                IGpxObject obj = item.GetWrappedObject() as IGpxObject;
                if (obj == null) obj = item.Source.ConvertTo<GpxTrack>();
                retVal.Add((trkType)obj.GetUnderlayingObject());
            }
            return retVal;
        }

        private List<rteType> GetRoutes(IEnumerable<RouteVM> routes)
        {
            var retVal = new List<rteType>();
            foreach (var item in routes)
            {
                IGpxObject obj = item.GetWrappedObject() as IGpxObject;
                if (obj == null) obj = item.Source.ConvertTo<GpxRoute>();
                retVal.Add((rteType)obj.GetUnderlayingObject());
            }
            return retVal;
        }

        private List<wptType> GetWaypoints(IEnumerable<WaypointVM> waypoints)
        {
            var retVal = new List<wptType>();
            foreach (var item in waypoints)
            {
                IGpxObject obj = item.GetWrappedObject() as IGpxObject;
                if (obj == null) obj = item.Source.ConvertTo<GpxWaypoint>();
                retVal.Add((wptType)obj.GetUnderlayingObject());
            }
            return retVal;
        }

        private Bounds GetBounds(IEnumerable<WaypointVM> waypoints, IEnumerable<RouteVM> routes, IEnumerable<TrackVM> tracks)
        {
            BoundsBuilder builder = new BoundsBuilder();
            foreach (var item in waypoints)
                item.GetBounds(builder);
            foreach (var item in routes)
                item.GetBounds(builder);
            foreach (var item in tracks)
                item.GetBounds(builder);
            return builder.GetBounds();
        }
    }
}