namespace Brjnk.GpxExplorer.Localization
{
    using System;
    using System.Collections.Generic;
    using System.ComponentModel;
    using System.Linq;
    using System.Windows.Data;

    /// <summary>
    /// </summary>
    public class LocalizationManager : System.ComponentModel.INotifyPropertyChanged
    {
        private IsoLanguageName language;
        private ILocalizationDataProvider provider;
        private IEnumerable<IsoLanguageName> availableLanguages;

        private LocalizationManager()
        {
        }

        private void LoadStrings()
        {
            if ((object)provider == null) strings = null;
            else
            {
                SaveChanges();
                strings = provider.GetStrings(language);
                customStrings = provider.GetCustomStrings(language);
            }
            OnPropertyChanged(string.Empty);
        }

        private string GetCurrentLanguage()
        {
            return System.Globalization.CultureInfo.CurrentUICulture.ThreeLetterISOLanguageName;
        }

        public IsoLanguageName Language
        {
            get { return language; }
            set
            {
                SetLanguage(new IsoLanguageName(value));
                LoadStrings();
                OnPropertyChanged("Language");
            }
        }

        public IEnumerable<IsoLanguageName> AvailableLanguages { get { return availableLanguages; } }

        private void SetLanguage(IsoLanguageName isoLanguageName)
        {
            if (provider == null) return;
            if (availableLanguages != null && availableLanguages.Contains(isoLanguageName))
                language = isoLanguageName;
            else
            {
                language = provider.GetNearestAvailableLanguage(isoLanguageName);
            }
        }

        public void SetProvider(ILocalizationDataProvider provider)
        {
            if (object.ReferenceEquals(provider, this.provider)) return;
            if (provider == null) throw new ArgumentNullException("provider");
            this.provider = provider;
            SetLanguage(new IsoLanguageName(GetCurrentLanguage()));
            availableLanguages = provider.GetAvailableLanguages().ToList();
            OnPropertyChanged("AvailableLanguages");
            LoadStrings();
        }

        /// <summary>
        /// Vrt lokalizovan etzec. Pokud nebyl kl nalezen, vrt kl.
        /// </summary>
        public string GetString(string key)
        {
            if (string.IsNullOrEmpty(key)) return "Empty localization key!!";
            if ((object)strings == null) return key;
            if (customStrings.ContainsKey(key)) return customStrings[key];
            else if (strings.ContainsKey(key)) return strings[key];
            return key;
        }

        public void Update(string key, string newValue)
        {
            customStrings[key] = newValue;
            OnPropertyChanged(string.Empty);
        }

        public void Resume(string key)
        {
            customStrings.Remove(key);
            OnPropertyChanged(string.Empty);
        }

        public void SaveChanges()
        {
            if ((object)customStrings != null)
                provider.UpdateCustomStrings(language, customStrings);
        }

        /// <summary>
        /// Vrt data binding na lokalizovan text.
        /// </summary>
        public Binding GetBinding(string key)
        {
            Binding b = new Binding("[" + key + "]");
            b.Source = this;
            return b;
        }

        public string this[string key]
        {
            get { return GetString(key); }
        }

        private IDictionary<string, string> customStrings;
        private IDictionary<string, string> strings;

        private static LocalizationManager instance;

        public static LocalizationManager Instance
        {
            get
            {
                if (instance == null) instance = new LocalizationManager();
                return instance;
            }
        }

        public event System.ComponentModel.PropertyChangedEventHandler PropertyChanged;

        private void OnPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }
    }
}