namespace Brjnk.GpxExplorer.Localization
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Windows;
    using System.Xml.Linq;

    /// <summary>
    /// </summary>
    public class XmlLocalizationDataProvider : ILocalizationDataProvider
    {
        private Uri baseUri;

        public XmlLocalizationDataProvider(Uri baseUri)
        {
            this.baseUri = baseUri;
        }

        public IDictionary<string, string> GetStrings(IsoLanguageName ISOLanguageName)
        {
            Dictionary<string, string> strings = new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase);
            List<IsoLanguageName> loaded = new List<IsoLanguageName>();

            LoadLocFile(ISOLanguageName, strings, loaded);
            LoadLocFile(new IsoLanguageName("ENG"), strings, loaded);
            LoadLocFile(new IsoLanguageName("CES"), strings, loaded);

            return strings;
        }

        private void LoadLocFile(IsoLanguageName ISOLanguageName, Dictionary<string, string> strings, List<IsoLanguageName> loaded)
        {
            IsoLanguageName baseLanguage = ISOLanguageName;
            while (true)
            {
                if (loaded.Contains(ISOLanguageName)) break;
                var stream = GetStream(baseLanguage);
                if ((object)stream == null) break;
                loaded.Add(baseLanguage);
                baseLanguage = new IsoLanguageName(ParseXml(stream, strings).ToUpper());
            }
        }

        private string ParseXml(System.IO.Stream stream, Dictionary<string, string> strings)
        {
            XDocument doc = XDocument.Load(stream);
            var root = doc.Root;
            if (root.Name != "LocalizationData")
            {
                System.Diagnostics.Debug.WriteLine("Dolo k chyb pi natn lokalizace.");
                return string.Empty;
            }

            var baseLang = root.Attribute("baseLanguage");
            var elements = root.Elements("LocalizationStrings").Elements("s");
            foreach (var i in elements)
            {
                var key = i.Attribute("n");
                if ((object)key == null) continue;
                if (strings.ContainsKey(key.Value)) continue;
                strings[key.Value] = i.Value;
            }

            if ((object)baseLang == null) return string.Empty;
            return baseLang.Value;
        }

        private System.IO.Stream GetStream(IsoLanguageName ISOLanguageName)
        {
            string fileName = ISOLanguageName + ".xml";

            //  Nast z Resources, pokud je jazyk dostupn
            var uri = new Uri(baseUri, new Uri(@"Localization/" + fileName, UriKind.Relative));
            var resourceStream = Application.GetResourceStream(uri);
            if ((object)resourceStream != null) return resourceStream.Stream;

            //  Nast z externho souboru ve sloce Localization
            var root = AppDomain.CurrentDomain.BaseDirectory;
            var path = System.IO.Path.Combine(root, @"Localization\" + fileName);
            if (System.IO.File.Exists(path)) return new System.IO.FileStream(path, System.IO.FileMode.Open);

            //  Lokalizan soubor se nepodailo nalzt
            return null;
        }

        private IDictionary<string, string> customStrings = new Dictionary<string, string>();

        public IDictionary<string, string> GetCustomStrings(IsoLanguageName ISOLanguageName)
        {
            return customStrings;
        }

        public void UpdateCustomStrings(IsoLanguageName ISOLanguageName, IDictionary<string, string> strings)
        {
            customStrings = strings;

            HashSet<string> exportedKeys = new HashSet<string>();
            XDocument doc = new XDocument();
            XElement root = new XElement("LocalizationData");
            XElement stringRoot = new XElement("LocalizationStrings");

            foreach (var customString in customStrings.Concat(this.GetStrings(ISOLanguageName)))
            {
                if (exportedKeys.Contains(customString.Key)) continue;
                var item = new XElement("s");
                item.SetAttributeValue("n", customString.Key);
                item.Add(customString.Value);
                stringRoot.Add(item);
            }
            root.Add(stringRoot);
            doc.Add(root);
            if (customStrings.Count != 0 && MessageBox.Show("Dolo ke zmn lokalizanch text. Chcete exportovat vechny lokalizan text do schrnky? V opanm ppad budou zmny ztraceny.", "", MessageBoxButton.YesNo) == MessageBoxResult.Yes)
            {
                Clipboard.SetText(doc.ToString());
                MessageBox.Show("Lokalizan data byly vloeny do schrnky (Clipboard)");
            }
        }

        public IsoLanguageName GetNearestAvailableLanguage(IsoLanguageName language)
        {
            var available = GetAvailableLanguages().ToList();
            if (available.Contains(language)) return language;
            else return new IsoLanguageName("ENG");
        }

        public IEnumerable<IsoLanguageName> GetAvailableLanguages()
        {
            var uri = new Uri(baseUri, new Uri(@"Localization/AvailableLanguages.txt", UriKind.Relative));
            var resourceStream = Application.GetResourceStream(uri);
            if ((object)resourceStream != null)
            {
                var reader = new System.IO.StreamReader(resourceStream.Stream);
                var all = reader.ReadToEnd();
                var names = all.Split(';');
                return names.Select(n => new IsoLanguageName(n));
            }
            return null;
        }
    }
}