namespace Brjnk.GpxExplorer
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Windows;
    using System.Windows.Controls;
    using System.Windows.Media;
    using System.Windows.Media.Imaging;
    using Brjnk.Common;
    using Brjnk.Common.UI;
    using Brjnk.CoordinateSystem2D;
    using Brjnk.CoordinateSystem2D.Layers;
    using Brjnk.Maps;

    /// <summary>
    /// </summary>
    public class MapLayer : LayerBase
    {
        private List<TileInfo> tiles = new List<TileInfo>();

        private VisualHost host = new VisualHost();
        private bool fullRefresh = false; //pokud je nastaveno na true, budou vechny dladice znovu nateny (nap. po zmn mapovho zdroje)
        private ZoomLevel zoom;
        private TextBlock infoTextBlock = new System.Windows.Controls.TextBlock(); //Textblock zobrzujc souadnice v InfoPopup

        public MapLayer()
        {
            host.Render += new EventHandler<RenderEventArgs>(host_Render);
            this.Content = host;

            this.InfoContent = infoTextBlock;
            infoTextBlock.FontSize = 9;
        }

        private void host_Render(object sender, RenderEventArgs e)
        {
            var context = e.Context;
            var rotationCenter = view.ToRelative(new Point(0, 0));
            context.PushTransform(new RotateTransform(Rotation, rotationCenter.X, rotationCenter.Y));
            foreach (var tile in oldTiles.Concat(tiles))
            {
                if (!tile.Task.IsCompleted) continue;
                var bitmap = GetBitmap(tile);
                context.DrawImage(bitmap, GetDestination(tile));
            }
            context.Pop();
        }

        protected override void ViewUpdated()
        {
            RequestRedraw();
        }

        protected override void ViewUpdateCompleted()
        {
            RefreshTileList();
            RequestRedraw();
        }

        protected override void ViewChanged(ViewInfoBase oldValue, ViewInfoBase viewInfoBase)
        {
            if (view != null)
            {
                System.Diagnostics.Debug.Assert(view.AxisX is LinearAxis, "Osa X mus bt linern pro zobrazen mapy");
                System.Diagnostics.Debug.Assert(view.AxisY is LinearAxis, "Osa Y mus bt linern pro zobrazen mapy");
                System.Diagnostics.Debug.Assert((view.AxisY as LinearAxis).Reverse, "Pro zobrazen mapy mus bt osa Y otoen (Reverse=true)");
            }
        }

        protected override void InfoPopupOpening(object sender, Brjnk.CoordinateSystem2D.OpenPopupEventArgs e)
        {
            infoTextBlock.Inlines.Clear();
            if (Map == null || !ShowGpsCoordinatesInPopup) return;
            var coord = GetGpsCoordinates(e.Position);
            infoTextBlock.Inlines.Add(coord.ToString(Brjnk.Common.UnitsSettings.Instance.Gps));
        }

        /// <param name="position">Souadnice v Area2D, nikoliv relativn</param>
        public GpsCoordinates GetGpsCoordinates(Point position)
        {
            return Map.Projection.ToGpsCoordinates(MathEx.RotatePointD(position, -Rotation)); ;
        }

        /// <summary>
        /// Vygeneruje seznam dladic, kter jsou aktuln zobrazen
        /// </summary>
        private void RefreshTileList()
        {
            if (view == null) return;
            if (fullRefresh || Map == null) { tiles.Clear(); oldTiles.Clear(); fullRefresh = false; };
            this.GpsCoordinates = Map.Projection.ToGpsCoordinates(MathEx.RotatePointD(view.Center, -Rotation));

            if (SelectZoomLevel() == false) return;

            int tilesInRow = (int)Math.Ceiling(Math.Max(view.LeftToRight, view.TopToBottom) /
                Math.Min(zoom.InternalTileSize.Width, zoom.InternalTileSize.Height) *
                Math.Sqrt(2));

            int tilesHalf = (int)Math.Ceiling(tilesInRow / 2.0);

            Point center = MathEx.RotatePointD(view.Center, -Rotation); //intern souadnice (v ppad rotace pepotan zpt)
            Point centerT = Map.GetTileCoordinates(center, zoom.Num); //dladicov souadnice, sted pohledu
            Point centerTileLeftTop = new Point(center.X - (centerT.X - Math.Floor(centerT.X)) * zoom.InternalTileSize.Width,
                center.Y - (centerT.Y - Math.Floor(centerT.Y)) * zoom.InternalTileSize.Height); //intern souadnice levho hornho rohu stedov dladiky
            var centerTile = new TileId((long)Math.Floor(centerT.X), (long)Math.Floor(centerT.Y), zoom.Num);

            BeginTileRefreshing();
            //for (int y = 0; y <= 2 * tilesHalf; y++)
            //{
            //    int row = y - tilesHalf;
            //    for (int x = 0; x <= 2 * tilesHalf; x++)
            //    {
            foreach (var r in GetTilesOrder(tilesHalf))
            {
                int column = r.Item2;
                int row = r.Item1;
                var tileId = new TileId(centerTile.X + column, centerTile.Y + row, zoom.Num);
                if (Map.IsTileAvailable(tileId))
                {
                    var destination = new Point(centerTileLeftTop.X + column * zoom.InternalTileSize.Width, centerTileLeftTop.Y + row * zoom.InternalTileSize.Height);
                    AddTile(tileId, destination);
                };
            }
            //    }
            //}
            ShowInfo();
            EndTileRefreshing();
        }

        private IEnumerable<Tuple<int, int>> GetTilesOrder(int halfCount)
        {
            for (int i = 0; i <= halfCount; i++)
            {
                if (i == 0) { yield return new Tuple<int, int>(0, 0); }
                else
                {
                    for (int x = -i; x <= i; x++)
                    {
                        yield return new Tuple<int, int>(x, i);
                        yield return new Tuple<int, int>(x, -i);
                    }

                    for (int y = -i + 1; y < i; y++)
                    {
                        yield return new Tuple<int, int>(i, y);
                        yield return new Tuple<int, int>(-i, y);
                    }
                }
            }
        }

        private bool SelectZoomLevel()
        {
            var internalTileSize = view.GetInternalSize(Map.TileDIPSize);
            ZoomLevel selectedLevel = default(ZoomLevel); bool specified = false;
            foreach (var level in Map.ZoomLevels)
            {
                if (specified == false) { specified = true; selectedLevel = level; }
                if (level.InternalTileSize.Width >= internalTileSize.Width && level.InternalTileSize.Height >= internalTileSize.Height)
                    selectedLevel = level;
            }
            if (specified == false)
            {
                System.Diagnostics.Debug.WriteLine("Neexistuje vhodn ZoomLevel pro zobrazen mapy.");
                return false;
            }
            this.zoom = selectedLevel;
            return true;
        }

        //informace pro debugovn
        private void ShowInfo()
        {
            System.Diagnostics.Debug.WriteLine(created, "Created");
            System.Diagnostics.Debug.WriteLine(reused, "Reused");
            System.Diagnostics.Debug.WriteLine("------------");
            created = 0; reused = 0;
        }

        private void RequestRedraw()
        {
            host.InvalidateVisual();
        }

        private Rect GetDestination(TileInfo tile)
        {
            var leftTop = view.ToRelative(tile.RenderPosition);
            var size = view.GetDipSize(Map.ZoomLevels.Where(z => z.Num == tile.Task.TileId.Zoom).FirstOrDefault().InternalTileSize);
            return new Rect(leftTop, size);
        }

        /// <summary>
        /// Vrt dladici jako bitmapu. Pokud byla dladice spn zskna, vrt ji.
        /// V opanm ppad vygeneruje dladici s chybovm oznmanm.
        /// Pokud probh zskvn dladice, vrt null;
        /// </summary>
        private BitmapSource GetBitmap(TileInfo tile)
        {
            if (tile.Task.IsCompleted)
            {
                return tile.Tile;
            }
            return null;
        }

        private int tilesCount = 0; //poet dladic v seznamu (pro metodu EndTileRefreshing)

        private void BeginTileRefreshing()
        {
            if (tiles.Capacity > 400) System.Diagnostics.Debug.WriteLine("Neoekvn velk velikost kolekce tiles!");
            tilesCount = 0;
            if (fullRefresh) { tiles.Clear(); oldTiles.Clear(); fullRefresh = false; };
        }

        private void AddTile(TileId tile, Point position)
        {
            for (int i = 0; i < tiles.Count; i++)
            {
                var t = tiles[i];
                //pokud ji dladice je v seznamu, bude znovu pouita
                if (t.Task.TileId.Equals(tile))
                {
                    var temp = tiles[tilesCount];
                    tiles[tilesCount] = t;
                    tiles[i] = temp;
                    tilesCount++;
                    reused++;
                    return;
                }
            }

            //pokud dladice neexistuje, bude vytvoena
            var info = new TileInfo(Map.GetTile(tile, Maps.Downloader.DownloadTaskPriority.Normal, NetworkMode), position, RequestRedraw);
            created++;
            if (tiles.Count == tilesCount) { tiles.Add(info); tilesCount++; } //pokud je kolekce pili mal, je volno Add()
            else { var temp = tiles[tilesCount]; tiles[tilesCount] = info; tiles.Add(temp); tilesCount++; } //pidn nov dladice za posledn
        }

        private void EndTileRefreshing()
        {
            for (int i = tiles.Count - 1; i >= tilesCount; i--)
            {
                var tile = tiles[i];
                tile.Destroy();
                if (tile.Task.IsCompleted) AddOldTile(tile);
            }
            int count = tilesCount;// Math.Min(Math.Max(tilesCount, 300), tiles.Count);
            tiles.RemoveRange(count, tiles.Count - count);
        }

        /// <summary>
        /// Pid dladici do fronty starch dladic, kter jsou pouity, nejsou-li k dispozici dladice poadovan rovn piblen.
        /// </summary>
        private void AddOldTile(TileInfo tile)
        {
            oldTiles.Enqueue(tile);
            if (oldTiles.Count > 100) oldTiles.Dequeue();
        }

        private Queue<TileInfo> oldTiles = new Queue<TileInfo>();

        //debug informace
        private int created = 0;

        private int reused = 0;

        #region Properties

        public IMapProvider Map
        {
            get { return (IMapProvider)GetValue(MapProperty); }
            set { SetValue(MapProperty, value); }
        }

        public static readonly DependencyProperty MapProperty =
            DependencyProperty.Register("Map", typeof(IMapProvider), typeof(MapLayer), new UIPropertyMetadata(null, MapProviderChanged));

        public double Rotation
        {
            get { return (double)GetValue(RotationProperty); }
            set { SetValue(RotationProperty, value); }
        }

        public static readonly DependencyProperty RotationProperty =
            DependencyProperty.Register("Rotation", typeof(double), typeof(MapLayer), new UIPropertyMetadata(0.0, RotationChanged));

        public bool Animations
        {
            get { return (bool)GetValue(AnimationsProperty); }
            set { SetValue(AnimationsProperty, value); }
        }

        public static readonly DependencyProperty AnimationsProperty =
            DependencyProperty.Register("Animations", typeof(bool), typeof(MapLayer), new UIPropertyMetadata(false));

        public NetworkMode NetworkMode
        {
            get { return (NetworkMode)GetValue(NetworkModeProperty); }
            set { SetValue(NetworkModeProperty, value); }
        }

        public static readonly DependencyProperty NetworkModeProperty =
            DependencyProperty.Register("NetworkMode", typeof(NetworkMode), typeof(MapLayer), new UIPropertyMetadata(NetworkMode.Online));

        public GpsCoordinates GpsCoordinates
        {
            get { return (GpsCoordinates)GetValue(GpsCoordinatesProperty); }
            protected set { SetValue(GpsCoordinatesPropertyKey, value); }
        }

        private static readonly DependencyPropertyKey GpsCoordinatesPropertyKey =
            DependencyProperty.RegisterReadOnly("GpsCoordinates", typeof(GpsCoordinates), typeof(MapLayer), new PropertyMetadata());

        public static readonly DependencyProperty GpsCoordinatesProperty = GpsCoordinatesPropertyKey.DependencyProperty;

        public bool ShowGpsCoordinatesInPopup
        {
            get { return (bool)GetValue(ShowGpsCoordinatesInPopupProperty); }
            set { SetValue(ShowGpsCoordinatesInPopupProperty, value); }
        }

        public static readonly DependencyProperty ShowGpsCoordinatesInPopupProperty =
            DependencyProperty.Register("ShowGpsCoordinatesInPopup", typeof(bool), typeof(MapLayer), new UIPropertyMetadata(true));

        private static void RotationChanged(object sender, DependencyPropertyChangedEventArgs e)
        {
            (sender as MapLayer).ViewUpdated();
        }

        private static void MapProviderChanged(object sender, DependencyPropertyChangedEventArgs e)
        {
            var target = (sender as MapLayer);
            target.fullRefresh = true;
            target.ViewUpdateCompleted();
        }

        #endregion Properties

        private class TileInfo
        {
            private Action redrawRequest;

            public TileInfo(GetTileTask tile, Point internalPosition, Action redrawDelegate)
            {
                redrawRequest = redrawDelegate;
                this.Task = tile;
                this.RenderPosition = internalPosition;
                if (tile.IsCompleted == false)
                {
                    tile.Completed += new EventHandler(tile_Completed);
                }
                else redrawRequest();
            }

            private void tile_Completed(object sender, EventArgs e)
            {
                Task.Completed -= new EventHandler(tile_Completed);
                if (Task.Exception != null)
                {
                    var args = TileGeneratorArgs.GetRed();
                    args.Text = Task.Exception.Message.ToString();
                    generatedTile = TileGenerator.Generate(args, 256, 256, 96);
                }
                redrawRequest();
            }

            public double Alpha { get; set; }

            public Point RenderPosition { get; protected set; }

            public GetTileTask Task { get; protected set; }

            private BitmapSource generatedTile;

            public BitmapSource Tile
            {
                get
                {
                    if (generatedTile != null) return generatedTile;
                    return Task.Image;
                }
            }

            internal void Destroy()
            {
                Task.CancelDownload();
            }
        }
    }
}