namespace Brjnk.GpxExplorer.Data
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using Brjnk.Common.Extensions;
    using Brjnk.GpxExplorer.Computing;
    using Brjnk.GpxExplorer.ViewModel;
    using Brjnk.GpxExplorer.Workspace;
    using Brjnk.MVVM;

    public class Photos : DataItem
    {
        private ObservableCollectionEx<TrackVM> tracks;

        private DataChangeWatcher watcher;

        public ObservableCollectionEx<Photo> Collection
        {
            get { return GetWatchedProperty<ObservableCollectionEx<Photo>>("Collection"); }
        }

        public Photos(ObservableCollectionEx<TrackVM> tracks, DataChangeWatcher watcher)
        {
            DefineWatchedProperty("Collection", new ObservableCollectionEx<Photo>());
            this.watcher = watcher;
            watcher.ObjectAdded += new EventHandler<WatcherEventArgs<object>>(watcher_ObjectAdded);
            watcher.ObjectRemoved += new EventHandler<WatcherEventArgs<object>>(watcher_ObjectRemoved);
            watcher.TracksReseted += new EventHandler(watcher_TracksReseted);
            watcher.TrackDataChanged += new EventHandler<WatcherEventArgs<TrackVM>>(watcher_TrackDataChanged);
            watcher.PhotosReseted += new EventHandler(watcher_PhotosReseted);
            watcher.PhotoTimeChanged += new EventHandler<WatcherEventArgs<Photo>>(watcher_PhotoTimeChanged);
            this.tracks = tracks;
        }

        private void watcher_PhotosReseted(object sender, EventArgs e)
        {
            RefreshAll();
        }

        private void watcher_PhotoTimeChanged(object sender, WatcherEventArgs<Photo> e)
        {
            RefreshPhoto(e.Object);
        }

        private void watcher_ObjectRemoved(object sender, WatcherEventArgs<object> e)
        {
            if (e.ObjectType == typeof(TrackVM))
            {
                foreach (var p in Collection)
                {
                    p.Locations.RemoveWhere(lp => object.ReferenceEquals(lp.Locator, e.Object));
                }
            }
            RefreshCounters();
        }

        private void watcher_TrackDataChanged(object sender, WatcherEventArgs<TrackVM> e)
        {
            RefreshTrack(e.Object);
        }

        private void watcher_TracksReseted(object sender, EventArgs e)
        {
            RefreshAll();
        }

        private void watcher_ObjectAdded(object sender, WatcherEventArgs<object> e)
        {
            if (e.ObjectType == typeof(TrackVM))
            {
                RefreshTrack(e.Object as TrackVM);
            }
            else if (e.ObjectType == typeof(Photo))
            {
                RefreshPhoto(e.Object as Photo);
            }
        }

        private void RefreshAll()
        {
            var dateTimes = Collection.Select(p => p.CorrectedCreationTime);
            Dictionary<TrackVM, IDictionary<DateTime, FindTimeResult>> results = CreateFindTimeResults(tracks, dateTimes);

            foreach (var photo in Collection)
            {
                photo.Locations.Clear();
                CreateLocatedPhoto(photo, results);
            }
            RefreshCounters();
        }

        private void RefreshPhoto(Photo photo)
        {
            photo.Locations.Clear();
            var results = CreateFindTimeResults(tracks, Enumerable.Repeat(photo.CorrectedCreationTime, 1));
            CreateLocatedPhoto(photo, results);
            RefreshCounters();
        }

        private void RefreshTrack(TrackVM track)
        {
            foreach (var photo in this.Collection)
            {
                photo.Locations.RemoveWhere(lp => lp.Locator == track);
                var results = CreateFindTimeResults(Enumerable.Repeat(track, 1), Enumerable.Repeat(photo.CorrectedCreationTime, 1));
                CreateLocatedPhoto(photo, results);
            }
            RefreshCounters();
        }

        /// <summary>
        /// Vrt FindTimeResult pro vechny stopy a pedan datumy.
        /// </summary>
        private Dictionary<TrackVM, IDictionary<DateTime, FindTimeResult>> CreateFindTimeResults(IEnumerable<TrackVM> tracks, IEnumerable<DateTime> dateTimes)
        {
            Dictionary<TrackVM, IDictionary<DateTime, FindTimeResult>> results = new Dictionary<TrackVM, IDictionary<DateTime, FindTimeResult>>();

            foreach (var track in tracks)
            {
                var result = track.FindTime(dateTimes.ToList(), false);
                if (result != null) results.Add(track, result);
            }
            return results;
        }

        /// <summary>
        /// Pd fotografii nov umstn (LocatedPhoto) podle pedanch vsledk.
        /// Pvodn LocatedPhoto NEJSOU odstranny.
        /// </summary>
        private void CreateLocatedPhoto(Photo photo, Dictionary<TrackVM, IDictionary<DateTime, FindTimeResult>> results)
        {
            foreach (var trackResults in results)
            {
                FindTimeResult r;
                if (trackResults.Value.TryGetValue(photo.CorrectedCreationTime, out r) != false)
                {
                    photo.Locations.Add(new LocatedPhoto(photo, trackResults.Key, r));
                }
            }
        }

        private void RefreshCounters()
        {
            int located = 0, unlocated = 0, ambiguous = 0; int withoutExif = 0;
            foreach (var photo in Collection)
            {
                var c = photo.Locations.Count;
                if (c == 0) unlocated++;
                else if (c == 1) located++;
                else ambiguous++;
                if (photo.FileCreationTimeUsed) withoutExif++;
            }
            LocatedCount = located;
            UnlocatedCount = unlocated;
            AmbiguousCount = ambiguous;
            WithoutExifCount = withoutExif;
        }

        private int withoutExifCount;

        public int WithoutExifCount
        {
            get { return withoutExifCount; }
            set { withoutExifCount = value; OnPropertyChanged("WithoutExifCount"); }
        }

        private int locatedCount;

        public int LocatedCount
        {
            get { return locatedCount; }
            set { locatedCount = value; OnPropertyChanged("LocatedCount"); }
        }

        private int unlocatedCount;

        public int UnlocatedCount
        {
            get { return unlocatedCount; }
            set { unlocatedCount = value; OnPropertyChanged("UnlocatedCount"); }
        }

        private int ambiguousCount;

        public int AmbiguousCount
        {
            get { return ambiguousCount; }
            set { ambiguousCount = value; OnPropertyChanged("AmbiguousCount"); }
        }
    }
}