using System;
using Brjnk.Common.ValueTypes;

namespace Brjnk.GpxExplorer.ViewModel
{
    /// <summary>
    /// Nullable tda, kter rozliuje 2 null stavy (Unavailable a Unset)
    /// </summary>
    public struct TrackPointAttributeValue<T> where T : struct
    {
        private T val;

        private AttributeValueState state;

        public TrackPointAttributeValue(T value, bool guessed)
            : this()
        {
            this.val = value;
            state = guessed ? AttributeValueState.ValueGuessed : AttributeValueState.ValueAvailable;
        }

        public TrackPointAttributeValue(T value)
            : this(value, false) { }

        public T Value
        {
            get
            {
                if (state == AttributeValueState.ValueUnavailable || state == AttributeValueState.ValueUnset)
                    throw new InvalidOperationException("TrackPointAttributeValue nem hodnotu");
                return val;
            }
        }

        public AttributeValueState State
        {
            get { return state; }
        }

        public static implicit operator TrackPointAttributeValue<T>(T value)
        {
            return new TrackPointAttributeValue<T>(value, false);
        }

        public override string ToString()
        {
            switch (state)
            {
                case AttributeValueState.ValueAvailable:
                    return val.ToString();
                case AttributeValueState.ValueUnset:
                    return Localization.LocalizationManager.Instance.GetString("unset");
                case AttributeValueState.ValueGuessed:
                    return string.Format("[{0}]", val.ToString());
                case AttributeValueState.ValueUnavailable:
                    return Localization.LocalizationManager.Instance.GetString("unavailable");
                default:
                    throw new Exception("Unknown AttributeValueState");
            }
        }

        public double GetDoubleValue()
        {
            double value;
            var type = val.GetType();

            if (typeof(IPhysicalQuantity).IsAssignableFrom(type)) value = ((IPhysicalQuantity)val).GetValue();
            else if (type == typeof(double)) value = (double)Convert.ChangeType(val, typeof(double));
            else if (type == typeof(TimeSpan)) return ((TimeSpan)Convert.ChangeType(this.val, typeof(TimeSpan))).TotalSeconds;
            else if (type == typeof(DateTime)) return ((DateTime)Convert.ChangeType(this.val, typeof(DateTime))).Ticks / 10000000;
            else throw new Exception("Unsupported Type");

            return value;
        }

        public TrackPointAttributeValue<double> ToDouble()
        {
            var retVal = new TrackPointAttributeValue<double>();
            retVal.val = this.GetDoubleValue();
            retVal.state = this.state;
            return retVal;
        }

        public bool HasValue
        {
            get
            {
                return state == AttributeValueState.ValueAvailable || state == AttributeValueState.ValueGuessed;
            }
        }

        public bool IsUnset
        {
            get
            {
                return state == AttributeValueState.ValueUnset;
            }
        }

        public bool IsGuessed
        {
            get
            {
                return state == AttributeValueState.ValueGuessed;
            }
        }

        public bool IsUnavailable
        {
            get
            {
                return state == AttributeValueState.ValueUnavailable;
            }
        }

        public static TrackPointAttributeValue<T> CreateUnset()
        {
            var retVal = new TrackPointAttributeValue<T>();
            retVal.state = AttributeValueState.ValueUnset;
            return retVal;
        }

        public static TrackPointAttributeValue<T> CreateUnavailable()
        {
            var retVal = new TrackPointAttributeValue<T>();
            retVal.state = AttributeValueState.ValueUnavailable;
            return retVal;
        }
    }

    public enum AttributeValueState : byte
    {
        ValueUnset,
        ValueAvailable,
        ValueGuessed,
        ValueUnavailable
    }
}