namespace Brjnk.GpxExplorer.Workspace
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Windows;
    using System.Windows.Media;
    using Brjnk.CoordinateSystem2D;
    using Brjnk.CoordinateSystem2D.Layers;
    using Brjnk.GpxExplorer.ViewModel;

    /// <summary>
    /// </summary>
    public abstract class DataAdapter : DataAdapterBase
    {
        public DataAdapter(DataModel model, SearchService searchService, DataView view)
            : base(model, searchService, view)
        {
        }

        #region Properties

        private double thickness;

        protected double Thickness
        {
            get { return thickness; }
            set { thickness = value; base.Invalidate(); }
        }

        private bool fillGaps;

        public bool FillGaps
        {
            get { return fillGaps; }
            set
            {
                fillGaps = value;
                base.Invalidate(DataAdapterItemType.Track);
            }
        }

        private bool colorHighlight;

        public bool ColorHighlight
        {
            get { return colorHighlight; }
            set
            {
                colorHighlight = value;
                base.Invalidate(DataAdapterItemType.Track);
            }
        }

        private double colorHighlightLevel = 0.5;

        public double ColorHighlightLevel
        {
            get { return colorHighlightLevel; }
            set
            {
                if (value < 0 || value > 1) throw new ArgumentException("ColorHighlightLevel mus bt od 0 do 1.");
                colorHighlightLevel = value;
                if (colorHighlight) base.Invalidate(DataAdapterItemType.Track);
            }
        }

        private TrackPointAttributeInfo highlightedAttribute;

        public TrackPointAttributeInfo HighlightedAttribute
        {
            get { return highlightedAttribute; }
            set
            {
                highlightedAttribute = value;
                base.Invalidate(DataAdapterItemType.Track);
            }
        }

        private bool includeGaps;

        public bool IncludeGaps
        {
            get { return includeGaps; }
            set
            {
                includeGaps = value;
                base.Invalidate(DataAdapterItemType.Track);
            }
        }

        #endregion Properties

        public void GetBoundingBox(BoundsBuilder builder, IEnumerable<object> items)
        {
            foreach (var item in items)
            {
                Bounds bounds;
                if (shownObjects.TryGetValue(item, out bounds))
                    builder.Add(bounds);
            }
        }

        public Bounds? GetBoundingBox()
        {
            var builder = new BoundsBuilder();
            foreach (var bounds in shownObjects.Values)
            {
                builder.Add(bounds);
            }
            if (builder.IsEmpty) return null;
            return builder.GetBounds();
        }

        protected Dictionary<object, Bounds> shownObjects = new Dictionary<object, Bounds>();

        protected Dictionary<object, PhotoRef> photosD = new Dictionary<object, PhotoRef>();

        protected override void AddObject(object obj, DataAdapterItemType type)
        {
            switch (type)
            {
                case DataAdapterItemType.Unknown:
                    throw new Exception("Unknown DataAdapterItemType");
                case DataAdapterItemType.SearchPositionRequest:
                    ProcessSearchRequest((SearchRequest)obj);
                    break;

                case DataAdapterItemType.SearchTimeRequest:
                    ProcessSearchRequest((SearchRequest)obj);
                    break;

                case DataAdapterItemType.SearchPositionResult:
                    ProcessSearchResult((SearchResult)obj);
                    break;

                case DataAdapterItemType.SearchTimeResult:
                    ProcessSearchResult((SearchResult)obj);
                    break;

                case DataAdapterItemType.SearchWaypointResult:
                    ProcessSearchWaypointResult((SearchPositionResult)obj);
                    break;

                case DataAdapterItemType.SearchPhotoResult:
                    ProcessSearchPhotoResult((SearchPositionResult)obj);
                    break;

                case DataAdapterItemType.LocatedPhoto:
                    ProcessLocatedPhoto((LocatedPhoto)obj);
                    break;

                case DataAdapterItemType.Waypoint:
                    ProcessWaypoint((WaypointVM)obj);
                    break;

                case DataAdapterItemType.Route:
                    ProcessRoute((RouteVM)obj);
                    break;

                case DataAdapterItemType.Track:
                    ProcessTrack((TrackVM)obj);
                    break;

                case DataAdapterItemType.RoutePoint:
                    ProcessRoutePoint((RoutePointVM)obj);
                    break;

                case DataAdapterItemType.TrackPoint:
                    ProcessTrackPoint((TrackPointVM)obj);
                    break;

                default:
                    throw new Exception("Unknown DataAdapterItemType");
            }
        }

        protected virtual void ProcessTrack(TrackVM track)
        {
            TrackPointVM previous = null, next = null, current = null;
            BoundsBuilder builder = new BoundsBuilder();

            List<GraphPoint> data = new List<GraphPoint>();
            GraphPoint? toAdd = null;
            double? minZ = null, maxZ = null;

            var segments = track.Segments;
            for (int s = 0; s < segments.Count; s++)
            {
                var segment = segments[s];
                var points = segment.Points;
                for (int p = 0; p < points.Count; p++)
                {
                    current = points[p];
                    if (p > 0) previous = points[p - 1];
                    else if (s > 0 && IncludeGaps) previous = segments[s - 1].Points.Last();

                    if (p < points.Count - 1) next = points[p + 1];
                    else if (s < segments.Count - 1 && includeGaps) next = segments[s + 1].Points.First();

                    double? zValue = GetZValue(track, previous, current, next);
                    if (!zValue.HasValue || double.IsInfinity(zValue.Value)) continue;
                    if (minZ.HasValue == false || zValue.Value < minZ) minZ = zValue.Value;
                    if (maxZ.HasValue == false || zValue > maxZ) maxZ = zValue.Value;

                    var pos = GetTrackPointPosition(track, previous, current, next);
                    if (!ValidatePoint(ref pos)) continue;

                    if (toAdd.HasValue) { data.Add(toAdd.Value); toAdd = null; }
                    toAdd = new GraphPoint(pos.Value, zValue.Value, current);
                    builder.AddPoint(pos.Value);
                }

                if (toAdd.HasValue)
                {
                    var p = toAdd.Value;
                    p.Interrupt = !FillGaps;
                    data.Add(p);
                    toAdd = null;
                }
            }
            var dataInfo = new GraphDataInfo(minZ ?? 1, maxZ ?? 1, GetGraphConstantInfo(track));
            Tracks.AddDataSerie(track, track, data, dataInfo, CreateGraphStyle(track));
            shownObjects.Add(track, builder.GetBounds());
        }

        protected double? GetZValue(TrackVM track, TrackPointVM previous, TrackPointVM current, TrackPointVM next)
        {
            if (ColorHighlight == false) return 1.0;
            var value = current.GetAttributeValue(HighlightedAttribute, includeGaps);
            if (value.State == AttributeValueState.ValueUnset && HighlightedAttribute.Type == TrackPointAttributeType.Type2 && (object)next != null)
                return next.GetAttributeValue(HighlightedAttribute, includeGaps).Value;
            if (value.HasValue) return value.Value;
            return null;
        }

        private void ProcessRoute(RouteVM route)
        {
            List<GraphPoint> data = new List<GraphPoint>();
            BoundsBuilder builder = new BoundsBuilder();
            var points = route.Points;

            RoutePointVM previous = null, next = null, current = null;

            for (int i = 0; i < points.Count; i++)
            {
                current = points[i];

                if (i == 0) previous = null;
                else previous = points[i - 1];

                if (i == points.Count - 1) next = null;
                else previous = points[i + 1];

                var pos = GetRoutePointPosition(route, previous, current, next);

                if (!ValidatePoint(ref pos)) continue;
                data.Add(new GraphPoint(pos.Value, 1.0, current));
                builder.AddPoint(pos.Value);
            }

            var dataInfo = new GraphDataInfo(1.0, 1.0, GetGraphConstantInfo(route));
            Routes.AddDataSerie(route, route, data, dataInfo, CreateGraphStyle(route));
            shownObjects.Add(route, builder.GetBounds());
        }

        private void ProcessTrackPoint(TrackPointVM trackPoint)
        {
            var track = (TrackVM)trackPoint.Parent.Parent;
            TrackPointVM previous, next;
            FindPreviousNext(trackPoint, out previous, out next, IncludeGaps);
            var pos = GetTrackPointPosition(track, previous, trackPoint, next);
            if (!ValidatePoint(ref pos)) return;
            TrackPoints.AddPoint(trackPoint, trackPoint, pos.Value, view.TrackPointStyle);
            shownObjects.Add(trackPoint, new Bounds(pos.Value));
        }

        private void ProcessRoutePoint(RoutePointVM routePoint)
        {
            var route = (RouteVM)routePoint.Parent;
            RoutePointVM previous, next;
            FindPreviousNext(routePoint, out previous, out next);
            var pos = GetRoutePointPosition(route, previous, routePoint, next);
            if (!ValidatePoint(ref pos)) return;
            RoutePoints.AddPoint(routePoint, routePoint, pos.Value, view.RoutePointStyle);
            shownObjects.Add(routePoint, new Bounds(pos.Value));
        }

        private void FindPreviousNext(RoutePointVM routePoint, out RoutePointVM previous, out RoutePointVM next)
        {
            var route = routePoint.Parent as RouteVM;
            var points = route.Points;
            int index = 0;
            for (int i = 0; i < points.Count; i++)
            {
                if (object.ReferenceEquals(points[i], routePoint))
                {
                    index = i;
                    break;
                }
            }

            if (index == 0) previous = null;
            else previous = points[index - 1];

            if (index == route.Points.Count - 1) next = null;
            else next = points[index + 1];
        }

        private void FindPreviousNext(TrackPointVM trackPoint, out TrackPointVM previous, out TrackPointVM next, bool joined)
        {
            var segment = trackPoint.Parent as TrackSegmentVM;
            IEnumerable<TrackPointVM> enumerable;
            if (joined)
            {
                enumerable = (segment.Parent as TrackVM).GetAllPoints();
            }
            else
            {
                enumerable = segment.Points;
            }

            TrackPointVM p = null, c = null, n = null; //previous, current, next

            foreach (var point in enumerable)
            {
                p = c;
                c = n;
                n = point;
                if (object.ReferenceEquals(c, trackPoint))
                {
                    previous = p;
                    next = n;
                    return;
                }
            }
            if (object.ReferenceEquals(n, trackPoint))
            {
                previous = c;
                next = null;
                return;
            }

            throw new Exception("Bod nebyl nalezen");
        }

        private void ProcessWaypoint(WaypointVM wp)
        {
            var pos = GetWaypointPosition(wp);
            if (!ValidatePoint(ref pos)) return;
            Points.AddPoint(wp, wp, pos.Value, view.WaypointStyle);
            shownObjects.Add(wp, new Bounds(pos.Value));
        }

        private void ProcessSearchPhotoResult(SearchPositionResult result)
        {
            var pos = GetSearchResultPosition(result);
            if (!ValidatePoint(ref pos)) return;
            var reference = new PhotoRef((LocatedPhoto)result.SearchedItem, pos.Value);
            photosD.Add(result, reference);
            Photos.Add(reference);
        }

        private void ProcessSearchWaypointResult(SearchPositionResult result)
        {
            var pos = GetSearchResultPosition(result);
            if (!ValidatePoint(ref pos)) return;
            Points.AddPoint(result, result, pos.Value, view.WaypointStyle);
            shownObjects.Add(result, new Bounds(pos.Value));
        }

        private void ProcessLocatedPhoto(LocatedPhoto photo)
        {
            var pos = GetPhotoPosition(photo);
            if (!ValidatePoint(ref pos)) return;
            var reference = new PhotoRef(photo, pos.Value);
            photosD.Add(photo, reference);
            Photos.Add(reference);
        }

        private void ProcessSearchResult(SearchResult searchResult)
        {
            var pos = GetSearchResultPosition(searchResult);
            if (!ValidatePoint(ref pos)) return;
            SearchResults.AddPoint(searchResult, searchResult, pos.Value, view.SearchPointStyle);
            shownObjects.Add(searchResult, new Bounds(pos.Value));
        }

        private void ProcessSearchRequest(SearchRequest obj)
        {
            var pos = GetSearchRequestPosition(obj as SearchRequest);
            if (!ValidatePoint(ref pos)) return;
            SearchResults.AddPoint(obj, obj, pos.Value, view.SearchPointStyle);
            shownObjects.Add(obj, new Bounds(pos.Value));
        }

        protected override void RemoveObject(object obj, DataAdapterItemType type)
        {
            if (shownObjects.ContainsKey(obj) || photosD.ContainsKey(obj))
            {
                if (type == DataAdapterItemType.SearchPositionRequest || type == DataAdapterItemType.SearchTimeRequest ||
                    type == DataAdapterItemType.SearchPositionResult || type == DataAdapterItemType.SearchTimeResult)
                    SearchResults.DeletePoint(obj);
                else if (type == DataAdapterItemType.SearchWaypointResult || type == DataAdapterItemType.Waypoint)
                    Points.DeletePoint(obj);
                else if (type == DataAdapterItemType.SearchPhotoResult || type == DataAdapterItemType.LocatedPhoto)
                    Photos.Remove(GetPhotoRef(obj));
                else if (type == DataAdapterItemType.Route)
                    Routes.RemoveDataSerie(obj);
                else if (type == DataAdapterItemType.Track)
                    Tracks.RemoveDataSerie(obj);
                else if (type == DataAdapterItemType.RoutePoint)
                    RoutePoints.DeletePoint(obj);
                else if (type == DataAdapterItemType.TrackPoint)
                    TrackPoints.DeletePoint(obj);
                else throw new Exception("Unknown DataAdapterItemType");
                shownObjects.Remove(obj);
            }
        }

        private PhotoRef GetPhotoRef(object photo)
        {
            PhotoRef retVal = photosD[photo];
            photosD.Remove(photo);
            return retVal;
        }

        protected override void UpdateObject(object obj, DataAdapterItemType type)
        {
            RemoveObject(obj, type);
            AddObject(obj, type);
        }

        protected override void SetSelectedState(ViewModel.ISelectable obj, DataAdapterItemType type)
        {
            if ((type == DataAdapterItemType.Route || type == DataAdapterItemType.Track)
                && ((IHideable)obj).IsVisible)
                UpdateObject(obj, type);
        }

        protected virtual GraphStyle CreateGraphStyle(RouteVM route)
        {
            var style = new GraphStyle();
            style.Thickness = this.Thickness;
            style.HighColor = route.Color;
            style.Highlight = route.IsSelected;
            style.LowColor = route.Color;
            style.UseGradient = false;
            return style;
        }

        protected virtual GraphStyle CreateGraphStyle(TrackVM track)
        {
            var style = new GraphStyle();
            style.Thickness = this.Thickness;
            style.HighColor = track.Color;
            style.Highlight = track.IsSelected;
            var maxColor = track.Color;
            double alpha = 255 - (colorHighlightLevel * 255);
            if (ColorHighlight) style.LowColor = Color.FromArgb((byte)alpha, maxColor.R, maxColor.G, maxColor.B);
            else style.LowColor = maxColor;

            style.UseGradient = HighlightedAttribute.Type != TrackPointAttributeType.Type2;
            return style;
        }

        private bool ValidatePoint(ref Point? input)
        {
            if (!input.HasValue) return false;
            if (double.IsNaN(input.Value.X) || double.IsNaN(input.Value.Y)) return false;
            if (double.IsPositiveInfinity(input.Value.X)) input = new Point(10000000, input.Value.Y);
            if (double.IsNegativeInfinity(input.Value.X)) input = new Point(-10000000, input.Value.Y);
            if (double.IsPositiveInfinity(input.Value.Y)) input = new Point(input.Value.X, 10000000);
            if (double.IsNegativeInfinity(input.Value.Y)) input = new Point(input.Value.X, -10000000);
            return true;
        }

        protected abstract Point? GetSearchRequestPosition(SearchRequest request);

        protected abstract Point? GetSearchResultPosition(SearchResult result);

        protected abstract Point? GetPhotoPosition(LocatedPhoto photo);

        protected abstract Point? GetWaypointPosition(WaypointVM waypoint);

        protected abstract Point? GetTrackPointPosition(TrackVM track, TrackPointVM previous, TrackPointVM current, TrackPointVM next);

        protected abstract Point? GetRoutePointPosition(RouteVM route, RoutePointVM previous, RoutePointVM current, RoutePointVM next);

        protected abstract AxisOrientation? GetGraphConstantInfo(RouteVM route);

        protected abstract AxisOrientation? GetGraphConstantInfo(TrackVM route);

        //protected override void AddObject(object obj, DataAdapterItemType type)
        //{
        //    System.Diagnostics.Debug.WriteLine(obj.ToString() + " " + obj.GetHashCode() + " " + type.ToString(), "Add");
        //}

        //protected override void RemoveObject(object obj, DataAdapterItemType type)
        //{
        //    System.Diagnostics.Debug.WriteLine(obj.ToString() + " " + obj.GetHashCode() + " " + type.ToString(), "Remove");
        //}

        //protected override void UpdateObject(object obj, DataAdapterItemType type)
        //{
        //    System.Diagnostics.Debug.WriteLine(obj.ToString() + " " + obj.GetHashCode() + " " + type.ToString(), "Update");
        //}

        //protected override void SetSelectedState(ViewModel.ISelectable obj, DataAdapterItemType type)
        //{
        //    System.Diagnostics.Debug.WriteLine(obj.ToString() + " " + obj.GetHashCode() + " " + type.ToString() + " " + obj.IsSelected.ToString(), "SetSelectedState");
        //}
    }
}