namespace Brjnk.GpxExplorer.Workspace
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using Brjnk.CoordinateSystem2D.Layers;
    using Brjnk.GpxExplorer.ViewModel;
    using Brjnk.MVVM;

    /// <summary>
    /// </summary>
    public abstract class DataAdapterBase
    {
        private DataModel model;

        private SearchService searchService;

        protected DataView view;

        private AreaManager area;

        private HashSet<object> visibleObjects = new HashSet<object>();

        public DataAdapterBase(DataModel model, SearchService searchService, DataView view)
        {
            this.model = model;
            this.searchService = searchService;
            this.view = view;
        }

        #region ChangesWatching

        private void AddHandlers(DataChangeWatcher dataChangeWatcher, SearchService searchService)
        {
            dataChangeWatcher.IsSelectedChanged += new EventHandler<WatcherEventArgs<ViewModel.ISelectable>>(dataChangeWatcher_IsSelectedChanged);
            dataChangeWatcher.IsVisibleChanged += new EventHandler<WatcherEventArgs<ViewModel.IHideable>>(dataChangeWatcher_IsVisibleChanged);
            dataChangeWatcher.ObjectAdded += new EventHandler<WatcherEventArgs<object>>(dataChangeWatcher_ObjectAdded);
            dataChangeWatcher.ObjectRemoved += new EventHandler<WatcherEventArgs<object>>(dataChangeWatcher_ObjectRemoved);
            dataChangeWatcher.RouteDataChanged += new EventHandler<WatcherEventArgs<ViewModel.RouteVM>>(dataChangeWatcher_RouteDataChanged);
            dataChangeWatcher.TrackDataChanged += new EventHandler<WatcherEventArgs<ViewModel.TrackVM>>(dataChangeWatcher_TrackDataChanged);
            dataChangeWatcher.WaypointDataChanged += new EventHandler<WatcherEventArgs<ViewModel.WaypointVM>>(dataChangeWatcher_WaypointDataChanged);
            dataChangeWatcher.PhotosReseted += new EventHandler(dataChangeWatcher_PhotosReseted);
            dataChangeWatcher.RoutesReseted += new EventHandler(dataChangeWatcher_RoutesReseted);
            dataChangeWatcher.TracksReseted += new EventHandler(dataChangeWatcher_TracksReseted);
            dataChangeWatcher.WaypointsReseted += new EventHandler(dataChangeWatcher_WaypointsReseted);

            searchService.Results.ItemsAdded += new EventHandler<MVVM.ItemsEventArgs<SearchResult>>(Results_ItemsAdded);
            searchService.Results.ItemsRemoved += new EventHandler<MVVM.ItemsEventArgs<SearchResult>>(Results_ItemsRemoved);
            searchService.SearchedItems.ItemsAdded += new EventHandler<MVVM.ItemsEventArgs<SearchRequest>>(SearchedItems_ItemsAdded);
            searchService.SearchedItems.ItemsRemoved += new EventHandler<MVVM.ItemsEventArgs<SearchRequest>>(SearchedItems_ItemsRemoved);
        }

        private void RemoveHandlers(DataChangeWatcher dataChangeWatcher, SearchService searchService)
        {
            dataChangeWatcher.IsSelectedChanged -= new EventHandler<WatcherEventArgs<ViewModel.ISelectable>>(dataChangeWatcher_IsSelectedChanged);
            dataChangeWatcher.IsVisibleChanged -= new EventHandler<WatcherEventArgs<ViewModel.IHideable>>(dataChangeWatcher_IsVisibleChanged);
            dataChangeWatcher.ObjectAdded -= new EventHandler<WatcherEventArgs<object>>(dataChangeWatcher_ObjectAdded);
            dataChangeWatcher.ObjectRemoved -= new EventHandler<WatcherEventArgs<object>>(dataChangeWatcher_ObjectRemoved);
            dataChangeWatcher.RouteDataChanged -= new EventHandler<WatcherEventArgs<ViewModel.RouteVM>>(dataChangeWatcher_RouteDataChanged);
            dataChangeWatcher.TrackDataChanged -= new EventHandler<WatcherEventArgs<ViewModel.TrackVM>>(dataChangeWatcher_TrackDataChanged);
            dataChangeWatcher.WaypointDataChanged -= new EventHandler<WatcherEventArgs<ViewModel.WaypointVM>>(dataChangeWatcher_WaypointDataChanged);
            dataChangeWatcher.PhotosReseted -= new EventHandler(dataChangeWatcher_PhotosReseted);
            dataChangeWatcher.RoutesReseted -= new EventHandler(dataChangeWatcher_RoutesReseted);
            dataChangeWatcher.TracksReseted -= new EventHandler(dataChangeWatcher_TracksReseted);
            dataChangeWatcher.WaypointsReseted -= new EventHandler(dataChangeWatcher_WaypointsReseted);

            searchService.Results.ItemsAdded -= new EventHandler<MVVM.ItemsEventArgs<SearchResult>>(Results_ItemsAdded);
            searchService.Results.ItemsRemoved -= new EventHandler<MVVM.ItemsEventArgs<SearchResult>>(Results_ItemsRemoved);
            searchService.SearchedItems.ItemsAdded -= new EventHandler<MVVM.ItemsEventArgs<SearchRequest>>(SearchedItems_ItemsAdded);
            searchService.SearchedItems.ItemsRemoved -= new EventHandler<MVVM.ItemsEventArgs<SearchRequest>>(SearchedItems_ItemsRemoved);
        }

        private void SearchedItems_ItemsRemoved(object sender, MVVM.ItemsEventArgs<SearchRequest> e)
        {
            foreach (var item in e.Items)
            {
                item.Changed -= new EventHandler(item_Changed);
                RemovePrivate(item);
            }
        }

        private void SearchedItems_ItemsAdded(object sender, MVVM.ItemsEventArgs<SearchRequest> e)
        {
            foreach (var item in e.Items)
            {
                item.Changed += new EventHandler(item_Changed);
                AddPrivate(item);
            }
        }

        private void item_Changed(object sender, EventArgs e)
        {
            UpdatePrivate(sender);
        }

        private void Results_ItemsRemoved(object sender, MVVM.ItemsEventArgs<SearchResult> e)
        {
            foreach (var item in e.Items)
            {
                RemovePrivate(item);
            }
        }

        private void Results_ItemsAdded(object sender, MVVM.ItemsEventArgs<SearchResult> e)
        {
            foreach (var item in e.Items)
            {
                if (item.FoundIn is IHideable == false) { AddPrivate(item); continue; }
                if ((item.FoundIn as IHideable).IsVisible) AddPrivate(item);
            }
        }

        private void dataChangeWatcher_WaypointsReseted(object sender, EventArgs e)
        {
            foreach (var item in visibleObjects.ToList())
            {
                if (item is WaypointVM) RemovePrivate(item);
            }
            ReadAllWaypoints();
        }

        private void dataChangeWatcher_TracksReseted(object sender, EventArgs e)
        {
            foreach (var item in visibleObjects.ToList())
            {
                if (item is TrackVM || item is TrackPointVM) RemovePrivate(item);
                if (item is SearchResult && (item as SearchResult).FoundIn is TrackVM) RemovePrivate(item);
            }
            ReadAllTracks();
        }

        private void dataChangeWatcher_RoutesReseted(object sender, EventArgs e)
        {
            foreach (var item in visibleObjects.ToList())
            {
                if (item is RouteVM || item is RoutePointVM) RemovePrivate(item);
                if (item is SearchResult && (item as SearchResult).FoundIn is RouteVM) RemovePrivate(item);
            }

            ReadAllRoutes();
        }

        private void dataChangeWatcher_PhotosReseted(object sender, EventArgs e)
        {
            foreach (var item in visibleObjects.ToList())
            {
                if (item is LocatedPhoto) RemovePrivate(item);
            }

            ReadAllPhotos();
        }

        private void dataChangeWatcher_WaypointDataChanged(object sender, WatcherEventArgs<ViewModel.WaypointVM> e)
        {
            UpdatePrivate(e.Object);
        }

        private void dataChangeWatcher_TrackDataChanged(object sender, WatcherEventArgs<ViewModel.TrackVM> e)
        {
            UpdatePrivate(e.Object);

            foreach (var item in visibleObjects.ToList())
            {
                if (item is TrackPointVM &&
                    object.ReferenceEquals((item as TrackPointVM).Parent.Parent, e.Object))
                    RemovePrivate(item);
            }

            ReadAllTrackPoints(e.Object);
        }

        private void dataChangeWatcher_RouteDataChanged(object sender, WatcherEventArgs<ViewModel.RouteVM> e)
        {
            UpdatePrivate(e.Object);

            foreach (var item in visibleObjects.ToList())
            {
                if (item is RoutePointVM &&
                    object.ReferenceEquals((item as RoutePointVM).Parent, e.Object))
                    RemovePrivate(item);
            }

            ReadAllRoutePoints(e.Object);
        }

        private void dataChangeWatcher_ObjectRemoved(object sender, WatcherEventArgs<object> e)
        {
            if (e.Object is RouteVM)
            {
                RemoveRoute(e.Object as RouteVM);
            }
            else if (e.Object is TrackVM)
            {
                RemoveTrack(e.Object as TrackVM);
            }
            else if (e.Object is Photo)
            {
                foreach (var l in (e.Object as Photo).Locations)
                    RemovePrivate(l);
            }
            else RemovePrivate(e.Object);
        }

        private void dataChangeWatcher_ObjectAdded(object sender, WatcherEventArgs<object> e)
        {
            if (e.Object is RouteVM)
            {
                ReadAllRoutePoints(e.Object as RouteVM);
                ReadAllRouteResults(e.Object as RouteVM);
            }
            else if (e.Object is TrackVM)
            {
                ReadAllTrackPoints(e.Object as TrackVM);
                ReadAllTrackResults(e.Object as TrackVM);
            }
            else if (e.Object is RoutePointVM || e.Object is TrackPointVM || e.Object is TrackSegmentVM) return; //nen teba je pidvat, jsou zpracovny udlost RouteDataChanged

            AddPrivate(e.Object);
        }

        private void dataChangeWatcher_IsVisibleChanged(object sender, WatcherEventArgs<ViewModel.IHideable> e)
        {
            if (e.Object.IsVisible)
            {
                if (e.Object is RouteVM)
                {
                    ReadAllRoutePoints(e.Object as RouteVM);
                    ReadAllRouteResults(e.Object as RouteVM);
                }
                else if (e.Object is TrackVM)
                {
                    ReadAllTrackPoints(e.Object as TrackVM);
                    ReadAllTrackResults(e.Object as TrackVM);
                }
                else if (e.Object is WaypointVM)
                {
                    //ReadAllTracks zahrnuje i zobrazen vech vsledk vyhledvn
                    ReadAllTracks();
                }
                AddPrivate(e.Object);
            }
            else
            {
                if (e.Object is RouteVM)
                {
                    RemoveRoute(e.Object as RouteVM);
                }
                else if (e.Object is TrackVM)
                {
                    RemoveTrack(e.Object as TrackVM);
                }
                else if (e.Object is WaypointVM)
                {
                    RemoveSearchResults(e.Object);
                    RemovePrivate(e.Object);
                }
                else RemovePrivate(e.Object);
            }
        }

        private void RemoveSearchResults(object searchedItem)
        {
            foreach (var item in visibleObjects.ToList())
            {
                if (item is SearchResult && (item as SearchResult).SearchedItem == searchedItem) RemovePrivate(item);
            }
        }

        private void dataChangeWatcher_IsSelectedChanged(object sender, WatcherEventArgs<ViewModel.ISelectable> e)
        {
            if (e.Object is TrackPointVM)
            {
                var trkPt = (e.Object as TrackPointVM);
                if (trkPt.IsSelected && (trkPt.Parent.Parent as TrackVM).IsVisible) AddPrivate(e.Object);
                else RemovePrivate(e.Object);
            }
            else if (e.Object is RoutePointVM)
            {
                var rtePt = (e.Object as RoutePointVM);
                if (rtePt.IsSelected && (rtePt.Parent as RouteVM).IsVisible) AddPrivate(e.Object);
                else RemovePrivate(e.Object);
            }

            SetSelectedState(e.Object, GetItemType(e.Object));
        }

        private void ReadAllTrackPoints(TrackVM track)
        {
            if (!track.IsVisible) return;
            foreach (var p in track.GetAllPoints())
            {
                if (p.IsSelected) AddPrivate(p);
            }
        }

        private void ReadAllRoutePoints(RouteVM route)
        {
            if (!route.IsVisible) return;
            foreach (var p in route.Points)
            {
                if (p.IsSelected) AddPrivate(p);
            }
        }

        private void ReadAllRouteResults(RouteVM route)
        {
            if (!route.IsVisible) return;
            foreach (var result in searchService.Results.Where(r => r.FoundIn == route))
            {
                AddPrivate(result);
            }
        }

        private void ReadAllTrackResults(TrackVM track)
        {
            if (!track.IsVisible) return;
            foreach (var result in searchService.Results.Where(r => r.FoundIn == track))
            {
                if (result.SearchedItem is WaypointVM && ((WaypointVM)result.SearchedItem).IsVisible == false) continue;
                AddPrivate(result);
            }
        }

        private void RemoveTrack(TrackVM obj)
        {
            foreach (var item in visibleObjects.ToList())
            {
                if (item is TrackPointVM && (item as TrackPointVM).Parent.Parent == obj) RemovePrivate(item);
                if (item is SearchResult && (item as SearchResult).FoundIn == obj) RemovePrivate(item);
            }
            RemovePrivate(obj);
        }

        private void RemoveRoute(RouteVM obj)
        {
            foreach (var item in visibleObjects.ToList())
            {
                if (item is RoutePointVM && (item as RoutePointVM).Parent == obj) RemovePrivate(item);
                if (item is SearchResult && (item as SearchResult).FoundIn == obj) RemovePrivate(item);
            }
            RemovePrivate(obj);
        }

        private void ReadAllWaypoints()
        {
            foreach (var p in model.Waypoints)
            {
                if ((p as IHideable).IsVisible) AddPrivate(p);
            }
        }

        private void ReadAllPhotos()
        {
            foreach (var p in model.Photos.Collection.SelectMany(p => p.Locations))
            {
                AddPrivate(p);
            }
        }

        private void ReadAllRoutes()
        {
            foreach (var r in model.Routes)
            {
                if ((r as IHideable).IsVisible)
                {
                    AddPrivate(r);
                    ReadAllRoutePoints(r);
                    ReadAllRouteResults(r);
                }
            }
        }

        private void ReadAllTracks()
        {
            foreach (var t in model.Tracks)
            {
                if ((t as IHideable).IsVisible)
                {
                    AddPrivate(t);
                    ReadAllTrackPoints(t);
                    ReadAllTrackResults(t);
                }
            }
        }

        private void ReadAllRequests()
        {
            foreach (var request in searchService.SearchedItems)
                AddPrivate(request);
        }

        #endregion ChangesWatching

        private void AddPrivate(object obj)
        {
            if (obj is IHideable)
            {
                if (!(obj as IHideable).IsVisible) return;
            }

            if (obj is SearchResult)
            {
                var result = obj as SearchResult;
                if (result.SearchedItem is IHideable)
                {
                    if (!((IHideable)result.SearchedItem).IsVisible) return;
                }
            }

            if (!visibleObjects.Contains(obj))
            {
                var type = GetItemType(obj);
                if (type == DataAdapterItemType.Unknown) return;
                visibleObjects.Add(obj);
                AddObject(obj, type);
            }
        }

        private void RemovePrivate(object obj)
        {
            if (visibleObjects.Contains(obj))
            {
                var type = GetItemType(obj);
                visibleObjects.Remove(obj);
                if (type == DataAdapterItemType.Unknown) return;
                RemoveObject(obj, type);
            }
        }

        private void UpdatePrivate(object obj)
        {
            if (visibleObjects.Contains(obj))
            {
                var type = GetItemType(obj);
                if (type == DataAdapterItemType.Unknown) return;
                UpdateObject(obj, type);
            }
        }

        protected abstract void AddObject(object obj, DataAdapterItemType type);

        protected abstract void RemoveObject(object obj, DataAdapterItemType type);

        protected abstract void UpdateObject(object obj, DataAdapterItemType type);

        protected abstract void SetSelectedState(ISelectable obj, DataAdapterItemType type);

        protected void Invalidate(Type t)
        {
            foreach (var obj in visibleObjects)
            {
                if (t.IsAssignableFrom(obj.GetType()))
                    UpdateObject(obj, GetItemType(obj));
            }
        }

        protected void Invalidate(DataAdapterItemType type)
        {
            foreach (var obj in visibleObjects)
            {
                var currentType = GetItemType(obj);
                if (currentType == type)
                    UpdateObject(obj, type);
            }
        }

        protected void Invalidate()
        {
            foreach (var obj in visibleObjects)
                UpdateObject(obj, GetItemType(obj));
        }

        protected DataAdapterItemType GetItemType(object obj)
        {
            var type = obj.GetType();
            if (obj is SearchPositionRequest) return DataAdapterItemType.SearchPositionRequest;
            else if (obj is SearchDateTimeRequest) return DataAdapterItemType.SearchTimeRequest;
            else if (obj is SearchPositionResult)
            {
                var result = obj as SearchPositionResult;
                if (result.SearchedItem is SearchPositionRequest) return DataAdapterItemType.SearchPositionResult;
                else if (result.SearchedItem is WaypointVM) return DataAdapterItemType.SearchWaypointResult;
                else if (result.SearchedItem is LocatedPhoto) return DataAdapterItemType.SearchPhotoResult;
                else return DataAdapterItemType.Unknown;
            }
            else if (obj is SearchDateTimeResult)
            {
                var result = obj as SearchDateTimeResult;
                if (result.SearchedItem is SearchDateTimeRequest) return DataAdapterItemType.SearchTimeResult;
                else return DataAdapterItemType.Unknown;
            }
            else if (obj is LocatedPhoto) return DataAdapterItemType.LocatedPhoto;
            else if (obj is WaypointVM) return DataAdapterItemType.Waypoint;
            else if (obj is RouteVM) return DataAdapterItemType.Route;
            else if (obj is TrackVM) return DataAdapterItemType.Track;
            else if (obj is RoutePointVM) return DataAdapterItemType.RoutePoint;
            else if (obj is TrackPointVM) return DataAdapterItemType.TrackPoint;
            return DataAdapterItemType.Unknown;
        }

        private bool used = false;

        public void Attach(AreaManager area)
        {
            if (used) throw new InvalidOperationException("Objekt neme bt znovu pouit. Je teba vytvoit novou instanci.");
            used = true;

            this.area = area;
            ClearArea();
            InitializeArea();
            ReadData();
            AddHandlers(model.Watcher, searchService);
        }

        private void ReadData()
        {
            ReadAllWaypoints();
            ReadAllRoutes();
            ReadAllTracks();
            ReadAllPhotos();
            ReadAllRequests();
        }

        public void Detach(bool clear)
        {
            if (area == null) throw new InvalidOperationException("Ped volnm Detach mus bt zavolna Attach. Metoda Detach me bt volna pouze jednou.");
            if (clear) ClearArea();
            RemoveHandlers(model.Watcher, searchService);
            area = null;
        }

        private void ClearArea()
        {
            area.PhotosLayer.Photos = null;
            area.PointsLayer.DataAdapter = null;
            area.RoutePointsLayer.DataAdapter = null;
            area.RoutesLayer.DataAdapter = null;
            area.SearchResultsLayer.DataAdapter = null;
            area.TrackPointsLayer.DataAdapter = null;
            area.TracksLayer.DataAdapter = null;
        }

        protected ObservableCollectionEx<PhotoRef> Photos { get; private set; }

        protected GenericPointsDataAdapter Points { get; private set; }

        protected GenericPointsDataAdapter RoutePoints { get; private set; }

        protected GenericPointsDataAdapter TrackPoints { get; private set; }

        protected GenericPointsDataAdapter SearchResults { get; private set; }

        protected GenericGraphDataAdapter Tracks { get; private set; }

        protected GenericGraphDataAdapter Routes { get; private set; }

        private void InitializeArea()
        {
            Photos = new ObservableCollectionEx<PhotoRef>();
            Points = new GenericPointsDataAdapter();
            RoutePoints = new GenericPointsDataAdapter();
            TrackPoints = new GenericPointsDataAdapter();
            SearchResults = new GenericPointsDataAdapter();
            Tracks = new GenericGraphDataAdapter();
            Routes = new GenericGraphDataAdapter();

            area.PhotosLayer.Photos = Photos;
            area.PointsLayer.DataAdapter = Points;
            area.RoutePointsLayer.DataAdapter = RoutePoints;
            area.RoutesLayer.DataAdapter = Routes;
            area.SearchResultsLayer.DataAdapter = SearchResults;
            area.TrackPointsLayer.DataAdapter = TrackPoints;
            area.TracksLayer.DataAdapter = Tracks;
        }
    }

    public enum DataAdapterItemType
    {
        Unknown,
        SearchPositionRequest,
        SearchTimeRequest,
        SearchPositionResult,
        SearchTimeResult,
        SearchWaypointResult,
        SearchPhotoResult,
        LocatedPhoto,
        Waypoint,
        Route,
        Track,
        RoutePoint,
        TrackPoint
    }
}