namespace Brjnk.GpxExplorer.Workspace
{
    using System;
    using System.Collections;
    using System.Linq;
    using Brjnk.Common;
    using Brjnk.Common.Extensions;
    using Brjnk.GpxExplorer.Data;
    using Brjnk.GpxExplorer.UI;
    using Brjnk.GpxExplorer.ViewModel;

    /// <summary>
    /// </summary>
    public class DataEditor
    {
        private DataModel model;

        public DataEditor(DataModel model)
        {
            this.model = model;
            CreateCommands();
        }

        private void ViewGeocache(object arg)
        {
            var window = new GeocacheDetailWindow(arg as WaypointVM);
            window.ShowDialog();
        }

        private void OpenGeocache(object arg)
        {
            var wp = ((WaypointVM)arg).Source;
            if (string.IsNullOrEmpty(wp.Name)) return;
            string url;
            if (wp.Name.StartsWith("gc", StringComparison.InvariantCultureIgnoreCase))
                url = string.Format("http://www.geocaching.com/seek/cache_details.aspx?wp={0}&Submit6=Go", wp.Name);
            else
            {
                UserMessages.ShowError(Localization.LocalizationManager.Instance.GetString("errorWaypointIsNotGc"));
                return;
            }
            System.Diagnostics.Process.Start(url);
        }

        /// <summary>
        /// Odebere jeden nebo vce Trackpoint
        /// </summary>
        /// <param name="arg">TrackPointVM nebo IEnumerable pro odstrann vce bod</param>
        public void RemoveTrackPoints(object arg)
        {
            var trkPt = arg as TrackPointVM;
            if (trkPt != null)
            {
                var answer = UserMessages.Question(Localization.LocalizationManager.Instance.GetString("questionRemoveTrackPoint"),
                Localization.LocalizationManager.Instance.GetString("ok"),
                Localization.LocalizationManager.Instance.GetString("cancel"));

                if ((answer ?? 1) == 1) return;

                model.BeginUpdate();
                RemoveTrackPoint(trkPt);
                model.EndUpdate();
            }
            else
            {
                var trkPts = arg as IEnumerable;
                if (trkPts == null) return;

                var answer = UserMessages.Question(Localization.LocalizationManager.Instance.GetString("questionRemoveTrackPoints"),
                    Localization.LocalizationManager.Instance.GetString("ok"),
                    Localization.LocalizationManager.Instance.GetString("cancel"));
                if ((answer ?? 1) == 1) return;
                model.BeginUpdate();
                foreach (TrackPointVM point in trkPts.OfType<TrackPointVM>().ToList())
                {
                    RemoveTrackPoint(point);
                }
                model.EndUpdate();
            }
        }

        private void RemoveTrackPoint(TrackPointVM trkPt)
        {
            var parentSeg = (TrackSegmentVM)trkPt.Parent;
            parentSeg.Points.Remove(trkPt);
            if (parentSeg.Points.Count == 0)
            {
                var parentTrack = (TrackVM)parentSeg.Parent;
                parentTrack.Segments.Remove(parentSeg);
            }
        }

        private void RemoveRoutePoints(object arg)
        {
            var rtePt = arg as RoutePointVM;
            if (rtePt != null)
            {
                var answer = UserMessages.Question(Localization.LocalizationManager.Instance.GetString("questionRemoveRoutePoint"),
                    Localization.LocalizationManager.Instance.GetString("ok"),
                    Localization.LocalizationManager.Instance.GetString("cancel"));

                if ((answer ?? 1) == 1) return;

                model.BeginUpdate();
                RemoveRoutePoint(rtePt);
                model.EndUpdate();
            }
            else
            {
                var rtePts = arg as IEnumerable;
                if (rtePts == null) return;

                var answer = UserMessages.Question(Localization.LocalizationManager.Instance.GetString("questionRemoveRoutePoints"),
                    Localization.LocalizationManager.Instance.GetString("ok"),
                    Localization.LocalizationManager.Instance.GetString("cancel"));
                if ((answer ?? 1) == 1) return;
                model.BeginUpdate();
                foreach (RoutePointVM point in rtePts.OfType<RoutePointVM>().ToList())
                {
                    RemoveRoutePoint(point);
                }
                model.EndUpdate();
            }
        }

        private void RemoveRoutePoint(RoutePointVM rtePt)
        {
            var parent = (RouteVM)rtePt.Parent;
            parent.Points.Remove(rtePt);
        }

        private void SimplifyTrack(object arg)
        {
            var track = arg as TrackVM;
            if (track != null)
            {
                var window = new SimplifyTrackWindow();
                if (window.ShowDialog() ?? false)
                {
                    SimlifyTrack(track, window.MaxDistance, window.KeepElevationLocalExtremes);
                }
            }
        }

        private void SimlifyTrack(TrackVM track, Distance distance, bool keepElevationExtremes)
        {
            track.BeginUpdate();

            foreach (var segment in track.Segments)
            {
                TrackPointVM first = null, previous = null, current = null;
                for (int i = 0; i < segment.Points.Count; i++)
                {
                    current = segment.Points[i];
                    if (first == null) first = current;
                    else if (!object.ReferenceEquals(previous, first))
                    {
                        if (current.DistanceJoined.GetDoubleValue() - first.DistanceJoined.GetDoubleValue() < distance
                            && (!keepElevationExtremes || previous.ElevationLocalExtreme == ElevationExtreme.None))
                        {
                            segment.Points.RemoveAt(i - 1);
                            i--;
                        }
                        else
                        {
                            first = previous;
                        }
                    }

                    previous = current;
                }
            }
            track.EndUpdate();
        }

        private void CreateWaypoint(object arg)
        {
            if (arg == null || arg.GetType() != typeof(GpsCoordinates))
            {
                CreateWaypoint(new GpsCoordinates());
            }
            else
            {
                CreateWaypoint((GpsCoordinates)arg);
            }
        }

        public void CreateWaypoint(GpsCoordinates coordinates)
        {
            var window = new CreateWaypointWindow(GetNewWaypointName(), coordinates);
            if (window.ShowDialog() ?? false)
            {
                model.Waypoints.Add(window.NewWaypoint);
            }
        }

        private string GetNewWaypointName()
        {
            for (int i = 0; i < 1000; i++)
            {
                string name = "Wpt" + i.ToString();

                if (!model.Waypoints.Any(w => string.Equals(w.Source.Name, name, StringComparison.InvariantCultureIgnoreCase)))
                {
                    return name;
                }
            }
            return "Wpt";
        }

        private string GetNewRouteName()
        {
            for (int i = 0; i < 1000; i++)
            {
                string name = "Route" + i.ToString();

                if (!model.Routes.Any(w => string.Equals(w.Source.Name, name, StringComparison.InvariantCultureIgnoreCase)))
                {
                    return name;
                }
            }
            return "Route";
        }

        private void CreateRoute(object arg)
        {
            var window = new CreateRouteWindow(GetNewRouteName(), model.Waypoints);
            if ((window.ShowDialog() ?? false))
            {
                model.Routes.Add(window.NewRoute);
            }
        }

        private void SortObjects(object arg)
        {
            var wpts = model.Waypoints.OrderBy(w => w.Source.Name).ToList();
            var rtes = model.Routes.OrderBy(w => w.Source.Name).ToList();
            var trks = model.Tracks.OrderBy(w => w.Source.Name).ToList();

            model.Waypoints.Clear();
            model.Routes.Clear();
            model.Tracks.Clear();

            model.Waypoints.AddRange(wpts);
            model.Routes.AddRange(rtes);
            model.Tracks.AddRange(trks);
        }

        private void MakeVisible(object arg)
        {
            SetVisibility(arg, true);
        }

        private void MakeInvisible(object arg)
        {
            SetVisibility(arg, false);
        }

        private void SetVisibility(object arg, bool visibility)
        {
            if (arg is IEnumerable)
            {
                foreach (IHideable item in ((IEnumerable)arg).Cast<IHideable>())
                    item.IsVisible = visibility;
            }
            else if (arg is IHideable)
            {
                ((IHideable)arg).IsVisible = visibility;
            }
        }

        private void InvertVisibility(object arg)
        {
            if (arg == null)
            {
                foreach (IHideable item in model.GetAllObjects().Cast<IHideable>())
                    item.IsVisible = !item.IsVisible;
            }
            else if (arg is IHideable)
            {
                var item = ((IHideable)arg);
                item.IsVisible = !item.IsVisible;
            }
        }

        private void InvertSelection(object arg)
        {
            if (arg == null)
            {
                foreach (ISelectable item in model.GetAllObjects().Cast<IHideable>())
                    item.IsSelected = !item.IsSelected;
            }
            else if (arg is ISelectable)
            {
                var item = ((ISelectable)arg);
                item.IsSelected = !item.IsSelected;
            }
        }

        private void InverseRoute(object arg)
        {
            var route = (RouteVM)arg;
            route.BeginUpdate();
            for (int i = 0; i < Math.Ceiling(route.Points.Count / 2.0); i++)
            {
                var tmp = route.Points[i];
                route.Points[i] = route.Points[route.Points.Count - 1 - i];
                route.Points[route.Points.Count - 1 - i] = tmp;
            }
            route.EndUpdate();
        }

        private void InverseTrack(object arg)
        {
            var track = (TrackVM)arg;
            track.BeginUpdate();
            for (int i = 0; i < Math.Ceiling(track.Segments.Count / 2.0); i++)
            {
                var tmp = track.Segments[i];
                track.Segments[i] = track.Segments[track.Segments.Count - 1 - i];
                track.Segments[track.Segments.Count - 1 - i] = tmp;
            }
            for (int i = 0; i < track.Segments.Count; i++)
                InverseTrackSegment(track.Segments[i]);
            track.EndUpdate();
        }

        private void InverseTrackSegment(TrackSegmentVM segment)
        {
            for (int i = 0; i < Math.Ceiling(segment.Points.Count / 2.0); i++)
            {
                var tmp = segment.Points[i];
                segment.Points[i] = segment.Points[segment.Points.Count - 1 - i];
                segment.Points[segment.Points.Count - 1 - i] = tmp;
            }
        }

        private void RemoveTrackTimestamp(object arg)
        {
            TrackVM track = (TrackVM)arg;
            track.BeginUpdate();
            track.GetAllPoints().ForEach(p => p.Source.Time = null);
            track.EndUpdate();
        }

        private void RemoveTrackElevation(object arg)
        {
            TrackVM track = (TrackVM)arg;
            track.BeginUpdate();
            track.GetAllPoints().ForEach(p => p.Source.Elevation = null);
            track.EndUpdate();
        }

        private void FlattenTrack(object arg)
        {
            TrackVM track = (TrackVM)arg;
            track.BeginUpdate();
            var points = track.GetAllPoints().ToList();
            track.Segments.Clear();
            var segment = new TrackSegmentVM(new TrackSegmentDefault());
            segment.BeginUpdate();
            foreach (var p in points) segment.Points.Add(p);
            segment.EndUpdate();
            track.Segments.Add(segment);
            track.EndUpdate();
        }

        private void EditTrackPoint(object arg)
        {
            var trackPoint = (TrackPointVM)arg;
            var window = new TrackPointWindow(trackPoint);
            window.Show();
        }

        private void EditRoutePoint(object arg)
        {
            var routePoint = (RoutePointVM)arg;
            var window = new RoutePointWindow(routePoint);
            window.Show();
        }

        #region commands

        private void CreateCommands()
        {
            createGeocacheInfo = new GenericCommand(o => (o as WaypointVM).Source.Geocache = new GeocacheInfoDefault());
            removeGeocacheInfo = new GenericCommand(o => (o as WaypointVM).Source.Geocache = null);
            viewGeocacheCommand = new GenericCommand(ViewGeocache);
            openGeocacheCommand = new GenericCommand(OpenGeocache);
            RemoveTrackPointsCommand = new GenericCommand(RemoveTrackPoints);
            RemoveRoutePointsCommand = new GenericCommand(RemoveRoutePoints);
            SimplifyTrackCommand = new GenericCommand(SimplifyTrack);
            CreateWaypointCommand = new GenericCommand(CreateWaypoint);
            CreateRouteCommand = new GenericCommand(CreateRoute);
            SortObjectsCommand = new GenericCommand(SortObjects);
            MakeVisibleCommand = new GenericCommand(MakeVisible);
            MakeInvisibleCommand = new GenericCommand(MakeInvisible);
            InvertVisibilityCommand = new GenericCommand(InvertVisibility);
            InvertSelectionCommand = new GenericCommand(InvertSelection);
            ReverseRouteCommand = new GenericCommand(InverseRoute);
            ReverseTrackCommand = new GenericCommand(InverseTrack);
            RemoveTrackTimestampCommand = new GenericCommand(RemoveTrackTimestamp);
            RemoveTrackElevationCommand = new GenericCommand(RemoveTrackElevation);
            FlattenTrackCommand = new GenericCommand(FlattenTrack);
            EditTrackPointCommand = new GenericCommand(EditTrackPoint);
            EditRoutePointCommand = new GenericCommand(EditRoutePoint);
        }

        private GenericCommand createGeocacheInfo;

        public GenericCommand CreateGeocacheInfo { get { return createGeocacheInfo; } }

        private GenericCommand removeGeocacheInfo;

        public GenericCommand RemoveGeocacheInfo { get { return removeGeocacheInfo; } }

        private GenericCommand viewGeocacheCommand;

        public GenericCommand ViewGeocacheCommand { get { return viewGeocacheCommand; } }

        private GenericCommand openGeocacheCommand;

        public GenericCommand OpenGeocacheCommand { get { return openGeocacheCommand; } }

        public GenericCommand RemoveTrackPointsCommand { get; private set; }

        public GenericCommand RemoveRoutePointsCommand { get; private set; }

        public GenericCommand SimplifyTrackCommand { get; private set; }

        public GenericCommand CreateWaypointCommand { get; private set; }

        public GenericCommand CreateRouteCommand { get; private set; }

        public GenericCommand SortObjectsCommand { get; private set; }

        public GenericCommand MakeVisibleCommand { get; private set; }

        public GenericCommand MakeInvisibleCommand { get; private set; }

        public GenericCommand InvertVisibilityCommand { get; private set; }

        public GenericCommand InvertSelectionCommand { get; private set; }

        public GenericCommand ReverseRouteCommand { get; private set; }

        public GenericCommand ReverseTrackCommand { get; private set; }

        public GenericCommand RemoveTrackTimestampCommand { get; private set; }

        public GenericCommand RemoveTrackElevationCommand { get; private set; }

        public GenericCommand FlattenTrackCommand { get; private set; }

        public GenericCommand EditTrackPointCommand { get; private set; }

        public GenericCommand EditRoutePointCommand { get; private set; }

        #endregion commands
    }
}