namespace Brjnk.GpxExplorer.Workspace
{
    using System.ComponentModel;
    using System.Linq;
    using System.Windows;
    using System.Windows.Data;
    using System.Windows.Media;
    using System.Windows.Media.Imaging;
    using Brjnk.Common;
    using Brjnk.Common.Extensions;
    using Brjnk.CoordinateSystem2D;
    using Brjnk.GpxExplorer.ViewModel;
    using Brjnk.Maps;

    /// <summary>
    /// </summary>
    public class DataView : INotifyPropertyChanged
    {
        private DataModel model;

        private AreaManager area;

        private SearchService searchService;

        private MapManager map;

        private GraphDataAdapter graphDataAdapter;

        private MapDataAdapter mapDataAdapter;

        public DataView(DataModel model, SearchService searchService, AreaManager area, MapManager map)
        {
            this.model = model;
            this.area = area;
            this.searchService = searchService;
            this.map = map;

            WaypointStyle = App.Instance.Settings.WaypointStyle;
            RoutePointStyle = App.Instance.Settings.RoutePointStyle;
            TrackPointStyle = App.Instance.Settings.TrackPointStyle;
            SearchPointStyle = App.Instance.Settings.SearchPointStyle;

            CreateCommands();
            mapMode = true;
            SetMapMode();
        }

        #region Options

        private bool highlightEnabled = true;

        public bool HighlightEnabled
        {
            get { return highlightEnabled; }
            set
            {
                if (graphDataAdapter != null) graphDataAdapter.ColorHighlight = value;
                if (mapDataAdapter != null) mapDataAdapter.ColorHighlight = value;
                highlightEnabled = value;
                OnPropertyChanged("HighlightEnabled");
            }
        }

        private TrackPointAttributeInfo highlightedAttribute = TrackPointAttributeInfo.Elevation;

        public TrackPointAttributeInfo HighlightedAttribute
        {
            get { return highlightedAttribute; }
            set
            {
                System.Diagnostics.Debug.WriteLine(value);
                if (value == null) return;
                if (graphDataAdapter != null) graphDataAdapter.HighlightedAttribute = value;
                if (mapDataAdapter != null) mapDataAdapter.HighlightedAttribute = value;
                highlightedAttribute = value;
                OnPropertyChanged("HighlightedAttribute");
            }
        }

        private double highlightLevel = 0.7;

        public double HighlightLevel
        {
            get { return highlightLevel; }
            set
            {
                if (graphDataAdapter != null) graphDataAdapter.ColorHighlightLevel = value;
                if (mapDataAdapter != null) mapDataAdapter.ColorHighlightLevel = value;
                highlightLevel = value;
                OnPropertyChanged("HighlightLevel");
            }
        }

        private bool includeGaps = true;

        public bool IncludeGaps
        {
            get { return includeGaps; }
            set
            {
                if (graphDataAdapter != null) graphDataAdapter.IncludeGaps = value;
                if (mapDataAdapter != null) mapDataAdapter.IncludeGaps = value;
                includeGaps = value;
                OnPropertyChanged("IncludeGaps");
            }
        }

        private bool fillGaps;

        public bool FillGaps
        {
            get { return fillGaps; }
            set
            {
                if (graphDataAdapter != null) graphDataAdapter.FillGaps = value;
                if (mapDataAdapter != null) mapDataAdapter.FillGaps = value;
                fillGaps = value;
                OnPropertyChanged("FillGaps");
            }
        }

        private bool graphShowPhotos = true;

        public bool GraphShowPhotos
        {
            get { return graphShowPhotos; }
            set
            {
                if (graphDataAdapter != null) graphDataAdapter.ShowPhotos = value;
                graphShowPhotos = value;
                OnPropertyChanged("GraphShowPhotos");
            }
        }

        private bool graphShowWaypoints = true;

        public bool GraphShowWaypoints
        {
            get { return graphShowWaypoints; }
            set
            {
                if (graphDataAdapter != null) graphDataAdapter.ShowWaypoints = value;
                graphShowWaypoints = value;
                OnPropertyChanged("GraphShowWaypoints");
            }
        }

        private bool graphShowSearchResults = true;

        public bool GraphShowSearchResults
        {
            get { return graphShowSearchResults; }
            set
            {
                if (graphDataAdapter != null) graphDataAdapter.ShowSearchResults = value;
                graphShowSearchResults = value;
                OnPropertyChanged("GraphShowSearchResults");
            }
        }

        #endregion Options

        #region Mode

        private bool mapMode;

        public bool MapMode
        {
            get { return mapMode; }
            set
            {
                if (value) { graphMode = false; mapMode = true; SetMapMode(); }
                else { graphMode = true; mapMode = false; SetGraphMode(); }
                OnPropertyChanged("GraphMode");
                OnPropertyChanged("MapMode");
            }
        }

        private bool graphMode;

        public bool GraphMode
        {
            get { return graphMode; }
            set
            {
                if (value) { graphMode = true; mapMode = false; SetGraphMode(); }
                else { graphMode = false; mapMode = true; SetMapMode(); }
                OnPropertyChanged("GraphMode");
                OnPropertyChanged("MapMode");
            }
        }

        #endregion Mode

        #region AxisX

        private TrackPointAttributeInfo graphX = TrackPointAttributeInfo.Distance;

        private bool graphXJoined = true;

        public bool GraphXDistance
        {
            get { return graphX.Attribute == TrackPointAttributeInfo.Distance.Attribute && graphXJoined == false; }
            set { graphX = TrackPointAttributeInfo.Distance; graphXJoined = false; AxisXChanged(); }
        }

        public bool GraphXDistanceJoined
        {
            get { return graphX.Attribute == TrackPointAttributeInfo.Distance.Attribute && graphXJoined; }
            set { graphX = TrackPointAttributeInfo.Distance; graphXJoined = true; AxisXChanged(); }
        }

        public bool GraphXDuration
        {
            get { return graphX.Attribute == TrackPointAttributeInfo.Duration.Attribute && graphXJoined == false; }
            set { graphX = TrackPointAttributeInfo.Duration; graphXJoined = false; AxisXChanged(); }
        }

        public bool GraphXDurationJoined
        {
            get { return graphX.Attribute == TrackPointAttributeInfo.Duration.Attribute && graphXJoined; }
            set { graphX = TrackPointAttributeInfo.Duration; graphXJoined = true; AxisXChanged(); }
        }

        public bool GraphXDateTime
        {
            get { return graphX.Attribute == TrackPointAttributeInfo.DateTime.Attribute; }
            set { graphX = TrackPointAttributeInfo.DateTime; AxisXChanged(); }
        }

        private void AxisXChanged()
        {
            OnPropertyChanged("GraphXDistance");
            OnPropertyChanged("GraphXDistanceJoined");
            OnPropertyChanged("GraphXDuration");
            OnPropertyChanged("GraphXDurationJoined");
            OnPropertyChanged("GraphXDateTime");
            if (graphDataAdapter != null)
            {
                graphDataAdapter.XAttribute = graphX;
                graphDataAdapter.XJoined = graphXJoined;
                CreateAxisX();
                var boundingBox = graphDataAdapter.GetBoundingBox();
                ResetView();
            }
        }

        #endregion AxisX

        #region AxisY

        private TrackPointAttributeInfo graphY = TrackPointAttributeInfo.Elevation;

        public bool GraphYDistance
        {
            get { return graphY.Attribute == TrackPointAttribute.Distance; }
            set
            {
                graphY = TrackPointAttributeInfo.Distance;
                AxisYChanged();
            }
        }

        public bool GraphYDuration
        {
            get { return graphY.Attribute == TrackPointAttribute.Duration; }
            set
            {
                graphY = TrackPointAttributeInfo.Duration;
                AxisYChanged();
            }
        }

        public bool GraphYElevation
        {
            get { return graphY.Attribute == TrackPointAttribute.Elevation; }
            set
            {
                graphY = TrackPointAttributeInfo.Elevation;
                AxisYChanged();
            }
        }

        public bool GraphYSpeed
        {
            get { return graphY.Attribute == TrackPointAttribute.Speed; }
            set
            {
                graphY = TrackPointAttributeInfo.Speed;
                AxisYChanged();
            }
        }

        public bool GraphYVerticalSpeed
        {
            get { return graphY.Attribute == TrackPointAttribute.VerticalSpeed; }
            set
            {
                graphY = TrackPointAttributeInfo.VerticalSpeed;
                AxisYChanged();
            }
        }

        public bool GraphYAcceleration
        {
            get { return graphY.Attribute == TrackPointAttribute.Acceleration; }
            set
            {
                graphY = TrackPointAttributeInfo.Acceleration;
                AxisYChanged();
            }
        }

        public bool GraphYSlope
        {
            get { return graphY.Attribute == TrackPointAttribute.Slope; }
            set { graphY = TrackPointAttributeInfo.Slope; AxisYChanged(); }
        }

        private void AxisYChanged()
        {
            OnPropertyChanged("GraphYDistance");
            OnPropertyChanged("GraphYDuration");
            OnPropertyChanged("GraphYElevation");
            OnPropertyChanged("GraphYSpeed");
            OnPropertyChanged("GraphYVerticalSpeed");
            OnPropertyChanged("GraphYAcceleration");
            OnPropertyChanged("GraphYSlope");
            if (graphDataAdapter != null)
            {
                graphDataAdapter.YAttribute = graphY;
                CreateAxisY();
                ResetView();
            }
        }

        #endregion AxisY

        #region Styles

        private WaypointStyle waypointStyle;

        public WaypointStyle WaypointStyle
        {
            get { return waypointStyle; }
            set { waypointStyle = value; OnPropertyChanged("WaypointStyle"); }
        }

        private RoutePointStyle routePointStyle;

        public RoutePointStyle RoutePointStyle
        {
            get { return routePointStyle; }
            set { routePointStyle = value; OnPropertyChanged("RoutePointStyle"); }
        }

        private TrackPointStyle trackPointStyle;

        public TrackPointStyle TrackPointStyle
        {
            get { return trackPointStyle; }
            set { trackPointStyle = value; OnPropertyChanged("TrackPointStyle"); }
        }

        private SearchPointStyle searchPointStyle;

        public SearchPointStyle SearchPointStyle
        {
            get { return searchPointStyle; }
            set { searchPointStyle = value; OnPropertyChanged("SearchResultPointStyle"); }
        }

        #endregion Styles

        private void SetMapMode()
        {
            area.AxisY = new LinearAxis() { Reverse = true };
            area.AxisX = new LinearAxis() { Reverse = false };

            //Nastav soustavu souadnic do pozice kde se nachz mapa
            //Pokud je pohled do soustavy souadnic velmi vzdlen, trv vykreslen mapy velmi dlouho.
            area.Area.View.SetCenter(new Point(0.5, 0.5));
            area.Area.View.SetZoom(100);
            area.Area.ShowCoordinatesInPopup = false;

            area.Show(area.MapLayer);
            area.Hide(area.GridLayer);

            area.Area.View.SetCenter(new Point(0.5, 0.5));
            area.Area.View.SetZoom(100, 100);

            area.ShowAxisX = false;
            area.ShowAxisY = false;

            if (graphDataAdapter != null) { graphDataAdapter.Detach(true); graphDataAdapter = null; }
            mapDataAdapter = new MapDataAdapter(model, searchService, this, map);
            InitializeAdapter(mapDataAdapter);
            mapDataAdapter.Attach(area);

            var boundingBox = mapDataAdapter.GetBoundingBox();
            SetView(boundingBox, true);
        }

        private void SetView(Bounds? boundingBox, bool isMapMode)
        {
            if (!boundingBox.HasValue)
            {
                area.Area.View.SetCenter(new Point(0.5, 0.5));
                area.Area.View.SetZoom(100, 100);
            }
            else if (boundingBox.Value.IsOnePoint && isMapMode)
            {
                area.Area.View.SetCenter(boundingBox.Value.Min);
                ZoomLevel? level = map.ZoomLevels.First(l => true);
                if (level.HasValue) map.SetZoomLevel(level.Value);
            }
            else
            {
                area.Area.View.SetBounds(boundingBox.Value.Min, boundingBox.Value.Max, isMapMode);
            }
        }

        private void ResetView()
        {
            if (graphDataAdapter != null)
            {
                var boundingBox = graphDataAdapter.GetBoundingBox();
                SetView(boundingBox, false);
            }
            else
            {
                var boundingBox = mapDataAdapter.GetBoundingBox();
                SetView(boundingBox, true);
            }
        }

        private void SetGraphMode()
        {
            area.Hide(area.MapLayer);
            area.Show(area.GridLayer);
            area.Area.ShowCoordinatesInPopup = true;

            CreateAxisX();
            CreateAxisY();

            area.ShowAxisX = true;
            area.ShowAxisY = true;

            if (mapDataAdapter != null) { mapDataAdapter.Detach(true); mapDataAdapter = null; }
            graphDataAdapter = new GraphDataAdapter(model, searchService, this);
            InitializeAdapter(graphDataAdapter);
            graphDataAdapter.Attach(area);

            var boundingBox = graphDataAdapter.GetBoundingBox();
            SetView(boundingBox, false);
        }

        private void InitializeAdapter(DataAdapter adapter)
        {
            adapter.ColorHighlight = this.HighlightEnabled;
            adapter.FillGaps = this.FillGaps;
            adapter.HighlightedAttribute = this.HighlightedAttribute;
            adapter.IncludeGaps = this.IncludeGaps;
            adapter.ColorHighlightLevel = this.HighlightLevel;
            if (adapter is GraphDataAdapter)
            {
                var graphAdapter = adapter as GraphDataAdapter;
                graphAdapter.ShowPhotos = this.GraphShowPhotos;
                graphAdapter.ShowWaypoints = this.graphShowWaypoints;
                graphAdapter.ShowSearchResults = this.GraphShowSearchResults;
                graphAdapter.XAttribute = graphX;
                graphDataAdapter.XJoined = graphXJoined;
                graphAdapter.YAttribute = graphY;
            }
        }

        private void CreateAxisY()
        {
            AxisDefinition axis;
            if (graphY.Attribute == TrackPointAttribute.Duration)
            {
                axis = new DateTimeAxis() { IsDateTime = false };
            }
            else axis = new LinearAxis();

            var caption = Brjnk.GpxExplorer.Localization.LocalizationManager.Instance.GetBinding("graph" + graphY.Attribute.ToString());
            BindingOperations.SetBinding(axis, LinearAxis.CaptionProperty, caption);
            area.AxisY = axis;
        }

        private void CreateAxisX()
        {
            AxisDefinition axis;

            if (graphX.Attribute == TrackPointAttribute.Distance)
            {
                axis = new LinearAxis();
                var caption = Brjnk.GpxExplorer.Localization.LocalizationManager.Instance.GetBinding("graphDistance");
                BindingOperations.SetBinding(axis, LinearAxis.CaptionProperty, caption);
            }
            else if (graphX.Attribute == TrackPointAttribute.Duration)
            {
                axis = new DateTimeAxis() { IsDateTime = false };
                var caption = Brjnk.GpxExplorer.Localization.LocalizationManager.Instance.GetBinding("graphDuration");
                BindingOperations.SetBinding(axis, LinearAxis.CaptionProperty, caption);
            }
            else if (graphX.Attribute == TrackPointAttribute.DateTime)
            {
                axis = new DateTimeAxis() { IsDateTime = true };
                var caption = Brjnk.GpxExplorer.Localization.LocalizationManager.Instance.GetBinding("graphDateTime");
                BindingOperations.SetBinding(axis, LinearAxis.CaptionProperty, caption);
            }
            else
            {
                System.Diagnostics.Debug.WriteLine("Neimplementovan typ osy X");
                axis = new LinearAxis();
            }
            area.AxisX = axis;
        }

        private void ExportBitmap(bool clipboard)
        {
            int pixelWidth = (int)(area.Area.ActualWidth * (300.0 / 96.0));
            int pixelHeight = (int)(area.Area.ActualHeight * (300.0 / 96.0));
            var renderTarget = new System.Windows.Media.Imaging.RenderTargetBitmap(pixelWidth, pixelHeight, 300, 300, PixelFormats.Default);
            renderTarget.Render(GetWhiteBackground());
            renderTarget.Render(area.Area);

            if (clipboard)
                System.Windows.Clipboard.SetImage(renderTarget);
            else
            {
                var sfd = new Microsoft.Win32.SaveFileDialog();
                if ((sfd.ShowDialog() ?? false) == true)
                {
                    var encoder = new PngBitmapEncoder();
                    encoder.Frames.Add(BitmapFrame.Create(renderTarget));
                    using (var stream = new System.IO.FileStream(sfd.FileName, System.IO.FileMode.Create))
                    {
                        encoder.Save(stream);
                    }
                }
            }
        }

        private Visual GetWhiteBackground()
        {
            DrawingVisual visual = new DrawingVisual();
            using (var context = visual.RenderOpen())
            {
                context.DrawRectangle(Brushes.White, null, new Rect(0, 0, area.Area.ActualWidth, area.Area.ActualHeight));
            }
            return visual;
        }

        private void UnselectTrackPoints(TrackVM track)
        {
            if (track == null) return;
            foreach (var point in track.GetAllPoints())
            {
                point.IsSelected = false;
            }
        }

        private void SelectElevationExtremeTrackPoints(TrackVM track)
        {
            if (track == null) return;
            foreach (var point in track.GetAllPoints())
            {
                point.IsSelected = point.ElevationLocalExtreme != ElevationExtreme.None;
            }
        }

        #region commands

        private void CreateCommands()
        {
            showAllWaypointsCommand = new GenericCommand(o => model.Waypoints.ForEach(i => i.IsVisible = true));
            showAllRoutesCommand = new GenericCommand(o => model.Routes.ForEach(i => i.IsVisible = true));
            showAllTracksCommand = new GenericCommand(o => model.Tracks.ForEach(i => i.IsVisible = true));
            showAllGCsCommand = new GenericCommand(o => model.Waypoints.Where(i => i.Source.Geocache != null).ForEach(i => i.IsVisible = true));

            hideAllWaypointsCommand = new GenericCommand(o => model.Waypoints.ForEach(i => i.IsVisible = false));
            hideAllRoutesCommand = new GenericCommand(o => model.Routes.ForEach(i => i.IsVisible = false));
            hideAllTracksCommand = new GenericCommand(o => model.Tracks.ForEach(i => i.IsVisible = false));
            hideAllGCsCommand = new GenericCommand(o => model.Waypoints.Where(i => i.Source.Geocache != null).ForEach(i => i.IsVisible = false));

            exportBitmapCommand = new GenericCommand(o => ExportBitmap(false));
            exportBitmapToClipboardCommand = new GenericCommand(o => ExportBitmap(true));

            resetViewCommand = new GenericCommand(o => ResetView());

            UnselectAllTrackPointsCommand = new GenericCommand(o => UnselectTrackPoints(o as TrackVM));
            SelectElevationExtremeTrackPointsCommand = new GenericCommand(o => SelectElevationExtremeTrackPoints(o as TrackVM));
        }

        private GenericCommand showAllWaypointsCommand;

        public GenericCommand ShowAllWaypointsCommand { get { return showAllWaypointsCommand; } }

        private GenericCommand showAllRoutesCommand;

        public GenericCommand ShowAllRoutesCommand { get { return showAllRoutesCommand; } }

        private GenericCommand showAllTracksCommand;

        public GenericCommand ShowAllTracksCommand { get { return showAllTracksCommand; } }

        private GenericCommand showAllGCsCommand;

        public GenericCommand ShowAllGCsCommand { get { return showAllGCsCommand; } }

        private GenericCommand hideAllWaypointsCommand;

        public GenericCommand HideAllWaypointsCommand { get { return hideAllWaypointsCommand; } }

        private GenericCommand hideAllRoutesCommand;

        public GenericCommand HideAllRoutesCommand { get { return hideAllRoutesCommand; } }

        private GenericCommand hideAllTracksCommand;

        public GenericCommand HideAllTracksCommand { get { return hideAllTracksCommand; } }

        private GenericCommand hideAllGCsCommand;

        public GenericCommand HideAllGCsCommand { get { return hideAllGCsCommand; } }

        private GenericCommand resetViewCommand;

        public GenericCommand ResetViewCommand { get { return resetViewCommand; } }

        private GenericCommand exportBitmapCommand;

        public GenericCommand ExportBitmapCommand { get { return exportBitmapCommand; } }

        private GenericCommand exportBitmapToClipboardCommand;

        public GenericCommand ExportBitmapToClipboardCommand { get { return exportBitmapToClipboardCommand; } }

        public GenericCommand UnselectAllTrackPointsCommand { get; private set; }

        public GenericCommand SelectElevationExtremeTrackPointsCommand { get; private set; }

        #endregion commands

        public event PropertyChangedEventHandler PropertyChanged;

        private void OnPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }
    }
}