namespace Brjnk.GpxExplorer.Workspace
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.ComponentModel;
    using System.Linq;
    using System.Windows;
    using System.Windows.Data;
    using System.Windows.Markup;
    using Brjnk.Common;
    using Brjnk.CoordinateSystem2D;
    using Brjnk.GpxExplorer.UI;
    using Brjnk.Maps;
    using Brjnk.Maps.Cache;
    using GpxExplorer;

    /// <summary>
    /// </summary>
    public class MapManager : INotifyPropertyChanged
    {
        private AreaManager area;

        private UsageLogger usageLogger = new UsageLogger();

        private IMapCache cache;

        private ViewInfoBase view;

        public MapManager(AreaManager area, ViewInfoBase view)
        {
            this.area = area;
            this.view = view;
            cache = new MapStorage(usageLogger);
            CreateSources();
            CreateCommands();
        }

        private void CreateSources()
        {
            //builtInSources = new ObservableCollection<IMapSource>();
            allSources = new System.Windows.Data.CompositeCollection();
            var builtInMaps = App.Instance.FindResource("MapSources") as IEnumerable;
            var customMaps = App.Instance.Settings.CustomMapSources;
            string mapName = App.Instance.Settings.SelectedMapSource;
            allSources.Add(new CollectionContainer() { Collection = builtInMaps });
            allSources.Add(new CollectionContainer() { Collection = customMaps });

            Source = builtInMaps.OfType<IMapSource>().Where(m => m.UniqueName == mapName).FirstOrDefault();
            if (Source == null) Source = customMaps.Where(m => m.UniqueName == mapName).FirstOrDefault();
            if (Source == null) Source = (IMapSource)builtInMaps.OfType<IMapSource>().First();
        }

        private void CreateProvider()
        {
            if (Source == null)
            {
                Provider = null; zoomLevels = null;
            }
            else
            {
                if (provider != null) provider.Abort();
                if (Source is WebMapSource)
                {
                    var webMapSource = Source as WebMapSource;
                    Provider = new WebMapProvider(cache, webMapSource, usageLogger);
                    ZoomLevels = webMapSource.ZoomLevels;
                }
                else
                {
                    System.Diagnostics.Debug.Fail("Neznm typ mapovho zdroje");
                }
            }
        }

        private void OpenInWebbrowser(object uri)
        {
            var uriString = uri.ToString();
            if (uriString.Contains("{0") && uriString.Contains("{1"))
            {
                uriString = string.Format(System.Globalization.CultureInfo.InvariantCulture, uriString, area.MapLayer.GpsCoordinates.LonTotalDegree, area.MapLayer.GpsCoordinates.LatTotalDegree);
                System.Diagnostics.Process.Start(uriString);
            }
        }

        public void SetZoomLevel(ZoomLevel level)
        {
            view.SetZoom(source.TileDIPSize.Width / level.InternalTileSize.Width, source.TileDIPSize.Height / level.InternalTileSize.Height);
        }

        private void SetZoomLevel(object level)
        {
            if (level == null || !(level is ZoomLevel) || source == null) return;
            var l = (ZoomLevel)level;
            SetZoomLevel(l);
        }

        private void AddCustomSource(object arg)
        {
            CustomSources.Add(new CustomWebMapSource()
            {
                Type = CustomWebMapType.MercatorOSM,
                MaxZoom = 18,
                SourceNameSerializable = "New map source " + DateTime.Now.ToString(),
                TileDIPSizeSerializable = new Size(256, 256)
            });
        }

        private void RemoveCustomSource(object source)
        {
            if (Source == source) Source = (allSources[0] as CollectionContainer).Collection.OfType<IMapSource>().FirstOrDefault();
            CustomSources.Remove(source as CustomWebMapSource);
        }

        private void ManageSources(object arg)
        {
            var window = new ManageMapSourcesWindow(this);
            window.ShowDialog();
        }

        private void ShowStatistics(object arg)
        {
            var window = new MapStatisticsWindow(usageLogger);
            window.Show();
        }

        private System.Windows.Data.CompositeCollection allSources;

        public IEnumerable AllSources
        {
            get { return allSources; }
        }

        public ObservableCollection<CustomWebMapSource> CustomSources
        {
            get { return App.Instance.Settings.CustomMapSources; }
        }

        private IEnumerable<ZoomLevel> zoomLevels;

        public IEnumerable<ZoomLevel> ZoomLevels
        {
            get { return zoomLevels; }
            set { zoomLevels = value; OnPropertyChanged("ZoomLevels"); }
        }

        private IMapSource source;

        public IMapSource Source
        {
            get { return source; }
            set
            {
                if (source is CustomWebMapSource)
                {
                    var customSource = (source as CustomWebMapSource);
                    customSource.Changed -= new EventHandler(MapManager_Changed);
                    customSource.PropertyChanged -= new PropertyChangedEventHandler(customSource_PropertyChanged);
                }
                source = value;
                if (source is CustomWebMapSource)
                {
                    var customSource = (source as CustomWebMapSource);
                    customSource.Changed += new EventHandler(MapManager_Changed);
                    customSource.PropertyChanged += new PropertyChangedEventHandler(customSource_PropertyChanged);
                    IsSourceValid = !customSource.IsInvalid;
                }
                else
                    IsSourceValid = true;

                CreateProvider();
                App.Instance.Settings.SelectedMapSource = source == null ? string.Empty : source.UniqueName;
                OnPropertyChanged("Source");
            }
        }

        private void customSource_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == "IsInvalid")
            {
                var customSource = (source as CustomWebMapSource);

                //pokud dolo ke zmn, znovuvytvoit provider
                if (IsSourceValid == customSource.IsInvalid) CreateProvider();
                IsSourceValid = !customSource.IsInvalid;
            }
        }

        private void MapManager_Changed(object sender, EventArgs e)
        {
            CreateProvider();
        }

        private bool isSourceValid;

        public bool IsSourceValid
        {
            get { return isSourceValid; }
            private set { isSourceValid = value; OnPropertyChanged("IsSourceValid"); }
        }

        private IMapProvider provider;

        public IMapProvider Provider
        {
            get { return provider; }
            private set { provider = value; area.MapLayer.Map = value; OnPropertyChanged("Provider"); }
        }

        private double angle;

        public double Angle
        {
            get { return angle; }
            set
            {
                double delta = value - angle;
                angle = value;
                area.MapLayer.Rotation = value;
                RotateViewPosition(delta);
                OnPropertyChanged("Angle");
            }
        }

        private void RotateViewPosition(double delta)
        {
            view.SetCenter(view.TransformBack(MathEx.RotatePointD(view.CenterInternal, -delta)));
        }

        /// <summary>
        /// Promtne gps souadnice do internch 2D souadnice (vetn zohlednn rotace mapy)
        /// </summary>
        public Point Project(GpsCoordinates gpsCoordinates)
        {
            return MathEx.RotatePointD(Provider.Projection.ToInternal(gpsCoordinates), Angle);
        }

        public event PropertyChangedEventHandler PropertyChanged;

        private void OnPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }

        #region commands

        private void CreateCommands()
        {
            openInWebbrowserCommand = new GenericCommand(OpenInWebbrowser);
            setZoomLevelCommand = new GenericCommand(SetZoomLevel);
            manageSourcesCommand = new GenericCommand(ManageSources);
            addCustomSourceCommand = new GenericCommand(AddCustomSource);
            removeCustomSourceCommand = new GenericCommand(RemoveCustomSource);
            showStatisticsCommnad = new GenericCommand(ShowStatistics);
        }

        private GenericCommand openInWebbrowserCommand;

        public GenericCommand OpenInWebbrowserCommand { get { return openInWebbrowserCommand; } }

        private GenericCommand setZoomLevelCommand;

        public GenericCommand SetZoomLevelCommand { get { return setZoomLevelCommand; } }

        private GenericCommand manageSourcesCommand;

        public GenericCommand ManageSourcesCommand { get { return manageSourcesCommand; } }

        private GenericCommand addCustomSourceCommand;

        public GenericCommand AddCustomSourceCommand { get { return addCustomSourceCommand; } }

        private GenericCommand removeCustomSourceCommand;

        public GenericCommand RemoveCustomSourceCommand { get { return removeCustomSourceCommand; } }

        private GenericCommand showStatisticsCommnad;

        public GenericCommand ShowStatisticsCommand { get { return showStatisticsCommnad; } }

        #endregion commands
    }

    public class WebbrowserMap
    {
        public WebbrowserMap(string name, string address, Uri iconUri)
        {
            Name = name;
            IconUri = iconUri;
            Address = address;
        }

        public WebbrowserMap() { }

        [ConstructorArgument("name")]
        public string Name { get; set; }

        [ConstructorArgument("iconUri")]
        public Uri IconUri { get; set; }

        [ConstructorArgument("address")]
        public string Address { get; set; }
    }
}