namespace Brjnk.GpxExplorer.Workspace
{
    using System;
    using System.ComponentModel;
    using System.Linq;
    using Brjnk.Common;
    using Brjnk.Common.Extensions;
    using Brjnk.GpxExplorer.Computing;
    using Brjnk.GpxExplorer.UI;
    using Brjnk.GpxExplorer.ViewModel;
    using Brjnk.MVVM;

    /// <summary>
    /// </summary>
    public class SearchService
    {
        private DataModel model;

        public SearchService(DataModel model)
        {
            this.model = model;
            searchedItems.ItemsAdded += new EventHandler<ItemsEventArgs<SearchRequest>>(searchedItems_ItemsAdded);
            searchedItems.ItemsRemoved += new EventHandler<ItemsEventArgs<SearchRequest>>(searchedItems_ItemsRemoved);
            Configuration = App.Instance.Settings.SearchConfiguration;
            SearchAll();
            AddHandlers(model);

            CreateCommands();
        }

        #region EventHandlers

        private void AddHandlers(DataModel model)
        {
            model.Watcher.RouteDataChanged += new EventHandler<WatcherEventArgs<RouteVM>>(Watcher_RouteDataChanged);
            model.Watcher.TrackDataChanged += new EventHandler<WatcherEventArgs<TrackVM>>(Watcher_TrackDataChanged);
            model.Watcher.WaypointDataChanged += new EventHandler<WatcherEventArgs<WaypointVM>>(Watcher_WaypointDataChanged);

            model.Watcher.WaypointsReseted += new EventHandler(Watcher_WaypointsReseted);
            model.Watcher.RoutesReseted += new EventHandler(Watcher_RoutesReseted);
            model.Watcher.TracksReseted += new EventHandler(Watcher_TracksReseted);
            model.Watcher.PhotosReseted += new EventHandler(Watcher_PhotosReseted);

            model.Watcher.ObjectAdded += new EventHandler<WatcherEventArgs<object>>(Watcher_ObjectAdded);
            model.Watcher.ObjectRemoved += new EventHandler<WatcherEventArgs<object>>(Watcher_ObjectRemoved);
        }

        private void Watcher_ObjectRemoved(object sender, WatcherEventArgs<object> e)
        {
            if (e.ObjectType == typeof(WaypointVM))
                RemoveBySearchedItem(e.Object as WaypointVM);
            else if (e.ObjectType == typeof(RouteVM))
                RemoveRoute(e.Object as RouteVM);
            else if (e.ObjectType == typeof(TrackVM))
                RemoveTrack(e.Object as TrackVM);
            else if (e.ObjectType == typeof(LocatedPhoto))
                RemoveBySearchedItem(e.Object as LocatedPhoto);
            else if (e.ObjectType == typeof(Photo))
            {
                foreach (var located in ((Photo)e.Object).Locations)
                    RemoveBySearchedItem(located);
            }
        }

        private void Watcher_ObjectAdded(object sender, WatcherEventArgs<object> e)
        {
            if (e.ObjectType == typeof(WaypointVM))
                ProcessPoint(e.Object as WaypointVM, waypointSearchArgs);
            else if (e.ObjectType == typeof(RouteVM))
                ProcessPath(e.Object as RouteVM);
            else if (e.ObjectType == typeof(TrackVM))
                ProcessPath(e.Object as TrackVM);
            else if (e.ObjectType == typeof(LocatedPhoto))
                ProcessPoint(e.Object as LocatedPhoto, photoSearchArgs);
        }

        private void Watcher_TrackDataChanged(object sender, WatcherEventArgs<TrackVM> e)
        {
            RemoveTrack(e.Object);
            ProcessPath(e.Object);
        }

        private void Watcher_RouteDataChanged(object sender, WatcherEventArgs<RouteVM> e)
        {
            RemoveRoute(e.Object);
            ProcessPath(e.Object);
        }

        private void Watcher_WaypointDataChanged(object sender, WatcherEventArgs<WaypointVM> e)
        {
            RemoveBySearchedItem(e.Object);
            ProcessPoint(e.Object, photoSearchArgs);
        }

        private void Watcher_PhotosReseted(object sender, EventArgs e)
        {
            RemoveType(typeof(LocatedPhoto));
            SearchAllPhotos();
        }

        private void Watcher_TracksReseted(object sender, EventArgs e)
        {
            RemoveType(typeof(TrackVM));
            SearchAllTracks();
        }

        private void Watcher_RoutesReseted(object sender, EventArgs e)
        {
            RemoveType(typeof(RouteVM));
            SearchAllRoutes();
        }

        private void Watcher_WaypointsReseted(object sender, EventArgs e)
        {
            RemoveType(typeof(WaypointVM));
            SearchAllWaypoints();
        }

        private void searchedItems_ItemsAdded(object sender, ItemsEventArgs<SearchRequest> e)
        {
            foreach (var item in e.Items)
            {
                item.Changed += new EventHandler(item_Changed);
                ProcessSearchRequest(item);
            }
        }

        private void searchedItems_ItemsRemoved(object sender, ItemsEventArgs<SearchRequest> e)
        {
            foreach (var item in e.Items)
            {
                item.Changed -= new EventHandler(item_Changed);
                RemoveBySearchedItem(item);
            }
        }

        private void item_Changed(object sender, EventArgs e)
        {
            var request = sender as SearchRequest;
            RemoveBySearchedItem(request);
            ProcessSearchRequest(request);
        }

        #endregion EventHandlers

        #region RemovingResults

        private void RemoveRoutesTracks(Predicate<SearchResult> predicate)
        {
            foreach (var rte in model.Routes)
            {
                foreach (var rtePt in rte.Points)
                {
                    if (rtePt.SearchResults != null)
                    {
                        rtePt.SearchResults.RemoveWhere(predicate);
                        if (rtePt.SearchResults.Count == 0) rtePt.SearchResults = null;
                    }
                }
            }

            foreach (var trk in model.Tracks)
            {
                foreach (Brjnk.GpxExplorer.ViewModel.TrackPointVM trkPt in trk.GetAllPoints())
                {
                    if (trkPt.SearchResults != null)
                    {
                        trkPt.SearchResults.RemoveWhere(predicate);
                        if (trkPt.SearchResults.Count == 0) trkPt.SearchResults = null;
                    }
                }
            }

            results.RemoveWhere(predicate);
        }

        private void RemovePointsPhotosItems(Predicate<SearchResult> predicate)
        {
            foreach (var pht in model.Photos.Collection)
            {
                foreach (var locatedPhto in pht.Locations)
                {
                    if (locatedPhto.SearchResults != null)
                    {
                        locatedPhto.SearchResults.RemoveWhere(predicate);
                        if (locatedPhto.SearchResults.Count == 0) locatedPhto.SearchResults = null;
                    }
                }
            }

            foreach (var wpt in model.Waypoints)
            {
                if (wpt.SearchResults != null)
                {
                    wpt.SearchResults.RemoveWhere(predicate);
                    if (wpt.SearchResults.Count == 0) wpt.SearchResults = null;
                }
            }

            foreach (var item in searchedItems)
            {
                if (item.SearchResults != null)
                {
                    item.SearchResults.RemoveWhere(predicate);
                    if (item.SearchResults.Count == 0) item.SearchResults = null;
                }
            }
            results.RemoveWhere(predicate);
        }

        private void RemoveAll()
        {
            results.Clear();
            RemovePointsPhotosItems(r => true);
            RemoveRoutesTracks(r => true);
        }

        private void RemoveType(Type t)
        {
            Predicate<SearchResult> predicate;
            if (t == typeof(WaypointVM) || t == typeof(LocatedPhoto) || typeof(SearchRequest).IsAssignableFrom(t))
            {
                predicate = (r => t.IsAssignableFrom(r.SearchedItem.GetType()));
            }
            else if (t == typeof(RouteVM) || t == typeof(TrackVM))
                predicate = (r => t.IsAssignableFrom(r.FoundIn.GetType()));
            else
            {
                throw new Exception("Mon chyba");
            }

            RemovePointsPhotosItems(predicate);
            RemoveRoutesTracks(predicate);
            results.RemoveWhere(predicate);
        }

        private void RemoveBySearchedItem(ISearchResults item)
        {
            if (!(item is WaypointVM || item is LocatedPhoto || item is SearchRequest))
                throw new Exception(string.Format("Do metody byl pedn argument typu {0}.", item.GetType().ToString()));
            item.SearchResults = null;
            RemoveRoutesTracks(r => object.ReferenceEquals(r.SearchedItem, item));
            results.RemoveWhere(r => object.ReferenceEquals(r.SearchedItem, item));
        }

        private void RemoveTrack(TrackVM track)
        {
            foreach (TrackPointVM trkPt in track.GetAllPoints()) trkPt.SearchResults = null;
            RemovePointsPhotosItems(r => object.ReferenceEquals(r.FoundIn, track));
            results.RemoveWhere(r => object.ReferenceEquals(r.FoundIn, track));
        }

        private void RemoveRoute(RouteVM route)
        {
            foreach (var rtePt in route.Points) rtePt.SearchResults = null;
            RemovePointsPhotosItems(r => object.ReferenceEquals(r.FoundIn, route));
            results.RemoveWhere(r => object.ReferenceEquals(r.FoundIn, route));
        }

        #endregion RemovingResults

        #region CreatingResults

        private void SearchAll()
        {
            SearchAllRoutes();
            SearchAllTracks();
        }

        private void SearchAllWaypoints()
        {
            foreach (var wp in model.Waypoints) ProcessPoint(wp, waypointSearchArgs);
        }

        private void SearchAllRoutes()
        {
            foreach (var rte in model.Routes) ProcessPath(rte);
        }

        private void SearchAllTracks()
        {
            foreach (var trk in model.Tracks) ProcessPath(trk);
        }

        private void SearchAllPhotos()
        {
            foreach (var pht in model.Photos.Collection)
            {
                foreach (var locatedPhto in pht.Locations) ProcessPoint(locatedPhto, photoSearchArgs);
            }
        }

        private void SearchAllSearchedItems()
        {
            foreach (var item in searchedItems) ProcessSearchRequest(item);
        }

        private void Process(SearchDateTimeRequest request, TrackVM track)
        {
            var r = track.FindTime(request.DateTime, request.SearchInGaps);
            if (r == null) return;
            var result = new SearchDateTimeResult(r, request, track);
            AddResult(request, result);
            results.Add(result);
            AddResult(r.PreviousPoint as TrackPointVM, result);
        }

        private void Process(IPoint point, ISearchablePath path, FindPositionArgs args)
        {
            var r = path.FindPosition(point, args);
            foreach (var i in r)
            {
                var result = new SearchPositionResult(i, point, path);
                AddResult(point as ISearchResults, result);
                results.Add(result);
                AddResult(i.PreviousPoint as ISearchResults, result);
            }
        }

        private void AddResult(ISearchResults target, SearchResult result)
        {
            if (target.SearchResults == null) target.SearchResults = new ObservableCollectionEx<SearchResult>();
            target.SearchResults.Add(result);
        }

        private void ProcessSearchRequest(SearchRequest item)
        {
            if (item is SearchPositionRequest)
            {
                var request = item as SearchPositionRequest;
                ProcessPoint(request, request.GetArgs());
            }
            else if (item is SearchDateTimeRequest)
                ProcessDateTimeSearch(item as SearchDateTimeRequest);
            else throw new InvalidOperationException("Neznm typ");
        }

        private void ProcessPoint(IPoint point, FindPositionArgs args)
        {
            foreach (var path in model.Tracks)
                Process(point, path, args);

            foreach (var path in model.Routes)
                Process(point, path, args);
        }

        private void ProcessDateTimeSearch(SearchDateTimeRequest request)
        {
            foreach (var path in model.Tracks)
                Process(request, path);
        }

        private void ProcessPath(ISearchablePath path)
        {
            foreach (var wpt in model.Waypoints)
            {
                Process(wpt, path, waypointSearchArgs);
            }

            foreach (var item in searchedItems.Where(r => r is SearchPositionRequest))
            {
                var request = item as SearchPositionRequest;
                var args = request.GetArgs();
                Process(request, path, args);
            }

            foreach (var photo in model.Photos.Collection)
            {
                foreach (var locatedPhoto in photo.Locations)
                    Process(locatedPhoto, path, photoSearchArgs);
            }

            if (path is TrackVM)
            {
                foreach (var item in searchedItems.Where(r => r is SearchDateTimeRequest))
                {
                    Process(item as SearchDateTimeRequest, path as TrackVM);
                }
            }
        }

        #endregion CreatingResults

        #region Properties

        private ObservableCollectionEx<SearchRequest> searchedItems = new ObservableCollectionEx<SearchRequest>();

        public ObservableCollectionEx<SearchRequest> SearchedItems
        {
            get { return searchedItems; }
            set { searchedItems = value; OnPropertyChanged("SearchedItems"); }
        }

        private ObservableCollectionEx<SearchResult> results = new ObservableCollectionEx<SearchResult>();

        public ObservableCollectionEx<SearchResult> Results
        {
            get { return results; }
        }

        private bool isDisabled;

        public bool IsDisabled
        {
            get { return isDisabled; }
            set
            {
                if (value == isDisabled) return;

                isDisabled = value;
                if (isDisabled) RemoveAll();
                else SearchAll();
                OnPropertyChanged("IsDisabled");
            }
        }

        private SearchConfiguration configuration;

        public SearchConfiguration Configuration
        {
            get { return configuration; }
            set
            {
                configuration = value;
                if (configuration == null) throw new ArgumentNullException();
                //zamezit zmnm objektu, protoe nen hldn pes INotifyPropertyChanged
                configuration.Seal();
                //uloit do konfiguranho souboru
                App.Instance.Settings.SearchConfiguration = configuration;
                if (this.IsDisabled) return;

                //aktualizovat podle novho nastaven
                RemoveType(typeof(WaypointVM));
                RemoveType(typeof(LocatedPhoto));
                photoSearchArgs = configuration.GetPhotoSearchArgs();
                waypointSearchArgs = configuration.GetWaypointSearchArgs();
                SearchAllWaypoints(); SearchAllPhotos();

                OnPropertyChanged("Configuration");
            }
        }

        private FindPositionArgs photoSearchArgs;
        private FindPositionArgs waypointSearchArgs;

        #endregion Properties

        #region Commands

        private void CreateCommands()
        {
            searchPositionCommand = new GenericCommand(SearchPosition);
            searchDateTimeCommand = new GenericCommand(SearchDateTime);
            searchPlaceCommand = new GenericCommand(SearchPlace);
            showSearchListCommand = new GenericCommand(ShowSearchList);
            showSearchConfigurationCommand = new GenericCommand(ShowSearchConfiguration);
            removeItemCommand = new GenericCommand(RemoveItem);
            showDebugInfoCommand = new GenericCommand(ShowDebugInfo);
        }

        private void SearchPosition(object arg)
        {
            var window = new SearchPositionWindow(this);
            window.ShowDialog();
        }

        private void SearchDateTime(object arg)
        {
            var window = new SearchTimeWindow(this);
            window.ShowDialog();
        }

        private void SearchPlace(object arg)
        {
            var window = new SearchPlaceWindow();
            window.ShowDialog();
        }

        private void ShowSearchList(object arg)
        {
            var window = new SearchListWindow(this);
            window.ShowDialog();
        }

        private void ShowSearchConfiguration(object arg)
        {
            var window = new SearchConfigurationWindow(this);
            window.ShowDialog();
        }

        private void RemoveItem(object arg)
        {
            var item = arg as SearchRequest;
            SearchedItems.Remove(item);
        }

        private void ShowDebugInfo(object arg)
        {
            var window = new ObjectViewer();
            window.Object = this;
            window.ShowDialog();
        }

        private GenericCommand searchPositionCommand;

        public GenericCommand SearchPositionCommand { get { return searchPositionCommand; } }

        private GenericCommand searchDateTimeCommand;

        public GenericCommand SearchDateTimeCommand { get { return searchDateTimeCommand; } }

        private GenericCommand searchPlaceCommand;

        public GenericCommand SearchPlaceCommand { get { return searchPlaceCommand; } }

        private GenericCommand showSearchListCommand;

        public GenericCommand ShowSearchListCommand { get { return showSearchListCommand; } }

        private GenericCommand showSearchConfigurationCommand;

        public GenericCommand ShowSearchConfigurationCommand { get { return showSearchConfigurationCommand; } }

        private GenericCommand removeItemCommand;

        public GenericCommand RemoveItemCommand { get { return removeItemCommand; } }

        private GenericCommand showDebugInfoCommand;

        public GenericCommand ShowDebugInfoCommand { get { return showDebugInfoCommand; } }

        #endregion Commands

        public event PropertyChangedEventHandler PropertyChanged;

        protected void OnPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }
    }
}