using System.Collections.Generic;
using System.ComponentModel;
using Brjnk.Common;
using Brjnk.CoordinateSystem2D;
using Brjnk.CoordinateSystem2D.Layers;
using Brjnk.GpxExplorer.Localization;
using Brjnk.GpxExplorer.UI;

namespace Brjnk.GpxExplorer.Workspace
{
    /// <summary>
    /// </summary>
    public class Workspace : INotifyPropertyChanged
    {
        public Workspace()
        {
            CreateCommands();
        }

        public void Initialize()
        {
            Data = new DataModel();

            Editor = new DataEditor(Data);

            AreaManager = new AreaManager(FindResource<Area2D>("Area2D"),
                FindResource<PointsLayer>("PointsLayer"),
                FindResource<GraphLayer>("RoutesLayer"),
                FindResource<GraphLayer>("TracksLayer"),
                FindResource<PhotosLayer>("PhotosLayer"),
                FindResource<MapLayer>("MapLayer"),
               FindResource<Grid>("GridLayer"),
              FindResource<PointsLayer>("RoutePointsLayer"),
              FindResource<PointsLayer>("TrackPointsLayer"),
              FindResource<PointsLayer>("SearchResultsLayer")
               );

            areaManager.SetEditor(Editor);

            DataManager = new DataManager(Data, AreaManager);

            ViewManager = new ViewManager(AreaManager.Area.View);
            MapManager = new MapManager(AreaManager, AreaManager.Area.View);

            SearchService = new SearchService(Data);

            DataView = new DataView(Data, SearchService, AreaManager, MapManager);

            Clipboard = new GpxExplorer.Workspace.Clipboard(App.Instance.ClipboardWatcher, Data, SearchService);

            VersionInfo = new GpxExplorer.Workspace.VersionInfo();
        }

        private void ShowDebugWindow(object arg)
        {
#if DEBUG

            var window = new ObjectViewer();
            window.Object = App.Instance;
            window.ShowDialog();
#endif
        }

        private void ShowAboutWindow(object arg)
        {
            var window = new AboutWindow(VersionInfo);
            window.ShowDialog();
        }

        private void ShowSettings(object arg)
        {
            var window = new SettingsWindow();
            window.ShowDialog();
        }

        private T FindResource<T>(object key) where T : class
        {
            var res = System.Windows.Application.Current.FindResource(key);
            if (typeof(T).IsAssignableFrom(res.GetType())) return res as T;
            System.Diagnostics.Debug.WriteLine("Nepodailo se vrtit Resource. Nalezen objekt nen poadovanho typu. Kl: " + key.ToString());
            return null;
        }

        #region Objects

        private DataModel data;

        public DataModel Data
        {
            get { return data; }
            private set { data = value; OnPropertyChanged("Data"); }
        }

        private AreaManager areaManager;

        public AreaManager AreaManager
        {
            get { return areaManager; }
            private set { areaManager = value; OnPropertyChanged("AreaManager"); }
        }

        private DataManager dataManager;

        public DataManager DataManager
        {
            get { return dataManager; }
            private set { dataManager = value; OnPropertyChanged("DataManager"); }
        }

        private ViewManager viewManager;

        public ViewManager ViewManager
        {
            get { return viewManager; }
            private set { viewManager = value; OnPropertyChanged("ViewManager"); }
        }

        private MapManager mapManager;

        public MapManager MapManager
        {
            get { return mapManager; }
            private set { mapManager = value; OnPropertyChanged("MapManager"); }
        }

        private DataView dataView;

        public DataView DataView
        {
            get { return dataView; }
            private set { dataView = value; OnPropertyChanged("DataView"); }
        }

        private DataEditor editor;

        public DataEditor Editor
        {
            get { return editor; }
            private set { editor = value; OnPropertyChanged("Editor"); }
        }

        private SearchService searchService;

        public SearchService SearchService
        {
            get { return searchService; }
            private set { searchService = value; OnPropertyChanged("SearchService"); }
        }

        private VersionInfo versionInfo;

        public VersionInfo VersionInfo
        {
            get { return versionInfo; }
            private set { versionInfo = value; OnPropertyChanged("VersionInfo"); }
        }

        private Clipboard clipboard;

        public Clipboard Clipboard
        {
            get { return clipboard; }
            set { clipboard = value; OnPropertyChanged("Clipboard"); }
        }

        public IEnumerable<IsoLanguageName> Languages
        {
            get { return Localization.LocalizationManager.Instance.AvailableLanguages; }
        }

        #endregion Objects

        #region Commands

        public GenericCommand ShowDebugWindowCommand { get; private set; }

        public GenericCommand ShowAboutWindowCommand { get; private set; }

        public GenericCommand ShowSettingsCommand { get; private set; }

        private void CreateCommands()
        {
            ShowDebugWindowCommand = new GenericCommand(ShowDebugWindow);
            ShowAboutWindowCommand = new GenericCommand(ShowAboutWindow);
            ShowSettingsCommand = new GenericCommand(ShowSettings);
        }

        #endregion Commands

        private object selectedItem;

        public object SelectedItem
        {
            get { return selectedItem; }
            set { selectedItem = value; OnPropertyChanged("SelectedItem"); }
        }

        public Settings Settings { get { return App.Instance.Settings; } }

        public event PropertyChangedEventHandler PropertyChanged;

        private void OnPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }
    }
}