namespace Brjnk.Maps.Cache
{
    using System;
    using System.Configuration;
    using System.Diagnostics;
    using System.Linq;
    using System.Windows.Media.Imaging;

    /// <summary>
    /// Tda je Thread-Safe, protoe WebMapDownloader pouv instanci tto tdy v jinm vlkn.
    /// </summary>
    public class MapStorage : IMapCache
    {
        private const string cacheName = "DB Cache";

        private MapStorageContext context;
        private IDbUsageLogger logger;
        private object lockObject = new Object();

        private long actualSize = 0; private long maxSize = 0;

        /// <summary>
        /// Vchoz doba platnosti dladice (poet dn)
        /// </summary>
        private int defaultExpirationPeriodDays;

        public MapStorage(IDbUsageLogger logger) : this(logger, new MapStorageContext()) { }

        public MapStorage(IDbUsageLogger logger, MapStorageContext context)
        {
            ReadConfig();
            this.context = context;
            this.logger = logger;
            PurgeDb();
            LogCacheState();
        }

        private void ReadConfig()
        {
            long.TryParse(ConfigurationManager.AppSettings["dbMaxTotalSize"], out maxSize);
            int.TryParse(ConfigurationManager.AppSettings["dbExpirationPeriodDays"], out defaultExpirationPeriodDays);
            if (maxSize == 0) maxSize = 1000 * 1024 * 1024;
            if (defaultExpirationPeriodDays == 0) defaultExpirationPeriodDays = 120;
            Debug.WriteLine(maxSize, "DB - maximln velikost databze (B)");
            Debug.WriteLine(defaultExpirationPeriodDays, "DB - vchoz platnost obrzku v cache pamti (dny)");
        }

        /// <summary>
        /// Vyist nepotebn obrzky v databzi. Vrt True, pokud byla velikost obrzk snena pod limitn hodnoty.
        /// </summary>
        private bool PurgeDb()
        {
            FindOutDbSize();
            long sizeToClear = actualSize - maxSize;

            while (sizeToClear > 0)
            {
                const int maxCount = 100;
                var toDelete = context.Tiles.OrderBy(t => t.DownloadedTime).Take(maxCount).Select(tile => new { tile.X, tile.Y, tile.Zoom, tile.SourceId }).ToList();
                int count = 0;
                foreach (var tile in toDelete)
                {
                    context.DeleteTile(tile.X, tile.Y, tile.Zoom, tile.SourceId);
                    count++;
                }
                context.SaveChanges();

                FindOutDbSize();
                sizeToClear = actualSize - maxSize;
                if (sizeToClear > 0 && count < maxCount) return false;
            }
            return true;
        }

        /// <summary>
        /// Zjist velikost dat v databzi
        /// </summary>
        private void FindOutDbSize()
        {
            actualSize = context.GetTotalImagesSize();
        }

        private void LogCacheState()
        {
            if (logger == null) return;
            string message = string.Format("Velikost obrzku v databzi: {0} \nMaxTotalSize: {1} \nDefaultTileExpirationPeriod (Days): {2}", actualSize, maxSize, defaultExpirationPeriodDays);
            logger.LogCacheState(cacheName, context.Tiles.Count(), string.Empty);
        }

        private byte[] GetBytes(BitmapImage image)
        {
            var memory = new System.IO.MemoryStream();
            var encoder = new PngBitmapEncoder();
            encoder.Frames.Add(BitmapFrame.Create(image));
            encoder.Save(memory);
            return memory.ToArray();
        }

        private BitmapImage GetImage(byte[] data)
        {
            var image = new BitmapImage();
            image.BeginInit();
            image.StreamSource = new System.IO.MemoryStream(data);
            image.EndInit();
            image.Freeze();
            return image;
        }

        /// <param name="returnIfExpired">True, pokud maj bt vrceny dladice, jejich platnost ji vyprela</param>
        public GetCachedTileResult GetTile(string mapName, TileId tileId, bool returnIfExpired)
        {
            lock (lockObject)
            {
                var tile = context.Tiles.Include("Source").Where(t => t.Source.Name == mapName && t.X == tileId.X && t.Y == tileId.Y && t.Zoom == tileId.Zoom).FirstOrDefault();
                if (tile == null) return GetCachedTileResult.Failed(mapName, tileId);
                if (!returnIfExpired && (tile.DownloadedTime.Add(tile.Source.ExpirationPeriod) < DateTime.Now)) return GetCachedTileResult.Failed(mapName, tileId);
                return new GetCachedTileResult(mapName, tileId, GetImage(tile.Image), tile.DownloadedTime + tile.Source.ExpirationPeriod);
            }
        }

        int? id;

        public void AddToCache(string mapSource, TileId tileId, BitmapImage image, DateTime downloadTime)
        {
            lock (lockObject)
            {
                if (id == null) id = System.Threading.Thread.CurrentThread.ManagedThreadId;
                else if (id.Value != System.Threading.Thread.CurrentThread.ManagedThreadId)
                    System.Diagnostics.Debug.WriteLine("ThreadProblem");

                var source = GetOrCreateMapSource(mapSource);
                var data = GetBytes(image);
                if (data.Length > 100 * 1024 * 1024) throw new ArgumentException("Image is too large");
                var tile = context.Tiles.FirstOrDefault(t => t.X == tileId.X && t.Y == tileId.Y && t.Zoom == tileId.Zoom && t.Source.ID == source.ID);
                if (tile == null)
                {
                    tile = new TileEntity() { TileId = tileId, Image = data, DownloadedTime = downloadTime, Source = source };
                    context.Tiles.Add(tile);
                }
                else
                {
                    tile.Image = data; tile.DownloadedTime = downloadTime;
                }
                context.SaveChanges();
                IncrementSizeCounter(data.Length);
            }
        }

        private void IncrementSizeCounter(int length)
        {
            actualSize += length;
            if (actualSize > maxSize) PurgeDb();
        }

        private SourceEntity GetOrCreateMapSource(string mapSource)
        {
            var source = context.Sources.FirstOrDefault(s => s.Name == mapSource);
            if (source != null) return source;
            source = new SourceEntity();
            source.Name = mapSource;
            source.ExpirationPeriod = TimeSpan.FromDays(defaultExpirationPeriodDays);
            context.Sources.Add(source);
            context.SaveChanges();
            return source;
        }

        /// <summary>
        /// Odstran vechny data
        /// </summary>
        public void Clear()
        {
            lock (lockObject)
            {
                var toDelete = context.Tiles.Select(tile => new { tile.X, tile.Y, tile.Zoom, tile.SourceId });
                foreach (var tile in toDelete) context.DeleteTile(tile.X, tile.Y, tile.Zoom, tile.SourceId);
                context.SaveChanges();
            }
        }
    }
}