namespace Brjnk.Maps.Downloader
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Threading;

    /// <summary>
    /// Plnuje poad stahovanch soubor a uchovv jejich seznam.
    /// Thread-safe
    /// </summary>
    public class DownloadScheduler
    {
        private volatile bool downloaderWaiting = false;
        private volatile TaskUpdaterPair nextTask;
        private AutoResetEvent syncEvent = new AutoResetEvent(false);
        private PriorityQueue<TaskUpdaterPair> tasks = new PriorityQueue<TaskUpdaterPair>();

        private object lck = new Object();//zmek na funkci GetNextTask()

        private MultiTimer multiTimer;
        private IOverloadProtection overloadProtection;
        private ScheduledAction timer;

        public DownloadScheduler(IOverloadProtection overloadProtection, MultiTimer multiTimer)
        {
            this.multiTimer = multiTimer;
            this.overloadProtection = overloadProtection;
            timer = new ScheduledAction(new Action(StartDownload), null);
        }

        public void AddTask(DownloadTask task, DownloadTaskUpdater updater)
        {
            lock (tasks)
            {
                if (task == null) throw new ArgumentNullException();
                tasks.Enqueue((int)task.Priority, new TaskUpdaterPair(task, updater));

                //Upednostnit tuto lohu pokud m vt prioritu
                if (downloaderWaiting && (nextTask == null || nextTask.Task.Priority < task.Priority))
                    FindNextTask();
            }
        }

        /// <summary>
        /// Vrt nsledujc lohu ke staen a vechny dal lohy se stejnm URL. Ped vrcenm hodnoty vyk na as stahovn podle IOverloadProtection
        /// </summary>
        /// <remarks>Pokud nen k dispozici dn loha, vlkno zstane pozastaveno.</remarks>
        public IEnumerable<TaskUpdaterPair> GetNextTask()
        {
            lock (lck)
            {
                lock (tasks)
                {
                    FindNextTask(); //vyhledat dal lohu
                    downloaderWaiting = true;
                }
                syncEvent.WaitOne();

                TaskUpdaterPair toDownload;
                List<TaskUpdaterPair> otherTasks;
                lock (tasks)
                {
                    toDownload = nextTask;
                    nextTask = null;
                    otherTasks = tasks.Where(t => t.Task.Uri == toDownload.Task.Uri).ToList();
                    tasks.RemoveWhere(o => o.Task.Uri == toDownload.Task.Uri); //odstranit pedanou lohu z fronty
                }
                downloaderWaiting = false;
                return otherTasks;
            }
        }

        public TimeSpan GetSpeedRestriction(long downloadedData)
        {
            if (overloadProtection != null)
            {
                return overloadProtection.GetSpeedRestriction(downloadedData);
            }
            return TimeSpan.Zero;
        }

        public void RemoveTask(DownloadTask task)
        {
            if (task == null) throw new ArgumentNullException();
            lock (tasks)
            {
                tasks.RemoveWhere(o => object.ReferenceEquals(o.Task, task));
                //pokud je prv odstrann task dal ke staen, najde nov task
                if (nextTask != null && downloaderWaiting && object.ReferenceEquals(nextTask.Task, task)) FindNextTask();
            }
        }

        public void RemoveWithTag(object tag)
        {
            if (tag == null) throw new ArgumentNullException();
            lock (tasks)
            {
                tasks.RemoveWhere(o => object.ReferenceEquals(tag, o.Task.Tag));
                if (downloaderWaiting) FindNextTask();
            }
        }

        private void FindNextTask()
        {
            if (tasks.Count == 0) return;
            var next = tasks.Peek();
            if (next == null) { nextTask = null; return; }
            if (overloadProtection == null)
            {
                syncEvent.Set();
                nextTask = next;
                return;
            }
            var remaining = overloadProtection.GetRemainingTime(next.Task.Priority);
            //  if (downloaderWaiting == false) return; //v metod GetNextTask neek vlkno a nen tedy teba nastavovat asova

            if (downloaderWaiting && nextTask != null) //pokud downloader ek a ji b asova na dal lohu
            {
                if (multiTimer.Update(timer, TicksTime.Add(remaining.TotalMilliseconds)))
                {
                    nextTask = next;//pokud se podailo aktualizovat timer (tj. pokud jet nebyl vykonn)
                }
            }
            else //naasovat stahovn
            {
                multiTimer.Add(timer, TicksTime.Add(remaining.TotalMilliseconds));
                nextTask = next;
            }
        }

        private void StartDownload()
        {
            syncEvent.Set();
        }
    }
}