namespace Brjnk.Maps.Downloader
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Threading;
    using Brjnk.Common.Extensions;

    public class Downloader
    {
        private IDownloadLogger logger;

        private DownloadScheduler scheduler;

        private System.Net.WebClient wc = new System.Net.WebClient();

        private Thread workingThread;

        /// <remarks>Instance tto tdy vytv vlastn vlkno, ve kterm jsou data stahovna</remarks>
        public Downloader(DownloadScheduler scheduler, IDownloadLogger logger)
        {
            this.scheduler = scheduler;
            this.logger = logger;

            workingThread = new Thread(new ThreadStart(() => this.DownloadLoop()));
            workingThread.IsBackground = true;
            workingThread.Start();
        }

        ~Downloader()
        {
            if (workingThread != null) workingThread.Abort();
        }

        public event EventHandler<DownloadCompletedEventArgs> DownloadCompleted;

        public void CancelDownload(DownloadTask task)
        {
            scheduler.RemoveTask(task);
        }

        public void CancelDownload(object tag)
        {
            scheduler.RemoveWithTag(tag);
        }

        public DownloadTask Download(Uri url, DownloadTaskPriority priority)
        {
            DownloadTaskUpdater updater;
            var task = new DownloadTask(url, priority, out updater);
            scheduler.AddTask(task, updater);
            return task;
        }

        private void CompleteDownload(IEnumerable<TaskUpdaterPair> tasks, Exception ex)
        {
            foreach (var task in tasks)
            {
                task.Updater.updateData(new byte[] { });
                task.Updater.updateException(ex);
                task.Updater.updateState(DownloadTaskState.Failed);
                task.Updater.updateCompleted(true);
                if (DownloadCompleted != null)
                    DownloadCompleted(this, new DownloadCompletedEventArgs(task.Task));
            }
            if (logger != null) logger.LogDownloadCompleted(tasks.First().Task.Uri.Host, false);
        }

        private void CompleteDownload(IEnumerable<TaskUpdaterPair> tasks, System.IO.MemoryStream output)
        {
            foreach (var task in tasks)
            {
                task.Updater.updateData(output.ToArray());
                if (output.Length < task.Task.TotalBytes)
                    task.Updater.updateState(DownloadTaskState.FileIsTooShort);
                else
                    task.Updater.updateState(DownloadTaskState.Success);

                if (DownloadCompleted != null)
                    DownloadCompleted(this, new DownloadCompletedEventArgs(task.Task));
                task.Updater.updateCompleted(true);
            }
            if (logger != null) logger.LogDownloadCompleted(tasks.First().Task.Uri.Host, true);
        }

        private void DownloadLoop()
        {
            while (true)
            {
                var tasks = scheduler.GetNextTask();
                if (tasks.Count() == 0) continue;
                ProcessDownload(tasks);
            }
        }

        private void ProcessDownload(IEnumerable<TaskUpdaterPair> tasks)
        {
            System.IO.Stream input = null;
            System.IO.MemoryStream output = null;
            try
            {
                AddHttpHeaders();
                tasks.ForEach(t => t.Updater.updateState(DownloadTaskState.Initializing));
                input = wc.OpenRead(tasks.First().Task.Uri);
                output = new System.IO.MemoryStream();
                int total;
                int.TryParse(wc.ResponseHeaders["Content-Length"], out total);
                byte[] buffer = new byte[4096];
                tasks.ForEach(t => { t.Updater.updateTotalBytes(total); t.Updater.updateState(DownloadTaskState.DownloadInProgress); });
                var lastTime = DefaultTimingProvider.GetDefaultTimingProvider().GetTime();
                while (true)
                {
                    var len = input.Read(buffer, 0, buffer.Length);
                    if (len == 0)
                    {
                        CompleteDownload(tasks, output);
                        break;
                    }
                    else
                    {
                        var currentTime = DefaultTimingProvider.GetDefaultTimingProvider().GetTime();
                        UpdateProgress(tasks, len, currentTime - lastTime);
                        var restriction = scheduler.GetSpeedRestriction(len);
                        Wait(restriction);
                        output.Write(buffer, 0, len);
                        lastTime = currentTime;
                    }
                }
            }
            catch (Exception ex)
            {
                CompleteDownload(tasks, ex);
            }
            finally
            {
                if (input != null) input.Dispose();
                if (output != null) output.Dispose();
            }
        }

        private void AddHttpHeaders()
        {
            wc.Headers.Add("User-Agent", "Downloader");
            //wc.Headers.Add("User-Agent", @"Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.1 (KHTML, like Gecko) Maxthon/4.1.0.3000 Chrome/26.0.1410.43 Safari/537.1");
        }

        private void UpdateProgress(IEnumerable<TaskUpdaterPair> tasks, int length, TimeSpan timeSpan)
        {
            foreach (var task in tasks)
            {
                task.Updater.updateReceivedBytes(task.Task.ReceivedBytes + length);
                if (logger != null) logger.LogDownload(task.Task.Uri, length, timeSpan);
            }
        }

        private void Wait(TimeSpan restriction)
        {
            if (restriction.Ticks == 0) return;
            Thread.Sleep(restriction.Milliseconds);
        }

        public void Abort()
        {
            workingThread.Abort();
        }
    }
}