namespace Brjnk.Maps
{
    using System;
    using System.Collections.Generic;
    using System.ComponentModel;
    using System.Configuration;
    using System.Linq;
    using System.Xml.Serialization;
    using Brjnk.Maps.Sources;

    //Dal zdoje: http://wiki.openstreetmap.org/wiki/OpenLayers

    /// <summary>
    /// </summary>
    [SettingsSerializeAs(SettingsSerializeAs.Xml)]
    public class CustomWebMapSource : WebMapSource, INotifyPropertyChanged
    {
        private WebMapSource mapSource;

        private string sourceName;

        public string SourceNameSerializable
        {
            get { return sourceName; }
            set
            {
                sourceName = value; mapSource = null; CreateSource(); OnPropertyChanged("SourceNameSerializable");
            }
        }

        private bool isInvalid = true;

        [XmlIgnore()]
        public bool IsInvalid
        {
            get { return isInvalid; }
            private set { isInvalid = value; OnPropertyChanged("IsInvalid"); }
        }

        private string uniqueName = Guid.NewGuid().ToString();

        public string UniqueNameSerializable
        {
            get { return uniqueName; }
            set
            {
                uniqueName = value; mapSource = null; CreateSource(); OnPropertyChanged("UniqueNameSerializable");
            }
        }

        private CustomWebMapType type;

        public CustomWebMapType Type
        {
            get { return type; }
            set
            {
                type = value; mapSource = null; CreateSource(); OnPropertyChanged("Type");
            }
        }

        private int maxZoom;

        public int MaxZoom
        {
            get { return maxZoom; }
            set
            {
                maxZoom = value; mapSource = null; CreateSource(); OnPropertyChanged("MaxZoom");
            }
        }

        private System.Windows.Size tileDIPSize;

        private string urlPattern;

        public string UrlPattern
        {
            get { return urlPattern; }
            set
            {
                urlPattern = CorrectPattern(value);
                mapSource = null;
                CreateSource();
                OnPropertyChanged("UrlPattern");
            }
        }

        /// <summary>
        /// Oprav url adresu. Promn v adrese mus mt tento formt: {x}
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        private string CorrectPattern(string value)
        {
            value = value.ToLower();
            value = value.Replace("${", "{"); //pevede ${x} na {x}
            value = value.Replace("/x/", "/{x}/");
            value = value.Replace("/y/", "/{y}/");
            value = value.Replace("/z/", "/{z}/");
            value = value.Replace("[", "{");
            value = value.Replace("]", "}");
            value = value.Replace("/zoom/", "/{z}/");
            value = value.Replace("{zoom}", "{z}");
            value = value.Replace(@"$%7bz%7d", "{z}");
            value = value.Replace(@"$%7bx%7d", "{x}");
            value = value.Replace(@"$%7by%7d", "{y}");

            return value;
        }

        public System.Windows.Size TileDIPSizeSerializable
        {
            get { return tileDIPSize; }
            set
            {
                tileDIPSize = value; mapSource = null; CreateSource(); OnPropertyChanged("TileDIPSizeSerializable");
            }
        }

        [XmlIgnore()]
        public override IMapProjection Projection
        {
            get
            {
                if (IsInvalid) return new EmptyProjection();
                return mapSource.Projection;
            }
        }

        [XmlIgnore()]
        public override IEnumerable<ZoomLevel> ZoomLevels
        {
            get
            {
                if (IsInvalid) return Enumerable.Empty<ZoomLevel>();
                return mapSource.ZoomLevels;
            }
        }

        [XmlIgnore()]
        public override System.Windows.Size TileDIPSize
        {
            get
            {
                return TileDIPSizeSerializable;
            }
        }

        [XmlIgnore()]
        public override string SourceName
        {
            get
            {
                return sourceName;
            }
        }

        [XmlIgnore()]
        public override string UniqueName
        {
            get
            {
                return uniqueName;
            }
        }

        public override System.Windows.Point GetTileCoordinates(System.Windows.Point internalPoint, int zoomLevel)
        {
            if (IsInvalid) return internalPoint;
            return mapSource.GetTileCoordinates(internalPoint, zoomLevel);
        }

        public override Uri GetTileUrl(TileId tile)
        {
            if (IsInvalid) return null;
            return mapSource.GetTileUrl(tile);
        }

        public override bool IsTileAvailable(TileId tile)
        {
            if (IsInvalid) return false;
            return mapSource.IsTileAvailable(tile);
        }

        private void CreateSource()
        {
            if (mapSource != null) return;
            var isValid = Validate();
            if (!isValid) { IsInvalid = true; return; }

            switch (Type)
            {
                case CustomWebMapType.MercatorOSM:
                    mapSource = new OsmMapSource(sourceName, uniqueName, maxZoom, urlPattern);
                    break;

                default:
                    throw new Exception("Unknown map Type (CustomWebMapSource.Type)");
            }

            OnPropertyChanged("SourceName");
            OnPropertyChanged("UniqueName");
            OnPropertyChanged("TileDIPSize");
            OnPropertyChanged("ZoomLevels");
            OnPropertyChanged("Projection");
            if (Changed != null) Changed(this, EventArgs.Empty);
            IsInvalid = false;
        }

        /// <summary>
        /// Vrt True, pokud je instance validn
        /// </summary>
        private bool Validate()
        {
            return !(string.IsNullOrEmpty(sourceName) ||
                            string.IsNullOrEmpty(uniqueName) ||
                            maxZoom < 0 ||
                            string.IsNullOrEmpty(urlPattern) ||
                            !Enum.IsDefined(typeof(CustomWebMapType), Type) ||
                            !urlPattern.Contains("{x}") ||
                            !urlPattern.Contains("{y}") ||
                            !urlPattern.Contains("{z}"));
        }

        public event EventHandler Changed;

        public event PropertyChangedEventHandler PropertyChanged;

        private void OnPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }
    }

    public enum CustomWebMapType
    {
        /// <summary>
        /// Projekce Mercator pouvan OSM (open street map)
        /// </summary>
        MercatorOSM = 1
    }
}