namespace Brjnk.Maps
{
    using System;
    using System.Collections.Generic;
    using System.Windows;
    using Brjnk.Common;

    /// <summary>
    /// </summary>
    public class WebMapProvider : IMapProvider
    {
        private IMapCache cache;
        private Downloader.Downloader downloader;
        private RamCache ramCache = new RamCache(1000);
        private WebMapSource source;
        private AsyncProcessing<object> asyncProcessing = new AsyncProcessing<object>(Common.AsyncProcessingMode.ThreadPool);

        public WebMapProvider(IMapCache cache, WebMapSource source, IDownloadLogger logger)
        {
            if (source == null) throw new ArgumentNullException("source");
            this.cache = cache;
            this.source = source;
            CreateDownloader(logger);
        }

        public WebMapSource Source { get { return source; } }

        public System.Windows.Size TileDIPSize
        {
            get { return source.TileDIPSize; }
        }

        public IEnumerable<ZoomLevel> ZoomLevels
        {
            get { return source.ZoomLevels; }
        }

        public IMapProjection Projection
        {
            get { return source.Projection; }
        }

        public Point GetTileCoordinates(Point point, int zoom)
        {
            return source.GetTileCoordinates(point, zoom);
        }

        public GetTileTask GetTile(TileId tile, Downloader.DownloadTaskPriority priority, NetworkMode mode)
        {
            GetTileTaskUpdater updater;
            var result = new GetTileTask(tile, Cancel, out updater);
            asyncProcessing.StartAsync(() => GetTileInternal(tile, priority, mode, result, updater), null);
            return result;
        }

        /// <summary>
        /// Funkce zsk mapovou dladici. Tato metoda je volna z GetTile v jinm vlkn.
        /// Nvratov hodnota je vdy null. Funkce m nvratovou hodnotu, aby mohla bt vyvolna pomoc tdy AsyncProcessing.
        /// </summary>
        private object GetTileInternal(TileId tile, Downloader.DownloadTaskPriority priority, NetworkMode mode, GetTileTask result, GetTileTaskUpdater updater)
        {
            try
            {
                //Zskat z vyrovnvac pamti (RAM)
                var ramResult = ramCache.GetTile(tile);
                if (ramResult != null)
                {
                    updater.updateSource(GetTileSource.RamCache);
                    updater.completed(ramResult, null);
                    return null;
                }

                //Zskat z vyrovnavac pamti na pevnm disku
                bool expired = false;
                if (mode == NetworkMode.Offline || mode == NetworkMode.DataSaving) expired = true;
                var cachedResult = cache.GetTile(source.UniqueName, tile, expired);
                if (cachedResult.IsTileEmpty == false)
                {
                    updater.updateSource(GetTileSource.Cache);
                    updater.completed(cachedResult.Tile, null);
                    ramCache.AddTile(tile, cachedResult.Tile);
                }
                //Sthnout z internetu
                else if (mode != NetworkMode.Offline)
                {
                    var dTask = downloader.Download(source.GetTileUrl(tile), priority);
                    updater.updateSource(GetTileSource.Network);
                    updater.updateDonwloadTask(dTask);
                    result.Completed += new EventHandler(result_Completed);
                }
                else
                {
                    //Neexistuje dn zpsob jak zskat data (nap. z dvodu offline reimu)
                    updater.completed(null, new Exception("There is no possibility how to get map tile.  NetworkMode:" + mode.ToString()));
                }
                return null;
            }
            catch (Exception ex)
            {
                updater.completed(null, ex);
                return null;
            }
        }

        private void Cancel(Downloader.DownloadTask task)
        {
            downloader.CancelDownload(task);
        }

        private void CreateDownloader(IDownloadLogger logger)
        {
            Downloader.DownloadScheduler scheduler = new Downloader.DownloadScheduler(new NoOverloadProtection(), new MultiTimer());
            downloader = new Downloader.Downloader(scheduler, logger);
        }

        private void result_Completed(object sender, EventArgs e)
        {
            asyncProcessing.StartAsync(() => CompleteDownload(sender as GetTileTask), null);
        }

        /// <summary>
        /// Zpracuje staen data (ulo je do vyrovnvac pamti)
        /// Funkce je volna v jinm vlkn, nvratov hodnota je vdy null (umonuje pouit v AsyncProcessing)
        /// </summary>
        private object CompleteDownload(GetTileTask task)
        {
            task.Completed -= new EventHandler(result_Completed);

            if (task == null) System.Diagnostics.Debug.WriteLine("Task je null!!");
            else
            {
                if (task.Exception != null) return null;
                cache.AddToCache(source.UniqueName, task.TileId, task.Image, DateTime.Now);
                ramCache.AddTile(task.TileId, task.Image);
            }
            return null;
        }

        public bool IsTileAvailable(TileId id)
        {
            return source.IsTileAvailable(id);
        }

        public void Abort()
        {
            downloader.Abort();
        }
    }

    public enum NetworkMode
    {
        Online = 0,
        Offline = 1,

        /// <summary>
        /// spora dat
        /// </summary>
        DataSaving = 2,
    }
}