namespace Brjnk.Maps
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Linq;

    /// <summary>
    /// </summary>
    public class MultiTimer
    {
        /// <summary>
        /// as v milisekundch mezi spoutnm udlosti Tick();
        /// Thread-safe
        /// </summary>
        public const long TickInterval = 1000;

        private Dictionary<IScheduledAction, TicksTime> items = new Dictionary<IScheduledAction, TicksTime>();
        private object lck = new object();
        private bool tickTimerEnabled = false;
        private ITimingProvider timing;
        private ITimer tmr; //slou pro odpotvn asu do dal udlosti

        /// <param name="timer">asova pro naplnovn poloek</param>
        /// <param name="tickTimer">asova pro spoutn udlosti Tick</param>
        /// <param name="context">SynchronizationContext pro synchronizaci udlosti Tick</param>
        public MultiTimer()
        {
            this.timing = DefaultTimingProvider.GetDefaultTimingProvider();
            this.tmr = timing.GetTimer();
            tmr.Elapsed += new Action(Refresh);
        }

        /// <summary>
        /// Pid poloku do asovae. Pokud ji pedan akce v asovai je, neprovade dnou akci a vrt false.
        /// </summary>
        public bool Add(IScheduledAction action, TicksTime time)
        {
            lock (lck)
            {
                if (items.ContainsKey(action)) return false;
                items.Add(action, time);
                action.Reset(timing.GetTime(), time);
                if (action.ProgressRequired) EnableTickTimer();
                Refresh();
                return true;
            }
        }

        /// <summary>
        /// Pid pedanou akci do asovae. Pokud ji existuje, aktualizuje jej as.
        /// </summary>
        public void AddOrUpdate(IScheduledAction action, TicksTime time)
        {
            lock (lck)
            {
                if (items.ContainsKey(action))
                {
                    action.Reset(items[action], time);
                    items[action] = time;
                }
                else
                {
                    action.Reset(timing.GetTime(), time);
                    items.Add(action, time);
                    if (action.ProgressRequired) EnableTickTimer();
                }
                Refresh();
            }
        }

        /// <summary>
        /// Odebere akci z asovae. Pokud akce v asovai nen, vrt false.
        /// </summary>
        public bool Remove(IScheduledAction action)
        {
            lock (lck)
            {
                if (items.ContainsKey(action))
                {
                    items.Remove(action);
                }
                return false;
            }
        }

        /// <summary>
        /// Aktualizuje as pedan akci. Pokud pedan akce v asovai nen, neprovede dnou akci a vrt false;
        /// </summary>
        public bool Update(IScheduledAction action, TicksTime time)
        {
            lock (lck)
            {
                if (items.ContainsKey(action) == false) return false;
                action.Reset(items[action], time);
                items[action] = time;
                Refresh();
                return true;
            }
        }

        private void DisableTickTimer()
        {
            if (!tickTimerEnabled) return;
            timing.Tick -= new Action<TicksTime>(ProcessTick);
            tickTimerEnabled = false;
        }

        private void EnableTickTimer()
        {
            if (tickTimerEnabled) return;
            timing.Tick += new Action<TicksTime>(ProcessTick);
            tickTimerEnabled = true;
        }

        private void InvokeAction(KeyValuePair<IScheduledAction, TicksTime> kvp)
        {
            try
            {
                if (kvp.Key.ProgressRequired) kvp.Key.UpdateProgress(TimeSpan.Zero);
                kvp.Key.Execute(kvp.Value);
            }
            catch (Exception ex)
            {
                Trace.TraceError("Metoda IScheduledAction.Execute() vyvolala vyjmku.\n" + ex.ToString());
            }
        }

        private void ProcessTick(TicksTime time)
        {
            lock (lck)
            {
                var currentTime = timing.GetTime();
                bool noneUpdate = true;
                foreach (var kvp in items.Where(a => a.Key.ProgressRequired))
                {
                    noneUpdate = false;
                    kvp.Key.UpdateProgress(TimeSpan.FromMilliseconds(kvp.Value.Ticks - currentTime.Ticks));
                }
                if (noneUpdate) DisableTickTimer();
            }
        }

        private void Refresh()
        {
            lock (lck)
            {
                var currentTime = timing.GetTime();
                var elapsed = items.Where(p => p.Value <= currentTime).ToArray();
                foreach (var kvp in elapsed)
                {
                    InvokeAction(kvp);
                    items.Remove(kvp.Key);
                }

                UpdateTimer(currentTime);
            }
        }

        private void UpdateTimer(TicksTime currentTime)
        {
            if (items.Count == 0)
            {
                tmr.Change(-1, -1);
                return;
            }
            var min = items.Min(kvp => kvp.Value.Ticks);
            tmr.Change((min - currentTime.Ticks) / TicksTime.TicksPerMillisecond, -1);
        }
    }
}